/*=======================================================
  Author:   R. Muth
  Version:  $Id: check.c,v 1.23 1999/04/20 21:23:50 muth Exp $
  Comment:  This file contains code that checks the
            consistency of the datastructures 
=======================================================*/

#include <string.h>
#include "global.h"
#include "opcode.h"
#include "functions.h"
#include "flags.h"
#include "regs.h"
#include "system.coff.h"

/*=======================================================
=======================================================*/
#if 0
#undef __PROC__
#define __PROC__ "CheckIns"

static void CheckIns(FILE *fp)
{

    INDEX iins;
    INDEX jins;
    INDEX ibbl;
    INDEX ifun;
    
    FOREACH_LIVE_INS(iins)
    {
        ibbl = AINS_BBL(iins);
        ifun = ABBL_FUN(ibbl);

        if( !VALID(ibbl) )
        {
            ERROR(("Ins with invalid bbl\n"));
            PrintIns(fp,iins);
            continue;
        }
        else if( !VALID(ifun) )
        {
            ERROR(("Ins with invalid fun\n"));
            PrintIns(fp,iins);
            continue;
        }
           

        FOREACH_BBL_INS(ibbl,jins)
            if( jins == iins ) break;
        
        CASSERT( jins == iins,("ins %s not found in bbl: %d (forward)\n",iins,ibbl));

        FOREACH_BBL_INS_R(ibbl,jins)
            if( jins == iins ) break;
        
        CASSERT( jins == iins,("ins %s not found in bbl: %d (backward)\n",iins,ibbl));
        
        if( VALID(AINS_PREV(iins)) )
        {
            CASSERT( ABBL_INS_FIRST(ibbl) != iins,("good prev of first ins %d (b: %d)\n",iins,ibbl));
        }
        else
        {
            CASSERT( ABBL_INS_FIRST(ibbl) == iins,("bad prev of non first ins %d (b: %d)\n",iins,ibbl));
        }

        if( VALID(AINS_NEXT(iins)) )
        {
            CASSERT( ABBL_INS_LAST(ibbl) != iins,("good next of last ins %d (b: %d)\n",iins,ibbl));
        }
        else
        {
            CASSERT( ABBL_INS_LAST(ibbl) == iins,("bad next of non last ins %d (b: %d)\n",iins,ibbl));
        }
        
        
        switch( AINS_CODE(iins) )
        {
        case I_br:
            if( AINS_REG_C(iins) != REG_ZERO )
            {
                WARNING(("br with non-zero save reg %d in %d [%s]\n",AINS_REG_C(iins),ifun,AFUN_NAME(ifun)));
            }
            break;

        case I_bsr:
            if( AINS_REG_C(iins) != REG_RA )
            {
                WARNING(("bsr with non-ra save reg %d in %d [%s]\n",AINS_REG_C(iins),ifun,AFUN_NAME(ifun)));
            }
            break;

        case I_jmp:
            if( AINS_REG_C(iins) != REG_ZERO )
            {
                WARNING(("jmp with non-zero save reg %d in %d [%s]\n",AINS_REG_C(iins),ifun,AFUN_NAME(ifun)));
            }
            break;

        case I_jsr:
            if( AINS_REG_C(iins) != REG_RA )
            {
                WARNING(("jsr with non-ra save reg %d in %d [%s]\n",AINS_REG_C(iins),ifun,AFUN_NAME(ifun)));
            }

            if( AINS_REG_B(iins) != REG_T12 )
            {
                WARNING(("jsr with non-t12 jump reg %d in %d [%s]\n",AINS_REG_B(iins),ifun,AFUN_NAME(ifun)));
            }

            break;

        case I_ret:
            if( AINS_REG_C(iins) != REG_ZERO )
            {
                WARNING(("ret with non-zero save reg %d in %d [%s]\n",AINS_REG_C(iins),ifun,AFUN_NAME(ifun)));
            }
            if( AINS_REG_B(iins) != REG_RA )
            {
                WARNING(("ret with non-ra jump reg %d in %d [%s]\n",AINS_REG_B(iins),ifun,AFUN_NAME(ifun)));
            }
            break;
        }
    }

    InfoStatus(fp,"\t[CheckIns]\n");
}


static char *type[] =
{
    "R_ABS       ",
    "R_REFLONG   ",
    "R_REFQUAD   ",
    "R_GPREL32   ",
    "R_LITERAL   ",
    "R_LITUSE    ",
    "R_GPDISP    ",
    "R_BRADDR    ",
    "R_HINT      ",
    "R_SREL16    ",
    "R_SREL32    ",
    "R_SREL64    ",
    "R_OP_PUSH   ",
    "R_OP_STORE  ",
    "R_OP_PSUB   ",
    "R_OP_PRSHIFT",
    "R_GPVALUE   ",
    "R_GPRELHIGH ",
    "R_GPRELLOW  ",
    "R_IMMED     ",
};
#endif

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "CheckEdg"

static void CheckEdg(FILE *fp)
{
    INDEX iedg;

    FOREACH_LIVE_EDG(iedg)
    {
        INDEX hd = AEDG_HEAD(iedg);
        INDEX tl = AEDG_TAIL(iedg);
    
        if( ABBL_IS_KILLED(hd) )
        {
            ERROR(("live edge has dead head\n"));
            PrintEdgSuper(fp,iedg);
        }

        if( ABBL_IS_KILLED(tl) )
        {
            ERROR(("live edge has dead tail\n"));
            PrintEdgSuper(fp,iedg);
        }


        switch( AEDG_TYPE(iedg) )
        {
        case ET_FUN_CALL:
            if( ABBL_TYPE(hd) != BT_CALL && ABBL_TYPE(hd) != BT_PAL )
            {
                ERROR(("fun_call edg in  non bt_call bbl\n"));
                PrintEdgSuper(fp,iedg);
            }

            if( !ABBL_IS_FIRST(tl) )
            {
                ERROR(("call to non first bbl\n"));
                PrintEdgSuper(fp,iedg);
            }

            CERROR( ABBL_EDG_CALL(hd) != iedg,("bad br target for fun call edg\n"));

            if( AEDG_IS_KILLED(iedg+1) || AEDG_TYPE(iedg+1) != ET_FUN_RETURN ) 
            {
                ERROR(("bad return edge for live call edge killed\n"));
                PrintEdg(fp,iedg);
                PrintEdg(fp,iedg+1);
            }

            break;
                        
        case ET_FUN_RETURN:
            if( !ABBL_IS_LAST(hd) )
            {
                ERROR(("ret from non last bbl\n"));
                PrintEdgSuper(fp,iedg);
            }
            
            if( AEDG_IS_KILLED(iedg-1) || AEDG_TYPE(iedg-1) != ET_FUN_CALL )
            {
                ERROR(("call edge for live ret edge killed\n"));
                PrintEdg(fp,iedg);
                PrintEdg(fp,iedg-1);
            }

            break;
            

        case ET_FUN_LINK:
            if( ABBL_TYPE(hd) != BT_CALL && ABBL_TYPE(hd) != BT_PAL )
            {
                ERROR(("fun_link edg in  non bt_call bbl\n"));
                PrintEdgSuper(fp,iedg);
            }
            CERROR( ABBL_EDG_LINK(hd) != iedg,("bad edg_link in bt_call bbl\n"));
            break;
            
        case ET_NORM:
            if( ABBL_TYPE(hd) != BT_NORM)
            {
                ERROR(("et_norm edg in non bt_norm bbl\n"));
                PrintEdgSuper(fp,iedg);
            }
            CERROR( ABBL_EDG_NORM(hd) != iedg,("bad edg_norm in bt_norm bbl\n"));
            break;

        case ET_UNCOND:
            if( ABBL_TYPE(hd) != BT_UNCOND)
            {
                ERROR(("et_uncond edg in non bt_uncond bbl\n"));
                PrintEdgSuper(fp,iedg);
            }
            CERROR( ABBL_EDG_UNCOND(hd) != iedg,("bad edg_uncond in bt_uncond bbl\n"));
            break;

        case ET_EXIT_LINK:
            CERROR( ABBL_TYPE(hd) != BT_RET,("et_exit_link edg from non bt_ret bbl\n"));
            CERROR( ABBL_TYPE(tl) != BT_EXIT,("et_exit_link edg to non bt_exit bbl\n"));
            break;
            
        case ET_HELL:
            CERROR( ABBL_FUN(hd) != AFUN_PSEUDO_HELL &&
                    ABBL_FUN(hd) != AFUN_PSEUDO_HELL_LITE &&
                    ABBL_FUN(tl) != AFUN_PSEUDO_HELL &&
                    ABBL_FUN(tl) != AFUN_PSEUDO_HELL_LITE,
                    ("et_hell edg not to or from hell\n"));
            break;

        case ET_EXTRA:
            CERROR( ABBL_TYPE(hd) != BT_EXTRA,("et_extra edg from non bt_extra bbl\n"));
            break;

        case ET_TRUE:
            CERROR( ABBL_TYPE(hd) != BT_BRANCH,("et_true edg from non bt_branch bbl\n"));
            break;

        case ET_FALSE:
            CERROR( ABBL_TYPE(hd) != BT_BRANCH,("et_false edg from non bt_branch bbl\n"));
            break;
            
        case ET_SETLONGJMP:
            CERROR( ABBL_TYPE(hd) != BT_EXIT,("et_setlongjmp edg from non bt_exit bbl\n"));
            CERROR( ABBL_TYPE(tl) != BT_EXIT,("et_setlongjmp edg to non bt_exit bbl\n"));
            break;

        case ET_JUMP:
            CERROR( ABBL_TYPE(hd) != BT_JUMP,("et_jump edg from non bt_jump bbl\n"));
            break;
            
        case ET_SWITCH:
            CERROR( ABBL_TYPE(hd) != BT_SWITCH,("et_switch edg from non bt_switch bbl\n"));
            CERROR( !(ABBL_FLAGS(tl) & F_GPREL32),("et_switch(%d,%d) edg to non gprel32 bbl\n",hd,tl));
            break;

        case ET_COMPENSATE:
            CERROR( !ABBL_IS_LAST(hd),("bad head bbl in compensate-edg\n"));
            CERROR( !ABBL_IS_LAST(tl),("bad tail bbl in compensate-edg\n"));
            CERROR( AEDG_IS_KILLED(iedg-1),("compensating dead edge\n"));
            if( ABBL_FUN(hd) == ABBL_FUN(tl) )
            {
                ERROR(("unnecessary compensation edg\n"));
                PrintEdgSuper(fp,iedg-1);
                PrintEdgSuper(fp,iedg);
            }
            
            CERROR( !ABBL_IS_LAST(hd),("head bbl not last\n"));
            CERROR( !ABBL_IS_LAST(tl),("tail bbl not last\n"));
            break;
            
        default:
            ERROR(("unknown edge type\n"));
            PrintEdgSuper(fp,iedg);
        }

        
        if( ABBL_FUN(hd) != ABBL_FUN(tl) &&
            AEDG_TYPE(iedg) != ET_FUN_CALL &&
            AEDG_TYPE(iedg) != ET_FUN_RETURN &&
            AEDG_TYPE(iedg) != ET_COMPENSATE )
        {
            if( AEDG_IS_KILLED(iedg+1) || AEDG_TYPE(iedg+1) != ET_COMPENSATE ) 
            {
                ERROR(("bad compensate-edg for live interprocedural edg \n"));
                PrintEdg(fp,iedg);
                PrintEdg(fp,iedg+1);
            }
        }
        

    }

    InfoStatus(fp,"\t[CheckEdg]\n");
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "CheckBblLastIns"

static void CheckBblLastIns(FILE *fp)
{
    INDEX ibbl;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        INDEX iins = ABBL_INS_LAST(ibbl);
        
        if( !VALID(iins) )
        {
            if( ABBL_TYPE(ibbl) != BT_NORM && ABBL_TYPE(ibbl) != BT_EXIT )
            {
                ERROR(("empty bbl is not bt_norm or bt_exit\n"));
                PrintBblSuper(stdout,ibbl);
            }
            
            continue;
        }

        switch( AINS_TYPE(iins) )
        {
        case IT_PAL:
            if( AINS_DATA(iins) == 0 )
            {
                CERROR( ABBL_TYPE(ibbl) != BT_HALT,("bt_halt expected\n"));
                CERROR( VALID(ABBL_EDG_TRUE(ibbl)),("bt_halt with t edg\n"));
                CERROR( VALID(ABBL_EDG_FALSE(ibbl)),("bt_halt with f edg\n"));
            }
            else
            {
                CERROR( ABBL_TYPE(ibbl) != BT_PAL,("bt_pal expected\n"));
                CERROR( !VALID(ABBL_EDG_CALL(ibbl)),("bt_pal without t edg\n"));
                CERROR( !VALID(ABBL_EDG_LINK(ibbl)),("bt_pal withpout f edg\n"));
            }
            break;

        case IT_IBR:
        case IT_FBR:
            CERROR( ABBL_TYPE(ibbl) != BT_BRANCH,("bt_branch expected\n"));
            CERROR( !VALID(ABBL_EDG_TRUE(ibbl)),("bt_branch without t edg\n"));
            CERROR( !VALID(ABBL_EDG_FALSE(ibbl)),("bt_branch without f edg\n"));
            break;
            
        case IT_JMP:
            switch( AINS_CODE(iins) )
            {
            case I_jsr:
                CERROR( ABBL_TYPE(ibbl) != BT_CALL,("bt_call expected\n"));
                CERROR( !VALID(ABBL_EDG_TRUE(ibbl)),("bt_call without t edg\n"));
                CERROR( !VALID(ABBL_EDG_FALSE(ibbl)),("bt_call without f edg\n"));
                break;

            case I_jmp:
                if( ABBL_TYPE(ibbl) == BT_JUMP )
                {
                    CERROR( VALID(ABBL_EDG_FALSE(ibbl)),("bt_jump with f edg\n"));
                }
                else if( ABBL_TYPE(ibbl) == BT_SWITCH )
                {
                    CERROR( VALID(ABBL_EDG_TRUE(ibbl)),("bt_switch with t edg\n"));
                    CERROR( VALID(ABBL_EDG_FALSE(ibbl)),("bt_switch with f edg\n"));
                }
                else
                {
                    ERROR(("bt_jump or bt_switch expected\n"));
                }
                break;

            case I_ret:
                CERROR( ABBL_TYPE(ibbl) != BT_RET,("bt_ret expected\n"));                
                break;

            default:
                ERROR(("I_jmp_co encountered\n"));
                PrintIns(fp,iins);
                break;
            }
            break;

        case IT_BRA:
            switch( AINS_CODE(iins) )
            {
            case I_bsr:
                CERROR( ABBL_TYPE(ibbl) != BT_CALL,("bt_call expected\n"));
                CERROR( !VALID(ABBL_EDG_TRUE(ibbl)),("bt_call without t edg\n"));
                CERROR( !VALID(ABBL_EDG_FALSE(ibbl)),("bt_cal without f edg\n"));
                if( !VALID(ABBL_EDG_TRUE(ibbl)) )
                    PrintBblSuper(stdout,ibbl);
                break;

            case I_br:
                CERROR( ABBL_TYPE(ibbl)!=BT_UNCOND && ABBL_TYPE(ibbl)!=BT_EXTRA,("bt_uncond or bt_extra expected\n"));
                break;

            default:
                ERROR(("unknown branch instruction encountered\n"));
                PrintIns(fp,iins);
                break;
            }
            break;
            
            
        default:
            if( ABBL_TYPE(ibbl) != BT_NORM )
            {
                ERROR(("unexpected instruction\n"));
                PrintIns(stdout,iins);
                PrintBblSuper(stdout,ibbl);
            }
            break;
        }
    }

    InfoStatus(fp,"\t[CheckBblLastIns]\n");
}

    
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "CheckBbl"

static void CheckBbl(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Checkfun"); 
    INDEX ibbl;
    INDEX iedg;
    INDEX iins;
    INDEX ifun;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        ifun = ABBL_FUN(ibbl);
        
        if( AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) )
        {
            if( ABBL_HAS_INSS(ibbl) )
            {
                ERROR(("nonempty bbl in AFUN_HELL\n"));
                if( verbose )
                    PrintBblSuper(fp,ibbl);
            }
            
        }
        
        if( ABBL_FLAGS(ibbl) & F_REFQUAD )
        {
            if( !ABBL_IS_FIRST(ibbl) )
            {
                WARNING(("Non first block with refquad: %d in fun %d [%s]\n",
                         ibbl,ifun,AFUN_NAME(ifun)));
                if( verbose )
                PrintBblSuper(fp,ibbl);
            }
        }

        if( ABBL_IS_LAST(ibbl) )
        {
            if( VALID(ABBL_COUNT(ibbl)) )
            {
                ERROR(("exit bbl with valid count field\n"));
                if( verbose )
                    PrintBbl(fp,ibbl);
            }
        }
        
        else
        {
            if( CONFIG_USE_PROFILE )
                if( !VALID(ABBL_COUNT(ibbl)) )
                {
                    ERROR(("bbl with invalid count field, %d\n", ibbl));
                    if( verbose )
                        PrintBbl(fp,ibbl);
                }
        }
            

        
        FOREACH_BBL_PRED(ibbl,iedg)
        {
            if( AEDG_IS_KILLED(iedg) )
            {
                ERROR(("killed pred edge\n"));
                PrintEdgSuper(fp,iedg);
            }
                
            if( !VALID(AEDG_HEAD(iedg) ) )
            {
                ERROR(("bad pred edge head\n"));
                PrintEdgSuper(fp,iedg);
            }
                    
            if( AEDG_TAIL(iedg)!=ibbl )
            {
                ERROR(("bad pred edge tail\n"));
                PrintEdgSuper(fp,iedg);
            }
        }
            
            
        FOREACH_BBL_SUCC(ibbl,iedg)
        {
            if( AEDG_IS_KILLED(iedg) )
            {
                ERROR(("killed succ edge\n"));
                PrintEdgSuper(fp,iedg);
            }
                
            if( AEDG_HEAD(iedg)!=ibbl )
            {
                ERROR(("bad succ edge head\n"));
                PrintEdgSuper(fp,iedg);
            }
                    
            if( !VALID(AEDG_TAIL(iedg)) )
            {
                ERROR(("bad succ edge tail\n"));
                PrintEdgSuper(fp,iedg);
            }
        }

        FOREACH_BBL_INS(ibbl,iins)
        {    
            CERROR(AINS_IS_KILLED(iins),("Killed (i %d,b %d)\n",iins,ibbl));
            CERROR(!VALID(AINS_REG_A(iins)),("Bad reg A (i %d,b %d)\n",iins,ibbl));
            CERROR(!VALID(AINS_REG_B(iins)),("Bad reg B (i %d,b %d)\n",iins,ibbl));
            CERROR(!VALID(AINS_REG_C(iins)),("Bad reg C (i %d,b %d)\n",iins,ibbl));
            if( AINS_BBL(iins) != ibbl )
            {
                ERROR(("Wrong bbl in ins %d, is %d must be %d\n",iins,AINS_BBL(iins),ibbl));
                PrintIns(fp,iins);
                PrintBbl(fp,ibbl);
                PrintBbl(fp,AINS_BBL(iins));
            }
            
        }

        FOREACH_BBL_INS_R(ibbl,iins)
        {    
            CERROR(AINS_IS_KILLED(iins),("Killed (i %d,b %d)\n",iins,ibbl));
            CERROR(!VALID(AINS_REG_A(iins)),("Bad reg A (i %d,b %d)\n",iins,ibbl));
            CERROR(!VALID(AINS_REG_B(iins)),("Bad reg B (i %d,b %d)\n",iins,ibbl));
            CERROR(!VALID(AINS_REG_C(iins)),("Bad reg C (i %d,b %d)\n",iins,ibbl));
            if( AINS_BBL(iins) != ibbl )
            {
                ERROR(("Wrong bbl in ins %d, is %d must be %d\n",iins,AINS_BBL(iins),ibbl));
                PrintIns(fp,iins);
                PrintBbl(fp,ibbl);
                PrintBbl(fp,AINS_BBL(iins));
            }
            
        }

        CERROR( VALID(ABBL_INS_LAST(ibbl)) && !VALID(ABBL_INS_FIRST(ibbl)),
                ("valid last ins, invalid first ins in bbl %d\n",ibbl));
        
        CERROR( !VALID(ABBL_INS_LAST(ibbl)) && VALID(ABBL_INS_FIRST(ibbl)),
                ("invalid last ins, valid first ins in bbl %d\n",ibbl));

    }

    InfoStatus(fp,"\t[Checkbbl]\n");
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "CheckFun"

static void CheckFun(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Checkfun"); 
    INDEX ibbl;
    INDEX ifun;
    
    FOREACH_LIVE_FUN(ifun)
    {
        if( !VALID(AFUN_BBL_LAST(ifun)) )
        {
            ERROR(("invalid last bbl in live fun %d [%s]\n",ifun,AFUN_NAME(ifun)));
            if( verbose )
                PrintFun(fp,ifun);
        }
        if( !VALID(AFUN_BBL_FIRST(ifun)) )
        {
            ERROR(("invalid first bbl in live fun %d [%s]\n",ifun,AFUN_NAME(ifun)));
            if( verbose )
                PrintFun(fp,ifun);
        }
        
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            if( ABBL_FUN(ibbl) != ifun )
            {
                ERROR(("bad fun in  bbl %d in live fun %d [%s]\n",ifun,AFUN_NAME(ifun)));
                if( verbose )
                    PrintBbl(fp,ibbl);
            }
            
            if( ABBL_IS_KILLED(ibbl) )
            {
                ERROR(("Killed bbl %d in live fun %d [%s]\n",ibbl,ifun,AFUN_NAME(ifun)));
                if( verbose )
                    PrintBbl(fp,ibbl);
            }
        }
    }

    InfoStatus(fp,"\t[Checkfun]\n");
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "CheckTab"

static void CheckTab(FILE *fp)
{
    INDEX ibbl;
    INDEX jbbl;
    INDEX itab;
    INDEX jins;
    INDEX  isec = SysGetSwitchSection();
    NUMBER icount = (Global.SwitchStart - ASEC_START(isec))/4;
    INT32 data;
    INT64 address;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_SWITCH ) continue;
        FOREACH_BBL_TAB(ibbl,itab)
        {
            jbbl = ATAB_TARGET(itab);
            
            jins = AbblGetRelocIns(jbbl);

            CASSERT( ABBL_FLAGS(jbbl) & F_GPREL32,("bt_switch %d in fun %d [%s] target bbl %d is not F_GPREL32\n",
                                                   ibbl,ABBL_FUN(ibbl),AFUN_NAME(ABBL_FUN(ibbl)),jbbl));
            /*
            CASSERT( AINS_FLAGS(jins) & F_GPREL32,("bt_switch %d in fun %d [%s] target ins %d is not F_GPREL32\n",
                                                   ibbl,ABBL_FUN(ibbl),AFUN_NAME(ABBL_FUN(ibbl)),jins));
                                                   */

            data = AsecReadData32(isec,itab+icount);
            address = SysGetGlobalPointer() + data;
            if( Ins2Addr(jins) == data )
            {
                
                ERROR(("bad table entry in bbl: %d\n",ibbl));
            }
        }
    }

    InfoStatus(fp,"\t[CheckTab]\n");
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "CheckAll"

extern void CheckAll(FILE *fp)
{
    /*
    CheckIns(fp);
    */
    CheckEdg(fp);
    CheckBbl(fp);
    CheckFun(fp);
    CheckBblLastIns(fp);
    CheckTab(fp);

}

/*=======================================================
=======================================================*/




