/*=======================================================
  Author:   R. Muth
  Version:  $Id: global.c,v 1.57 1999/04/13 20:55:48 saw Exp $
  Comment:  This file contains the primitives operating
            on the datastructures
=======================================================*/

#include "global.h"
#include "functions.h"
#include "opcode.h"
#include "flags.h"
#include "regs.h"
#include <string.h>

#include <stdarg.h>

/*=======================================================
=======================================================*/
static NUMBER TheMagic = 6666666;

extern NUMBER GetMagic()
{
    return TheMagic++;
}

/*=======================================================
@FUN
Tests for occurence of pattern in test
@END
=======================================================*/
extern BOOL HasOption(STRING text, STRING pattern)
{
    return (0 != strstr(text,pattern));
}

/*=======================================================
=======================================================*/

extern void Error(char *mess, ... )
{
    va_list ap;
    
    va_start(ap,mess);
    vfprintf(stdout,mess,ap);
    fflush(stdout);
    va_end(ap);
}
/*=======================================================
=======================================================*/

extern void InfoPhase(FILE *fp,char *mess, ... )
{
    va_list ap;
    
    va_start(ap,mess);

    
    if( HasOption(PRINT_OPT_VERBOSE,"Phase") )
    {
        PrintSep2(fp);
        vfprintf(fp,mess,ap);
        fprintf(fp,"\n");
        PrintSep2(fp);
    }
    else
    {
        fprintf(fp,"[");
        vfprintf(fp,mess,ap);
        fprintf(fp,"]\n");
    }
    
    fflush(stdout);
    va_end(ap);
}


/*=======================================================
=======================================================*/

extern void InfoStatus(FILE *fp,char *mess, ... )
{
    va_list ap;
    
    va_start(ap,mess);

    
    if( HasOption(PRINT_OPT_VERBOSE,"Status") )
        vfprintf(fp,mess,ap);
    fflush(stdout);
    va_end(ap);
}

/*=======================================================
=======================================================*/

extern void InfoExtra(FILE *fp,char *mess, ... )
{
    va_list ap;
    
    va_start(ap,mess);
    vfprintf(fp,mess,ap);
    fflush(stdout);
    va_end(ap);
}

/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AstkInit"

static void AstkInit(INDEX istk)
{
    CASSERT( VALID(istk),("Bad argument stk: %d\n",istk));

    ASTK_OFFSET(istk) = INVALID;
    ASTK_FLAGS(istk) = 0;
    ASTK_READ_REG(istk) = INVALID;
    ASTK_WRITE_REG(istk) = INVALID;
    ASTK_READ_TYPE(istk) = INVALID;
    ASTK_WRITE_TYPE(istk) = INVALID;
    ASTK_WRITE_NUM(istk) = 0;
    ASTK_READ_NUM(istk) = 0;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AtabInit"

static void AtabInit(INDEX itab)
{
    CASSERT( VALID(itab),("Bad argument tab: %d\n",itab));

    ATAB_TARGET(itab) = INVALID;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblInit"

static void AbblInit(INDEX ibbl)
{
    CASSERT( VALID(ibbl),("Bad argument bbl: %d\n",ibbl));
    
    ABBL_FLAGS(ibbl) = 0;
    /*
    ABBL_OLD_ADDR(ibbl) = INVALID;
    */
/*     ABBL_NEW_ADDR(ibbl) = INVALID;                 */
    ABBL_JUMP_TAB(ibbl) = INVALID;
    ABBL_SUCC(ibbl) = INVALID;                   
    ABBL_PRED(ibbl) = INVALID;                   
    ABBL_PREV(ibbl)= INVALID;                    
    ABBL_NEXT(ibbl)= INVALID;             
    ABBL_FUN(ibbl) = INVALID;                    
    ABBL_INS_FIRST(ibbl) = INVALID;              
    ABBL_INS_LAST(ibbl)= INVALID;                
    ABBL_PRE_DOM(ibbl) = INVALID;                
    ABBL_LOOP_HEADER(ibbl) = INVALID;
    ABBL_LOOP_NEXT_MEMBER(ibbl) = INVALID;
    ABBL_LOOP_FIRST_MEMBER(ibbl) = INVALID;
    ABBL_POST_DOM(ibbl) = INVALID;               
    ABBL_EDG_TRUE(ibbl) = INVALID;              
    ABBL_EDG_FALSE(ibbl) = INVALID;               
    ABBL_TYPE(ibbl) = INVALID;
    
    ABBL_OUT(ibbl) = INVALID;
    
    ABBL_COUNT(ibbl) = INVALID;
}
/*=======================================================
@FUN
obsolete
@END
=======================================================*/
#if 0
#undef __PROC__
#define __PROC__ "AbblIsFallthru"

extern BOOL  AbblIsFallthru(INDEX ibbl)
{
    INDEX index;
    
    index = ABBL_PREV(ibbl);
    if( !VALID(index) ) return 0;

    index = ABBL_FALL_THRU(index);
    if( !VALID(index) ) return 0;

    index = AEDG_TAIL(index);
    if( index != ibbl ) return 0;

    return 1;
}
#endif
/*=======================================================
@FUN
Duplicate a basic block including the instructions therein.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblDup"

extern INDEX AbblDup(INDEX ibbl)
{
    INDEX iins;
    INDEX jbbl = AbblNew();

    CASSERT( VALID(ibbl) && !ABBL_IS_KILLED(ibbl),("Bad argument bbl: %d\n",ibbl));

    if( ABBL_TYPE(ibbl) == BT_JUMP )
    {
        ERROR(("duplicating bt_jump bbl %d\n",ibbl));
    }

    
    AbblInit(jbbl);
    ABBL_TYPE(jbbl) = ABBL_TYPE(ibbl);
    ABBL_FLAGS(jbbl) = ABBL_FLAGS(ibbl) & (FB_CRIT_PATH|FB_CRITICAL);
    ABBL_COUNT(jbbl) = ABBL_COUNT(ibbl);
    ABBL_OUT(jbbl) = ABBL_OUT(ibbl);
    
    /* we do not copy a lot especially not ABBL_FUN */
       
    FOREACH_BBL_INS(ibbl,iins)
    {
        INDEX jins = AinsDup(iins);
        AINS_BBL(jins) = jbbl;
        AinsAppend(jins);
    }
    
    return jbbl;
}

/*=======================================================
@FUN
Eliminate successor edges from ibbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblKillSuccEdges"

extern void AbblKillSuccEdges(INDEX ibbl)
{
    INDEX iedg;
                                 
    FOREACH_BBL_SUCC(ibbl,iedg)
    {
        if( AEDG_TYPE(iedg) == ET_FUN_LINK )
        {
            continue;
        }
        else
        {
            AedgKillFancy(iedg);
        }
    }

    /* now kill the link edges */
    
    FOREACH_BBL_SUCC(ibbl,iedg)
    {
        AedgKillFancy(iedg);
    }
}

/*=======================================================
@FUN
Move successor edges from jbbl to ibbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblMoveSuccEdges"

extern void AbblMoveSuccEdges(INDEX jbbl,INDEX ibbl)
{
    INDEX iedg;
    INDEX jedg;
    
    CASSERT( ABBL_TYPE(jbbl) != BT_SWITCH,("not implemented\n"));

    /* do link edges first */
    
    FOREACH_BBL_SUCC(jbbl,iedg)
    {
        if( AEDG_TYPE(iedg) != ET_FUN_LINK ) continue;
        jedg = AedgNewFancy(ibbl,AEDG_TAIL(iedg),ET_FUN_LINK);
        AedgKillFancy(iedg);
        AedgLinkFancy(jedg);
    }

    /* now the rest */

    FOREACH_BBL_SUCC(jbbl,iedg)
    {
        jedg = AedgNewFancy(ibbl,AEDG_TAIL(iedg),AEDG_TYPE(iedg));
        AedgKillFancy(iedg);
        AedgLinkFancy(jedg);
    }
}


/*=======================================================
@FUN
Eliminate predecessor edges from ibbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblKillPredEdges"

extern void AbblKillPredEdges(INDEX ibbl)
{
    INDEX iedg;
                                 
    FOREACH_BBL_PRED(ibbl,iedg)
    {
        if( AEDG_TYPE(iedg) == ET_FUN_LINK )
        {
            continue;
        }
        else if( AEDG_TYPE(iedg) == ET_FUN_RETURN )
        {
            AedgKillFancy(iedg-1);
        }
        else
        {
            AedgKillFancy(iedg);
        }
    }

    /* kill link edges */
    FOREACH_BBL_PRED(ibbl,iedg)
    {
        AedgKillFancy(iedg);
    }
}
/*=======================================================
@FUN
Move predecessor edges from jbbl to ibbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblMovePredEdges"

extern void AbblMovePredEdges(INDEX jbbl,INDEX ibbl)
{
    INDEX iedg;
    INDEX jedg;

    /* do link edges first */
    
    FOREACH_BBL_PRED(jbbl,iedg)
    {
        if( AEDG_TYPE(iedg) != ET_FUN_LINK ) continue;
        jedg = AedgNewFancy(AEDG_HEAD(iedg),ibbl,ET_FUN_LINK);
        AedgKillFancy(iedg);
        AedgLinkFancy(jedg);
    }

    /* now the rest */
    
    FOREACH_BBL_PRED(jbbl,iedg)
    {
        if( AEDG_TYPE(iedg) == ET_FUN_RETURN )
        {
            jedg = AedgNewFancy(AEDG_HEAD(iedg-1),AEDG_TAIL(iedg-1),AEDG_TYPE(iedg-1));
            AedgKillFancy(iedg-1);
            AedgLinkFancy(jedg);
        }
        else
        {
            jedg = AedgNewFancy(AEDG_HEAD(iedg),ibbl,AEDG_TYPE(iedg));
            AedgKillFancy(iedg);
            AedgLinkFancy(jedg);
        }
    }
}

/*=======================================================
@FUN
obsolete
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AstkNew"

extern INDEX AstkNew()
{
    CFATAL(CONFIG_USTK >= CONFIG_NSTK,("not enough stack\n"));
    AstkInit(CONFIG_USTK);
    return(CONFIG_USTK++);
}
/*=======================================================
@FUN
Allocate a new entry for a jump table (see BT_SWITCH)
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AtabNew"

extern INDEX AtabNew()
{
    CFATAL(CONFIG_UTAB >= CONFIG_NTAB,("not enough tab\n"));
    AtabInit(CONFIG_UTAB);
    return(CONFIG_UTAB++);
}
/*=======================================================
@FUN
Allocate a new basic block
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblNew"


extern INDEX AbblNew()
{
    CFATAL(CONFIG_UBBL >= CONFIG_NBBL,("not enough basic blocks\n"));
    AbblInit(CONFIG_UBBL);
    return(CONFIG_UBBL++);
}

/*=======================================================
@FUN
Returns the number of incoming edges into ibbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblNpreds"

extern NUMBER AbblNpreds(INDEX ibbl)
{
    INDEX  iedg;
    NUMBER i=0;
    CASSERT( VALID(ibbl),("Bad argument bbl: %d\n",ibbl));
    FOREACH_BBL_PRED(ibbl,iedg) i++;
    return i;
}

/*=======================================================
@FUN
Returns the number of outgoing edges from ibbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblNsuccs"

extern NUMBER AbblNsuccs(INDEX ibbl)
{
    INDEX  iedg;
    NUMBER i=0;
    CASSERT( VALID(ibbl),("Bad argument bbl: %d\n",ibbl));
    FOREACH_BBL_SUCC(ibbl,iedg) i++;
    return i;
}
/*=======================================================
@FUN
Append iins to the end of the bbl AINS_BBL(iins) 
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsAppend"

extern void AinsAppend(INDEX iins)
{
    INDEX ibbl;
    
    CASSERT( VALID(iins) && !AINS_IS_KILLED(iins),("Bad argument ins: %d\n",iins));
    ibbl = AINS_BBL(iins);
    CASSERT( VALID(ibbl) && !ABBL_IS_KILLED(ibbl),("invalid ins bbl: %d\n",ibbl));

    if( !VALID(ABBL_INS_LAST(ibbl)) )
    {
        ABBL_INS_LAST(ibbl) = iins;
        ABBL_INS_FIRST(ibbl) = iins;
        AINS_NEXT(iins) = INVALID;
        AINS_PREV(iins) = INVALID;
    }
    else
    {
        AINS_NEXT(iins) = INVALID;
        AINS_PREV(iins) = ABBL_INS_LAST(ibbl);
        AINS_NEXT(ABBL_INS_LAST(ibbl)) = iins;
        ABBL_INS_LAST(ibbl) = iins;
    }
}
/*=======================================================
@FUN
Prepend iins to the beginning of the bbl AINS_BBL(iins) 
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsPrepend"

extern void AinsPrepend(INDEX iins)
{
    INDEX ibbl;
    
    CASSERT( VALID(iins) && !AINS_IS_KILLED(iins),("Bad argument ins: %d\n",iins));
    ibbl = AINS_BBL(iins);
    CASSERT( VALID(ibbl) && !ABBL_IS_KILLED(ibbl),("invalid ins bbl: %d\n",ibbl));

    if( !VALID(ABBL_INS_FIRST(ibbl)) )
    {
        ABBL_INS_LAST(ibbl) = iins;
        ABBL_INS_FIRST(ibbl) = iins;
        AINS_NEXT(iins) = INVALID;
        AINS_PREV(iins) = INVALID;
    }
    else
    {
        AINS_PREV(iins) = INVALID;
        AINS_NEXT(iins) = ABBL_INS_FIRST(ibbl);
        AINS_PREV(ABBL_INS_FIRST(ibbl)) = iins;
        ABBL_INS_FIRST(ibbl) = iins;
    }
}
/*=======================================================
@FUN
Insert iins before next
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsInsertBefore"

extern void AinsInsertBefore(INDEX iins, INDEX next)
{
    INDEX ibbl;
    INDEX prev;
        
    CASSERT( VALID(iins) && !AINS_IS_KILLED(iins),("Bad argument iins: %d\n",iins));
    CASSERT( VALID(next) && !AINS_IS_KILLED(next),("Bad argument iins: %d\n",iins));
    CASSERT( AINS_BBL(iins) == AINS_BBL(next),("two iins disagree in bbl\n"));
    ibbl = AINS_BBL(iins);
    CASSERT( VALID(ibbl) && !ABBL_IS_KILLED(ibbl),("invalid ins bbl: %d\n",ibbl));

    prev = AINS_PREV(next);
    if( !VALID(prev) )
    {
        CASSERT( ABBL_INS_FIRST(ibbl) == next,("weird\n"));
        ABBL_INS_FIRST(ibbl) = iins;
    }
    else
    {
        AINS_NEXT(prev) = iins;
    }

    AINS_PREV(iins) = prev;
    AINS_NEXT(iins) = next;
    AINS_PREV(next) = iins;
}
/*=======================================================
@FUN
Insert iins after prev
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsInsertAfter"

extern void AinsInsertAfter(INDEX iins, INDEX prev)
{
    INDEX ibbl;
    INDEX next;
        
    CASSERT( VALID(iins) && !AINS_IS_KILLED(iins),("Bad argument iins: %d\n",iins));
    CASSERT( VALID(prev) && !AINS_IS_KILLED(prev),("Bad argument iins: %d\n",iins));
    CASSERT( AINS_BBL(iins) == AINS_BBL(prev),("two iins disagree in bbl\n"));
    ibbl = AINS_BBL(iins);
    CASSERT( VALID(ibbl) && !ABBL_IS_KILLED(ibbl),("invalid ins bbl: %d\n",ibbl));

    next = AINS_NEXT(prev);
    if( !VALID(next) )
    {
        CASSERT( ABBL_INS_LAST(ibbl) == prev,("weird\n"));
        ABBL_INS_LAST(ibbl) = iins;
    }
    else
    {
        AINS_PREV(next) = iins;
    }

    AINS_NEXT(iins) = next;
    AINS_PREV(iins) = prev;
    AINS_NEXT(prev) = iins;
}

/*=======================================================
@FUN
Remove iins from AINS_BBL(iins)

Do not use this unless you know what you are doing
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsUnlink"

extern void AinsUnlink(INDEX iins)
{
    INDEX ibbl;
    
    CASSERT( VALID(iins) && !AINS_IS_KILLED(iins),("Bad argument ins: %d\n",iins));
    ibbl = AINS_BBL(iins);
    if( !VALID(ibbl) || ABBL_IS_KILLED(ibbl) )
    {
        ERROR(("invalid ins bbl: %d\n",ibbl));
        PrintIns(stdout,iins);
        PrintBblSuper(stdout,ibbl);
        
    }

    if( VALID(AINS_NEXT(iins)) )
    {
        CASSERT( ABBL_INS_LAST(ibbl) != iins,("ins with valid next is last: %d (b: %d)\n",iins,ibbl));
        AINS_PREV(AINS_NEXT(iins)) = AINS_PREV(iins);
    }
    else
    {
        CASSERT( ABBL_INS_LAST(ibbl) == iins,("ins with invalid next is not last: %d (b: %d)\n",iins,ibbl));
        ABBL_INS_LAST(ibbl) = AINS_PREV(iins);
    }

    if( VALID(AINS_PREV(iins)) )
    {
        CASSERT( ABBL_INS_FIRST(ibbl) != iins,("ins with valid pred is first: %d (b: %d)\n",iins,ibbl));
        AINS_NEXT(AINS_PREV(iins)) = AINS_NEXT(iins);
    }
    else
    {
        CASSERT( ABBL_INS_FIRST(ibbl) == iins,("ins with invalid pred is not first: %d (b: %d)\n",iins,ibbl));
        ABBL_INS_FIRST(ibbl) = AINS_NEXT(iins);
    }
}
/*=======================================================
@FUN
Kill instructions.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsKill"

extern void AinsKill(INDEX iins)
{
    CASSERT( VALID(iins) && !AINS_IS_KILLED(iins),("Bad argument ins: %d\n",iins));
    AinsUnlink(iins);

    AINS_FLAGS(iins) |= F_KILLED;
#if 1
    AINS_CODE(iins) = INVALID;
    AINS_REG_A(iins) = AINS_REG_B(iins) = AINS_REG_C(iins) = REG_ZERO;
    AINS_FLAGS(iins) |= FI_CHANGED;
#endif

}

/*=======================================================
@FUN
Append  ibbl to the end of  ABBL_FUN(ibbl)

Do not use this unless you know what you are doing
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AbblAppend"

extern void AbblAppend(INDEX ibbl)
{
    INDEX ifun;
    
    CASSERT( VALID(ibbl) && !ABBL_IS_KILLED(ibbl),("Bad argument bbl: %d\n",ibbl));
    ifun = ABBL_FUN(ibbl);
    CASSERT( VALID(ifun) && !AFUN_IS_KILLED(ifun),("invalid bbl fun: %d\n",ifun));

    /*
    PrintBbl(stdout,ibbl);
    PrintFun(stdout,ifun);
    */
    
    if( !VALID(AFUN_BBL_LAST(ifun)) )
    {
        AFUN_BBL_LAST(ifun) = ibbl;
        AFUN_BBL_FIRST(ifun) = ibbl;
        ABBL_PREV(ibbl) = INVALID;
        ABBL_NEXT(ibbl) = INVALID;
    }
    else
    {
        ABBL_NEXT(ibbl) = INVALID;
        ABBL_PREV(ibbl) = AFUN_BBL_LAST(ifun);
        ABBL_NEXT(AFUN_BBL_LAST(ifun)) = ibbl;
        AFUN_BBL_LAST(ifun) = ibbl;
    }
}

/*=======================================================
@FUN
Remove ibbl from ABBL_FUN(ibbl)

Do not use this unless you know what you are doing
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblUnlink"

extern void AbblUnlink(INDEX ibbl)
{
    INDEX ifun;

    CASSERT( VALID(ibbl) && !ABBL_IS_KILLED(ibbl),("Bad argument bbl: %d\n",ibbl));
    ifun = ABBL_FUN(ibbl);
    CASSERT( VALID(ifun) && !AFUN_IS_KILLED(ifun),("invalid bbl fun: %d\n",ifun));

    if( !VALID(ABBL_PREV(ibbl)) )
    {
        CASSERT( AFUN_BBL_FIRST(ifun) == ibbl,("bbl with invalid pred is not first: %d (f: %d)\n",ibbl,ifun));
        if( AFUN_BBL_FIRST(ifun) != ibbl )
        {
            PrintBblSuper(stdout,ibbl);
            PrintFunSuper(stdout,ifun);
        }
        
        AFUN_BBL_FIRST(ifun) = ABBL_NEXT(ibbl);
        
    }
    else
    {
        CASSERT( AFUN_BBL_FIRST(ifun) != ibbl,("bbl with valid pred is first: %d (f: %d)\n",ibbl,ifun));
        ABBL_NEXT(ABBL_PREV(ibbl)) = ABBL_NEXT(ibbl);
    }
    
    if( !VALID(ABBL_NEXT(ibbl)) )
    {
        CASSERT( AFUN_BBL_LAST(ifun) == ibbl,("bbl with invalid next is not last: %d\n (f: %d)",ibbl,ifun));
        AFUN_BBL_LAST(ifun) = ABBL_PREV(ibbl);
    }
    else
    {
        CASSERT( AFUN_BBL_LAST(ifun) != ibbl,("bbl with valid next is last: %d (f: %d)\n",ibbl,ifun));
        if( AFUN_BBL_LAST(ifun) == ibbl )
        {
            PrintBblSuper(stdout,ibbl);
            
                         
        }
        
            
        ABBL_PREV(ABBL_NEXT(ibbl)) = ABBL_PREV(ibbl);
    }
}

/*=======================================================
@FUN
Insert ibbl after pred
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblInsertAfter"

extern void AbblInsertAfter(INDEX ibbl,INDEX pred)
{
    INDEX ifun;

    CASSERT( ABBL_FUN(ibbl) == ABBL_FUN(pred),("weird\n"));
    CASSERT( VALID(ibbl) && !ABBL_IS_KILLED(ibbl),("Bad argument bbl: %d\n",ibbl));
    ifun = ABBL_FUN(ibbl);
    CASSERT( VALID(ifun) && !AFUN_IS_KILLED(ifun),("invalid bbl fun: %d\n",ifun));

    ABBL_NEXT(ibbl) = ABBL_NEXT(pred);
    if( !VALID(ABBL_NEXT(ibbl)) )
    {
        CASSERT( AFUN_BBL_LAST(ifun) == pred,("weird\n"));
        AFUN_BBL_LAST(ifun) = ibbl;
    }
    else
    {
        ABBL_PREV(ABBL_NEXT(ibbl)) = ibbl;
    }

    ABBL_PREV(ibbl) = pred;
    ABBL_NEXT(pred) = ibbl;
}
/*=======================================================
@FUN
Kill basic block
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblKill"

extern void AbblKill(INDEX ibbl)
{
    INDEX iedg;
    INDEX iins;

    CASSERT( VALID(ibbl) && !ABBL_IS_KILLED(ibbl),("Bad argument bbl: %d\n",ibbl));
    CERROR( ABBL_FLAGS(ibbl) & F_REFQUAD,("cannot delete refquad bbl %d\n",ibbl));
    CWARNING( ABBL_FLAGS(ibbl) & F_GPREL32,("deleted gprel32 bbl %d [%s]\n",
                                            ibbl,AFUN_NAME(ABBL_FUN(ibbl))));

    /* Unlink from predecessor blocks, most be done first ??? */
    
    FOREACH_BBL_PRED(ibbl,iedg)
        {
        if( AEDG_TYPE(iedg) == ET_FUN_RETURN || AEDG_TYPE(iedg) == ET_COMPENSATE )
            AedgKillFancy(iedg-1);
        else
            AedgKillFancy(iedg);
        }
    
    /* Unlink from successor blocks, interaction with UnlinkEdge is tricky */
    
    FOREACH_BBL_SUCC(ibbl,iedg)     
        AedgKillFancy(iedg);

    /* Kill Instructions in basic block */

    FOREACH_BBL_INS(ibbl,iins)
        AinsKill(iins);

    AbblUnlink(ibbl);
    
    ABBL_FLAGS(ibbl) |= F_KILLED;
}

/*=======================================================
@FUN
This is a very important function that needs thorough
documentation. It is necessary to call this at times to
ensure that FindTargetBbl works correctly.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblMigrateRelocInfo"

extern void AbblMigrateRelocInfo(INDEX from, INDEX to)
{
    INDEX iins;
    INDEX itab;
    
    CASSERT( VALID(from) && !ABBL_IS_KILLED(from),("Bad argument bbl: %d\n",from));
    CASSERT( VALID(to) && !ABBL_IS_KILLED(to),("Bad argument bbl: %d\n",to));    

    if( !ABBL_IS_RELOC(from) ) return ;
#if 0
    printf("migrating info from: %d to %d\n",from,to);
#endif
    ABBL_FLAGS(to) |= ABBL_FLAGS(from);

    if( ABBL_FLAGS(from) & F_GPREL32 )
    {
        for( itab=0; itab < CONFIG_UTAB; itab++ )
        {
            if( ATAB_TARGET(itab) == from )
            {
                WARNING(("migrated switch target bbl from %d to %d\n",from,to));
                ATAB_TARGET(itab) = to;
            }
            
        }
    }
    
    /*
      even the killed instructions must keep the right bbl
      which is the one that later has the new address
      */

    FOREACH_INS(iins)
    {
        if( !(AINS_FLAGS(iins) & F_RELOC) ) continue;
        if( AINS_BBL(iins) != from ) continue;
        if( !AINS_IS_KILLED(iins) )
        {
            INDEX dup = AinsDup(iins);
            AINS_BBL(dup) = AINS_BBL(iins);
            AinsInsertAfter(dup,iins);
            AinsKill(iins);
        }
        
        AINS_BBL(iins) = to;
    }

    ABBL_FLAGS(from) &= ~(F_RELOC);
}
/*=======================================================
@FUN
Determine the called fun of callsite ibbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblGetCalledFun"

extern INDEX AbblGetCalledFun(INDEX ibbl)
{
    INDEX iedg;
    INDEX jbbl;
    CASSERT( VALID(ibbl) && !ABBL_IS_KILLED(ibbl),("invalid bbl: %d\n",ibbl));
    CASSERT( ABBL_TYPE(ibbl) == BT_CALL,("weird\n"));
    iedg = ABBL_EDG_CALL(ibbl);
    CASSERT( VALID(iedg),("invalid edg_call bbl: %d\n",ibbl));
    CASSERT( AEDG_TYPE(iedg) == ET_FUN_CALL,("invalid edge type in edg: %d\n",iedg));
    jbbl = AEDG_TAIL(iedg);
    CASSERT( VALID(jbbl) && !ABBL_IS_KILLED(jbbl),("invalid target bbl: %d\n",jbbl));
    if( !ABBL_IS_FIRST(jbbl) )
    {
        WARNING(("call to non first bbl\n"));
        PrintEdgSuper(stdout,iedg);
    }
    
    return ABBL_FUN(jbbl);
}

/*=======================================================
@FUN
Find bbl associated with address. Ie, find bbl which
represents the bbl that was once found at address in the
original binary.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "FindTargetBbl"

extern INDEX FindTargetBbl(ADDRESS address)
{
    INDEX iins;
    INDEX ibbl;
    
    iins = Addr2Ins(address);
    CASSERT( AINS_IS_RELOC(iins),("target ins %d is not reloc (adr: %016lx)\n",iins,address));
    ibbl =  AINS_BBL(iins);
    CASSERT( ABBL_IS_RELOC(ibbl),("branch target bbl %d is not reloc\n",ibbl));
    CASSERT( !ABBL_IS_KILLED(ibbl),("branch target bbl %d is dead\n",ibbl));
    return ibbl;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AedgNew"

static INDEX AedgNew()
{
    CFATAL(CONFIG_UEDG >= CONFIG_NEDG,("not enough edges\n"));
    return(CONFIG_UEDG++);
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AedgInit"

static void AedgInit(INDEX iedg)
{
    CASSERT( VALID(iedg),("Bad argument edg: %d\n",iedg));
    
    AEDG_FLAGS(iedg) = 0;
    AEDG_TYPE(iedg) = INVALID;
    AEDG_HEAD(iedg) = INVALID;           
    AEDG_TAIL(iedg) = INVALID;           
    AEDG_SUCC(iedg) = INVALID;           
    AEDG_PRED(iedg) = INVALID;           
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AedgDupFancyObsolete"

extern INDEX AedgDupFancyObsolete(INDEX iedg)
{
    INDEX jedg;
    
    CASSERT( VALID(iedg) && !AEDG_IS_KILLED(iedg),("Bad argument edg: %d\n",iedg));
    CERROR( AEDG_TYPE(iedg) == ET_FUN_RETURN,("weird\n"));

    jedg = AedgNew();
    AedgInit(jedg);
    AEDG_FLAGS(jedg) = AEDG_FLAGS(iedg);
    AEDG_TYPE(jedg) = AEDG_TYPE(iedg);
    AEDG_HEAD(jedg) = AEDG_HEAD(iedg);
    AEDG_TAIL(jedg) = AEDG_TAIL(iedg);

    if( AEDG_TYPE(iedg) == ET_FUN_CALL )   /* allocate return edge */
    {
        AedgNew();
    }
    
    return jedg;
}
/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "AedgUnlinkFancy"

extern void AedgUnlinkFancy(INDEX iedg)
{
    INDEX jedg;
    INDEX hd;
    INDEX tl;
    BOOL  found;

    CASSERT( VALID(iedg) && !AEDG_IS_KILLED(iedg),("Bad argument edg: %d\n",iedg));
    hd = AEDG_HEAD(iedg);
    tl = AEDG_TAIL(iedg);
    if( AEDG_IS_KILLED(iedg) )
    {
        PrintEdgSuper(stdout,iedg);
        PrintBblSuper(stdout,hd);
        PrintBblSuper(stdout,tl);
    }

    CASSERT( VALID(hd) && !ABBL_IS_KILLED(hd),("Bad edge head bbl: %d\n",hd));
    CASSERT( VALID(tl) && !ABBL_IS_KILLED(tl),("Bad edge tail bbl: %d\n",tl));

    /* remove edge form head */

    found = 0;
    if( ABBL_SUCC(hd) == iedg )
    {
        ABBL_SUCC(hd) = AEDG_SUCC(iedg);
        found = 1;
    }
    else
        FOREACH_BBL_SUCC(hd,jedg)
            if( AEDG_SUCC(jedg) == iedg )
            {
                AEDG_SUCC(jedg) = AEDG_SUCC(iedg);
                found = 1;
                break;
            }
    
    CASSERT( found ,("Could not remove edg %d from head bbl %d\n",iedg,hd));

    /* remove edge form head */

    found = 0;
    if( ABBL_PRED(tl) == iedg )
    {
        ABBL_PRED(tl) = AEDG_PRED(iedg);
        found = 1;
    }
    else
    {
        FOREACH_BBL_PRED(tl,jedg)
            if( AEDG_PRED(jedg) == iedg )
            {
                AEDG_PRED(jedg) = AEDG_PRED(iedg);
                found = 1;
                break;
            }
    }
    
    CASSERT( found ,("Could not remove edg %d from tail bbl %d\n",iedg,tl));

    /* remove edge form head */
    
    if( ABBL_EDG_FALSE(hd) == iedg )
        ABBL_EDG_FALSE(hd) = INVALID;

    if( ABBL_EDG_TRUE(hd) == iedg )
        ABBL_EDG_TRUE(hd) = INVALID;

    if( AEDG_TYPE(iedg+1) == ET_FUN_RETURN || AEDG_TYPE(iedg+1) == ET_COMPENSATE )
        AedgUnlinkFancy(iedg+1);
}

/*=======================================================
@FUN
Attach an  edge to its nodes. 
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AedgLinkFancy"

extern void AedgLinkFancy(INDEX iedg)
{
    INDEX hd;
    INDEX tl;
    
    CASSERT( VALID(iedg) && !AEDG_IS_KILLED(iedg),("Bad argument edg: %d\n",iedg));
    hd = AEDG_HEAD(iedg);
    tl = AEDG_TAIL(iedg);
    CASSERT( VALID(hd) && !ABBL_IS_KILLED(hd),("Bad %c edge head bbl: %d\n",AEDG_TYPE(iedg),hd));
    CASSERT( VALID(tl) && !ABBL_IS_KILLED(tl),("Bad %c edge tail bbl: %d\n",AEDG_TYPE(iedg),tl));

    /*
     *
     * do "fancy" treatment for all edge types
     *
     */

    switch( AEDG_TYPE(iedg) )
    {
    case ET_EXIT_LINK:
        CASSERT( ABBL_TYPE(hd) == BT_RET,("et_exit_link edg from non bt_ret bbl\n"));
        AEDG_TAIL(iedg) = tl = AFUN_BBL_LAST( ABBL_FUN(hd) );
        if( ABBL_TYPE(hd) != BT_RET )
        {
            PrintEdgSuper(stdout,iedg);
        }
        

        CASSERT( ABBL_TYPE(tl) == BT_EXIT,("et_exit_link edge to non bt_exit bbl\n"));
        break;

    case ET_FUN_RETURN:
        CASSERT( ABBL_TYPE(hd) == BT_EXIT,("et_fun_return edg from non bt_exit bbl\n"));
        break;

    case ET_NORM:
        CASSERT( ABBL_TYPE(hd) == BT_NORM,("et_norm edg from non bt_norm bbl \n"));
        CERROR( VALID(ABBL_EDG_NORM(hd)),("overwriting et_norm in bbl %d\n",hd));
        ABBL_EDG_NORM(hd) = iedg;
        break;
        
    case ET_UNCOND:
        CASSERT( ABBL_TYPE(hd) == BT_UNCOND,("et_uncond edg from non bt_uncond bbl \n"));
        CERROR( VALID(ABBL_EDG_UNCOND(hd)),("overwriting edg_uncond\n"));
        ABBL_EDG_UNCOND(hd) = iedg;
        break;

    case ET_JUMP:
        CASSERT( ABBL_TYPE(hd) == BT_JUMP,("et_jump edg from non bt_jump bbl\n"));
        CASSERT( ABBL_FUN(tl) == AFUN_PSEUDO_HELL,("jump to non hell\n"));
        CERROR( VALID(ABBL_EDG_JUMP(hd)),("overwriting edg_jump\n"));
        ABBL_EDG_JUMP(hd) = iedg;
        break;

    case ET_EXTRA:
        CASSERT( ABBL_TYPE(hd) == BT_EXTRA,("et_extra edg from non bt_extra bbl \n"));
        CERROR( VALID(ABBL_EDG_EXTRA(hd)),("overwriting edg_uncond\n"));
        ABBL_EDG_EXTRA(hd) = iedg;
        break;
        
    case ET_FUN_CALL:
        CASSERT( ABBL_TYPE(hd)==BT_CALL || ABBL_TYPE(hd)==BT_PAL,
                 ("et_fun_call edg from non bt_call bbl \n"));
        CERROR( VALID(ABBL_EDG_CALL(hd)),("overwriting edg_call %d\n",iedg));
#if 1
        if( !ABBL_IS_FIRST(tl) )
        {
            ERROR(("call to non first bbl\n"));
            PrintEdg(stdout,iedg);
            PrintBbl(stdout,hd);
            PrintBbl(stdout,tl);
        }
#endif
        ABBL_EDG_CALL(hd) = iedg;    
        break;
        
    case ET_FUN_LINK:
        CASSERT( ABBL_TYPE(hd)==BT_CALL || ABBL_TYPE(hd)==BT_PAL,
                 ("et_fun_link edg from non bt_call bbl \n"));
        CERROR( VALID(ABBL_EDG_LINK(hd)),("overwriting edg_link\n"));
        ABBL_EDG_LINK(hd) = iedg;
        break;

    case  ET_SWITCH:
        CASSERT( ABBL_TYPE(hd) == ET_SWITCH,("et_switch edg from non bt_switch bbl \n"));
        CASSERT( ABBL_IS_RELOC(tl),("et_switch to non reloc bbl\n"));
        break;
        
    case  ET_HELL:
        CASSERT( (ABBL_FUN(hd) == AFUN_PSEUDO_HELL) || 
		 (ABBL_FUN(hd) == AFUN_PSEUDO_HELL_LITE) ,("weird\n"));
        break;

    case  ET_SETLONGJMP:
        CASSERT( ABBL_TYPE(hd) == BT_EXIT,("bad head of br-setlongjmp-edg\n"));
        CASSERT( ABBL_TYPE(tl) == BT_EXIT,("bad tail of br-setlongjmp-edg\n"));
        break;
        

    case ET_COMPENSATE:
        CWARNING( ABBL_TYPE(hd) != BT_EXIT,("et_compensate edg from non bt_exit bbl - this is normal during layouting\n"));
        CWARNING( ABBL_TYPE(tl) != BT_EXIT,("et_compensate edg to non bt_exit bbl - this is normal during layouting\n"));
        break;
        
    case ET_TRUE:
        CASSERT( ABBL_TYPE(hd) == BT_BRANCH,("et_true edg from non bt_branch bbl %d\n",hd));
        CERROR( VALID(ABBL_EDG_TRUE(hd)),("overwriting edg_true %d\n",iedg));
        ABBL_EDG_TRUE(hd) = iedg;    
        break;
        
    case ET_FALSE:
        CASSERT( ABBL_TYPE(hd) == BT_BRANCH,("et_false edg from non bt_branch bbl %d\n",hd));
        CERROR( VALID(ABBL_EDG_FALSE(hd)),("overwriting edg_false %d\n",iedg));
        ABBL_EDG_FALSE(hd) = iedg;    
        break;
        
    default:    
        ERROR(("unexpected edge type\n"));
        PrintEdgSuper(stdout,iedg);
        break;
    }

    if( AEDG_TYPE(iedg+1) == ET_FUN_RETURN )
    {
        INDEX returnsite;
        INDEX exit;
        CASSERT( ABBL_TYPE(hd)==BT_CALL || ABBL_TYPE(hd)==BT_PAL,
                 ("function from non bt_call bbl\n"));
        CASSERT( VALID(ABBL_EDG_LINK(hd)),("bad function link\n"));
        exit = AFUN_BBL_LAST(ABBL_FUN(tl));
        returnsite = AEDG_TAIL(ABBL_EDG_LINK(hd));
        AEDG_HEAD(iedg+1) = exit;
        AEDG_TAIL(iedg+1) = returnsite;
        AedgLinkFancy(iedg+1);
    }

    if( AEDG_TYPE(iedg+1) == ET_COMPENSATE )
    {
        AEDG_HEAD(iedg+1) = AFUN_BBL_LAST(ABBL_FUN(tl));
        AEDG_TAIL(iedg+1) = AFUN_BBL_LAST(ABBL_FUN(hd));
        if( !VALID(AEDG_HEAD(iedg+1)) || !VALID(AEDG_TAIL(iedg+1)) )
        {
            WARNING(("Cannot find exit bbls for compensation edge - this is normal during layouting\n"));
        }
        else
        {
            CWARNING( AEDG_TAIL(iedg+1) == AEDG_HEAD(iedg+1),("useless compensation for edg %d\n",iedg));
            AedgLinkFancy(iedg+1);
        }
    }
    
    
    /* do the actual linking */
    
    AEDG_SUCC(iedg) = ABBL_SUCC( hd );
    ABBL_SUCC( hd ) = iedg;
    
    AEDG_PRED(iedg) = ABBL_PRED( tl );
    ABBL_PRED( tl ) = iedg;
}


/*=======================================================
@FUN
Create a new edge of a type et going from hd to tl.
This edge will not be attached to hd and tl unless you call
AedgLinkFancy.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AedgNewFancy"

extern INDEX AedgNewFancy(INDEX hd, INDEX tl, EDGETYPE et)
{
    INDEX iedg = AedgNew();
    
    CASSERT( VALID(hd) && !ABBL_IS_KILLED(hd),("Bad %c edge head bbl: %d\n",et,hd));
    CASSERT( VALID(tl) && !ABBL_IS_KILLED(tl),("Bad %c edge tail bbl: %d\n",et,tl));

    
    AedgInit(iedg);
    AEDG_TYPE(iedg) = et;
    AEDG_HEAD(iedg) = hd;
    AEDG_TAIL(iedg) = tl;

    CERROR( AEDG_TYPE(iedg) == ET_FUN_RETURN,("attempt to create a et_fun_return edge\n"));
    CERROR( AEDG_TYPE(iedg) == ET_COMPENSATE,("attempt to create a et_compensate edge\n"));

    if( et == ET_EXIT_LINK )
    {
        /* this is not really necessary since AedgLinkFancy also
           recomputes this.
           However, we avoid some warning messages in OptBranchMerge
           this way. */
        
        AEDG_TAIL(iedg) = tl = AFUN_BBL_LAST( ABBL_FUN(hd) );

    }
    
    if( et == ET_FUN_CALL || ABBL_FUN(hd) != ABBL_FUN(tl) )
    {
        INDEX jedg = AedgNew();
        AedgInit(jedg);
        AEDG_TYPE(jedg) = (et == ET_FUN_CALL) ? ET_FUN_RETURN : ET_COMPENSATE;
    }

    return iedg;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AedgKillFancy"

extern void AedgKillFancy(INDEX iedg)
{
    CASSERT( VALID(iedg) && !AEDG_IS_KILLED(iedg),("Bad argument edg: %d\n",iedg));
    CERROR( AEDG_TYPE(iedg) == ET_FUN_RETURN,("attempt to kill a fun-return-edg\n"));
    CERROR( AEDG_TYPE(iedg) == ET_COMPENSATE,("attempt to kill a compensate-edg\n")); 

    AedgUnlinkFancy(iedg);
    AEDG_FLAGS(iedg) |= F_KILLED;

    if( AEDG_TYPE(iedg+1) == ET_FUN_RETURN || AEDG_TYPE(iedg+1) == ET_COMPENSATE )
    {
        AEDG_FLAGS(iedg+1) |= F_KILLED;
    }

}

/*=======================================================
@FUN
Return the number of instructions in ibbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblNInss"

extern NUMBER AbblNInss(INDEX ibbl)
{
    INDEX  iins;
    NUMBER i=0;
    CASSERT( VALID(ibbl),("Bad argument bbl: %d\n",ibbl));
    FOREACH_BBL_INS(ibbl,iins) i++;
    return i;
}
/*=======================================================
@FUN
Return the number of marked instructions in ibbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblNMarkedInss"

extern NUMBER AbblNMarkedInss(INDEX ibbl)
{
    INDEX  iins;
    NUMBER i=0;
    CASSERT( VALID(ibbl),("Bad argument bbl: %d\n",ibbl));
    FOREACH_BBL_INS(ibbl,iins)
    {
        if( AINS_FLAGS(iins) & F_MARKED)
            i++;
    }

    return i;
}
/*=======================================================
@FUN
This "fancy" function appends a bbl to the end of a function.
Since the last bbl of a function is always the dummy bbl
the new bbl is actually placed second to last.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AbblAppendFancy"

extern void AbblAppendFancy(INDEX ibbl)
{
    INDEX ifun;
    INDEX exit;
    
    CASSERT( VALID(ibbl) && !ABBL_IS_KILLED(ibbl),("Bad argument bbl: %d\n",ibbl));
    ifun = ABBL_FUN(ibbl);
    CASSERT( VALID(ifun) && !AFUN_IS_KILLED(ifun),("invalid bbl fun: %d\n",ifun));
    exit = AFUN_BBL_LAST(ifun);
    CASSERT( VALID(exit) && !ABBL_IS_KILLED(exit),("invalid exit bbl: %d\n",exit));
    CASSERT( !(ABBL_TYPE(ibbl) == BT_EXIT),("extra exit bbl: %d\n",ibbl));
    
    ABBL_NEXT(ibbl) = exit;
    ABBL_PREV(ibbl) = ABBL_PREV(exit);
    ABBL_NEXT(ABBL_PREV(exit)) = ibbl;
    ABBL_PREV(exit) = ibbl;
    if( AFUN_BBL_LAST(ifun) == AFUN_BBL_FIRST(ifun) )
        AFUN_BBL_FIRST(ifun) = ibbl;

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbinNHotCallsites"

extern NUMBER AbinNHotCallsites()
{
    INDEX  ibbl;
    NUMBER count = 0;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_CALL ) continue;
        if( AINS_TYPE(ABBL_INS_LAST(ibbl)) == IT_JMP ) continue;
        CASSERT( AINS_TYPE(ABBL_INS_LAST(ibbl)) == IT_BRA,("bad call ins\n"));
        if( !ABBL_IS_HOT(ibbl) ) continue;
        count++;
    }
    
    return count;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbinNCallsites"

extern NUMBER AbinNCallsites()
{
    INDEX  ibbl;
    NUMBER count = 0;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_CALL ) continue;
        if( AINS_TYPE(ABBL_INS_LAST(ibbl)) == IT_JMP ) continue;
        CASSERT( AINS_TYPE(ABBL_INS_LAST(ibbl)) == IT_BRA,("bad call ins\n"));
        count++;
    }
    
    return count;
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "AbinNHotIndirectCallsites"

extern NUMBER AbinNHotIndirectCallsites()
{
    INDEX  ibbl;
    NUMBER count = 0;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_CALL ) continue;
        if( AINS_TYPE(ABBL_INS_LAST(ibbl)) == IT_BRA ) continue;
        CASSERT( AINS_TYPE(ABBL_INS_LAST(ibbl)) == IT_JMP,("bad call ins\n"));
        if( !ABBL_IS_HOT(ibbl) ) continue;
        count++;
    }
    
    return count;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbinNIndirectCallsites"

extern NUMBER AbinNIndirectCallsites()
{
    INDEX  ibbl;
    NUMBER count = 0;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_CALL ) continue;
        if( AINS_TYPE(ABBL_INS_LAST(ibbl)) == IT_BRA ) continue;
        CASSERT( AINS_TYPE(ABBL_INS_LAST(ibbl)) == IT_JMP,("bad call ins\n"));
        count++;
    }
    
    return count;
}
/*=======================================================
=======================================================*/


