/*=======================================================
  Author:   R. Muth
  Version:  $Id: instr.c,v 1.27 1999/04/18 20:01:02 muth Exp $
  Comment:  This file contains instruction related functions
=======================================================*/

#include "global.h"
#include "functions.h"
#include "regs.h"

#include "opcode.h"
#include "flags.h"
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsComesBefore"

extern BOOL AinsComesBefore(INDEX iins,INDEX jins)
{
    if( iins == jins ) return 0;
    for(;;)
    {
        iins = AINS_NEXT(iins);
        if( iins == jins ) return 1;
        if( iins == INVALID ) return 0;
    }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsHasSideEffect"

extern BOOL AinsHasSideEffect(INDEX iins)
{
    switch( AINS_TYPE(iins) )
    {
    case IT_BRA:
    case IT_JMP:
    case IT_IBR:
    case IT_FBR:
        
    case IT_IST:
    case IT_UST:
    case IT_CST:
    case IT_FST:
    case IT_LLD:

    case IT_PAL:
    case IT_MEM:
    case IT_MSC:
    case IT_RX:
    case IT_F2R:
    case IT_R2F:
        
        return 1;
    default:
        return 0;
    }
}

/*=======================================================
@FUN
An instruction can either cause a numerical exception such
as division by zero or overflow
Or, it can cause a memory exception.
We assume that stack accesses and fixed address accesses will
not cause an exception.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMayCauseException"

extern BOOL AinsMayCauseException(INDEX iins)
{
    if( AINS_RES_USE(iins) & (RES_FP|RES_TR) )
        return 1;
    if( AINS_RES_USE(iins) & (RES_LD|RES_ST) )
    {
        if( AINS_REG_B(iins) != REG_SP &&
            !(AINS_FLAGS(iins) & FI_EVAL_B) )
            return 1;
    }
    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsIsNoop"

extern BOOL AinsIsNoop(INDEX iins)
{
    if( AINS_REG_C(iins) != REG_ZERO && AINS_REG_C(iins) != REG_FZERO ) return 0;
    if( AinsHasSideEffect(iins) ) return 0;
    return 1;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsIsMove"

extern BOOL AinsIsMove(INDEX iins)
{
    if( AINS_TYPE(iins)==IT_LDA && AINS_DATA(iins)==0 )
        return 1;


    if( AINS_REG_B(iins) == REG_ZERO )
    {
            if( AINS_CODE(iins) == I_bis ||
                AINS_CODE(iins) == I_xor ||
                AINS_CODE(iins) == I_bic ||
                AINS_CODE(iins) == I_addq ||
                AINS_CODE(iins) == I_subq ||
                AINS_CODE(iins) == I_sll ||
                AINS_CODE(iins) == I_srl ||
                AINS_CODE(iins) == I_zap ||
                AINS_CODE(iins) == I_sra )
                return 1;
    }
    
    
    if( AINS_REG_A(iins) == REG_ZERO && AINS_REG_B(iins) != REG_LIT )
    {
        if( AINS_CODE(iins) == I_bis ||
            AINS_CODE(iins) == I_xor ||
            AINS_CODE(iins) == I_addq ||
            AINS_CODE(iins) == I_s4addl ||
            AINS_CODE(iins) == I_s4addq ||
            AINS_CODE(iins) == I_s8addl ||
            AINS_CODE(iins) == I_s8addq)
            return 1;
    }
        
   if( AINS_REG_A(iins) == AINS_REG_B(iins) )
   {
       if( AINS_CODE(iins) == I_bis ||
           AINS_CODE(iins) == I_and ||
           AINS_CODE(iins) == I_eqv || 
           AINS_CODE(iins) == I_cpys ||
           AINS_CODE(iins) == I_cpyse)  
           return 1;
   }

    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsDataType"

extern DATATYPE AinsDataType(INDEX iins)
{
    if( AINS_TYPE(iins) != IT_IST &&
        AINS_TYPE(iins) != IT_FST &&
        AINS_TYPE(iins) != IT_CST &&
        AINS_TYPE(iins) != IT_UST &&
        AINS_TYPE(iins) != IT_ULD &&
        AINS_TYPE(iins) != IT_LLD &&
        AINS_TYPE(iins) != IT_ILD &&
        AINS_TYPE(iins) != IT_FLD )
    {
        ERROR(("weird\n"));
        return INVALID;
    }
    
    switch( AINS_CODE(iins) )
    {
    case I_stb:
    case I_ldbu:
        return DT_BYTE;
    case I_stw:
    case I_ldwu:
        return DT_WORD;
    case I_stl:
    case I_stl_c:
    case I_ldl:
    case I_ldl_l:
        return DT_LONG;
    case I_stq:
    case I_stq_c:
    case I_stq_u:
    case I_ldq:
    case I_ldq_l:
    case I_ldq_u:
        return DT_QUAD;
    case I_stf:
    case I_ldf:
        return DT_FFLT;
    case I_stg:
    case I_ldg:
        return DT_GFLT;
    case I_sts:
    case I_lds:
        return DT_SFLT;
    case I_stt:
    case I_ldt:
        return DT_TFLT;
    }

    ERROR(("weird\n"));
    return INVALID;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMemWidth"

extern NUMBER AinsMemWidth(INDEX iins)
{
    if( AINS_TYPE(iins) != IT_IST &&
        AINS_TYPE(iins) != IT_FST &&
        AINS_TYPE(iins) != IT_CST &&
        AINS_TYPE(iins) != IT_UST &&
        AINS_TYPE(iins) != IT_ULD &&
        AINS_TYPE(iins) != IT_LLD &&
        AINS_TYPE(iins) != IT_ILD &&
        AINS_TYPE(iins) != IT_FLD )
    {
        ERROR(("bad instruction type %s\n",AINS_NAME(iins)));
        return 0;
    }

    switch( AINS_CODE(iins) )
    {
    case I_stb:
    case I_ldbu:
        return 1;
    case I_stw:
    case I_ldwu:
        return 2;
    case I_stl:
    case I_stl_c:
    case I_ldl:
    case I_ldl_l:
    case I_sts:
    case I_lds:
    case I_stf:
    case I_ldf:
        return 4;
    case I_stq:
    case I_stq_c:
    case I_stq_u:
    case I_ldq:
    case I_ldq_l:
    case I_ldq_u: 
    case I_stt:
    case I_ldt:
    case I_stg:
    case I_ldg:
        return 8;
    }

    ERROR(("weird\n"));
    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsInit"

static void AinsInit(INDEX iins)
{
    CASSERT( VALID(iins),("Bad argument ins: %d\n",iins));

    AINS_FLAGS(iins) = 0;
    AINS_PREV(iins) = INVALID;
    
    AINS_NEXT(iins) = INVALID;    
    AINS_CODE(iins) = INVALID;    
    AINS_DATA(iins) = INVALID;    
    AINS_REG_A(iins) = INVALID;   
    AINS_REG_B(iins) = INVALID;   
    AINS_REG_C(iins) = INVALID;   
    AINS_BBL(iins) = INVALID;     
    AINS_B_DEF(iins) = INVALID;
    AINS_A_DEF(iins) = INVALID;    
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsNew"

extern INDEX AinsNew()
{
    CFATAL(CONFIG_UINS >= CONFIG_NINS,("not enough instructions\n"));
    AinsInit(CONFIG_UINS);
    return(CONFIG_UINS++);
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsDup"

extern INDEX AinsDup(INDEX iins)
{
    INDEX jins = AinsNew();
    
    CASSERT( VALID(iins) && !AINS_IS_KILLED(iins),("Bad argument ins: %d\n",iins));

    AinsInit(jins);
    AINS_FLAGS(jins)  = AINS_FLAGS(iins) &
                        (FI_EVAL_A|FI_EVAL_B|FI_EVAL_C|FI_CHANGED);
    AINS_CODE(jins)   = AINS_CODE(iins);
    AINS_DATA(jins)   = AINS_DATA(iins);    
    AINS_REG_A(jins)  = AINS_REG_A(iins);   
    AINS_REG_B(jins)  = AINS_REG_B(iins);   
    AINS_REG_C(jins)  = AINS_REG_C(iins);   
    AINS_C_EVAL(jins) = AINS_C_EVAL(iins);
    AINS_B_EVAL(jins) = AINS_B_EVAL(iins);
    AINS_A_EVAL(jins) = AINS_A_EVAL(iins);

    /* do not copy the bbl value */
       
    return jins;
}

/*=====================================================================
  =====================================================================*/
#undef __PROC__
#define __PROC__ "AinsMove"

extern void AinsMove(INDEX instr, INDEX destination_bbl)
{
   INDEX new_instr;

   new_instr = AinsDup(instr);
   AINS_BBL(new_instr) = destination_bbl;
   AinsAppend(new_instr);
   AinsKill(instr);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeLdt"

extern void  AinsMakeLdt(INDEX iins,REG source, REG dest, NUMBER offset)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);

    CASSERT( offset >= -(1<<15) && offset < (1<<15),("bad offset %d for lda\n",offset));
    CASSERT( source != REG_LIT,("bad source register\n"));
    CASSERT( dest != REG_LIT,("bad dest register\n"));
    
    AINS_CODE(iins) = I_ldt;
    AINS_REG_A(iins) = REG_ZERO;
    AINS_REG_B(iins) = source;
    AINS_REG_C(iins) = dest;
    AINS_DATA(iins) = offset;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeLda"

extern void  AinsMakeLda(INDEX iins,REG source, REG dest, NUMBER offset)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);

    CASSERT( offset >= -(1<<15) && offset < (1<<15),("bad offset %d for lda\n",offset));
    CASSERT( source != REG_LIT,("bad source register\n"));
    CASSERT( dest != REG_LIT,("bad dest register\n"));
    
    AINS_CODE(iins) = I_lda;
    AINS_REG_A(iins) = REG_ZERO;
    AINS_REG_B(iins) = source;
    AINS_REG_C(iins) = dest;
    AINS_DATA(iins) = offset;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeLdq"

extern void  AinsMakeLdq(INDEX iins,REG source, REG dest, NUMBER offset)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);

    CASSERT( offset >= -(1<<15) && offset < (1<<15),("bad offset %d for lda\n",offset));
    CASSERT( source != REG_LIT,("bad source register\n"));
    CASSERT( dest != REG_LIT,("bad dest register\n"));
    
    AINS_CODE(iins) = I_ldq;
    AINS_REG_A(iins) = REG_ZERO;
    AINS_REG_B(iins) = source;
    AINS_REG_C(iins) = dest;
    AINS_DATA(iins) = offset;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeStq"

extern void  AinsMakeStq(INDEX iins,REG stored, REG dest, NUMBER offset)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);

    CASSERT( offset >= -(1<<15) && offset < (1<<15),("bad offset %d for lda\n",offset));
    CASSERT( stored != REG_LIT,("bad source register\n"));
    CASSERT( dest != REG_LIT,("bad dest register\n"));
    
    AINS_CODE(iins) = I_stq;
    AINS_REG_A(iins) = stored;
    AINS_REG_B(iins) = dest;
    AINS_REG_C(iins) = REG_ZERO;
    AINS_DATA(iins) = offset;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeStt"

extern void  AinsMakeStt(INDEX iins,REG stored, REG dest, NUMBER offset)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);

    CASSERT( offset >= -(1<<15) && offset < (1<<15),("bad offset %d for lda\n",offset));
    CASSERT( stored != REG_LIT,("bad source register\n"));
    CASSERT( dest != REG_LIT,("bad dest register\n"));
    
    AINS_CODE(iins) = I_stt;
    AINS_REG_A(iins) = stored;
    AINS_REG_B(iins) = dest;
    AINS_REG_C(iins) = REG_ZERO;
    AINS_DATA(iins) = offset;
}




/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeCpys"

extern void  AinsMakeCpys(INDEX iins,REG rega, REG regb, NUMBER regc)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);

    CASSERT( REG_IS_FLT(rega),("bad source register\n"));
    CASSERT( REG_IS_FLT(regb),("bad source register\n"));
    CASSERT( REG_IS_FLT(regc),("bad source register\n"));
    
    AINS_CODE(iins) = I_cpys;
    AINS_REG_A(iins) = rega;
    AINS_REG_B(iins) = regb;
    AINS_REG_C(iins) = regc;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeLdah"

extern void  AinsMakeLdah(INDEX iins,REG source, REG dest, NUMBER offset)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);

    CASSERT( offset >= -(1<<15) && offset < (1<<15),("bad offset %d for lda\n",offset));
    CASSERT( source != REG_LIT,("bad source register\n"));
    CASSERT( dest != REG_LIT,("bad dest register\n"));
    
    AINS_CODE(iins) = I_ldah;
    AINS_REG_A(iins) = REG_ZERO;
    AINS_REG_B(iins) = source;
    AINS_REG_C(iins) = dest;
    AINS_DATA(iins) = offset;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeBeq"     

extern void AinsMakeBeq(INDEX iins, REG reg, NUMBER offset)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_beq;
    AINS_REG_A(iins) = reg;
    AINS_REG_B(iins) = REG_ZERO;
    AINS_REG_C(iins) = REG_ZERO;
    AINS_DATA(iins) = offset;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeBge"     

extern void AinsMakeBge(INDEX iins, REG reg, NUMBER offset)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_bge;
    AINS_REG_A(iins) = reg;
    AINS_REG_B(iins) = REG_ZERO;
    AINS_REG_C(iins) = REG_ZERO;
    AINS_DATA(iins) = offset;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeBle"     

extern void AinsMakeBle(INDEX iins, REG reg, NUMBER offset)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_ble;
    AINS_REG_A(iins) = reg;
    AINS_REG_B(iins) = REG_ZERO;
    AINS_REG_C(iins) = REG_ZERO;
    AINS_DATA(iins) = offset;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeBr"     

extern void AinsMakeBr(INDEX iins, NUMBER offset)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_br;
    AINS_REG_A(iins) = REG_ZERO;
    AINS_REG_B(iins) = REG_ZERO;
    AINS_REG_C(iins) = REG_ZERO;
    AINS_DATA(iins) = offset;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeAddl"     

extern void AinsMakeAddl(INDEX iins,REG ra,REG rb, REG rc)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_addl;
    AINS_REG_A(iins) = ra;
    AINS_REG_B(iins) = rb;
    AINS_REG_C(iins) = rc;
    AINS_DATA(iins) = 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeAddl"     

extern void AinsMakeAddq(INDEX iins,REG ra,REG rb, REG rc)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_addq;
    AINS_REG_A(iins) = ra;
    AINS_REG_B(iins) = rb;
    AINS_REG_C(iins) = rc;
    AINS_DATA(iins) = 0;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeCmpeq"     

extern void AinsMakeCmpeq(INDEX iins,REG ra,REG rb, REG rc, INT64 data)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_cmpeq;
    AINS_REG_A(iins) = ra;
    AINS_REG_B(iins) = rb;
    AINS_REG_C(iins) = rc;
    AINS_DATA(iins) = data;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeCmplt"     

extern void AinsMakeCmplt(INDEX iins,REG ra,REG rb, REG rc, INT64 data)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_cmplt;
    AINS_REG_A(iins) = ra;
    AINS_REG_B(iins) = rb;
    AINS_REG_C(iins) = rc;
    AINS_DATA(iins) = data;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeCmple"     

extern void AinsMakeCmple(INDEX iins,REG ra,REG rb, REG rc, INT64 data)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_cmple;
    AINS_REG_A(iins) = ra;
    AINS_REG_B(iins) = rb;
    AINS_REG_C(iins) = rc;
    AINS_DATA(iins) = data;
}

#undef __PROC__
#define __PROC__ "AinsMakeCmpgeq"     

extern void AinsMakeCmpgeq(INDEX iins,REG ra,REG rb, REG rc, INT64 data)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_cmpgeq;
    AINS_REG_A(iins) = ra;
    AINS_REG_B(iins) = rb;
    AINS_REG_C(iins) = rc;
    AINS_DATA(iins) = data;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeNot"     

extern void AinsMakeNot(INDEX iins,REG rb,REG rc,INT64 data)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_ornot;
    AINS_REG_A(iins) = REG_ZERO;
    AINS_REG_B(iins) = rb;
    AINS_REG_C(iins) = rc;
    AINS_DATA(iins) = data;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeAnd"     

extern void AinsMakeAnd(INDEX iins, REG ra, REG rb,REG rc)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_and;
    AINS_REG_A(iins) = ra;
    AINS_REG_B(iins) = rb;
    AINS_REG_C(iins) = rc;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeSubq"     

extern void AinsMakeSubq(INDEX iins,REG ra,REG rb, REG rc)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_subq;
    AINS_REG_A(iins) = ra;
    AINS_REG_B(iins) = rb;
    AINS_REG_C(iins) = rc;
    AINS_DATA(iins) = 0;
}

#undef __PROC__
#define __PROC__ "AinsMakeBne"     

extern void AinsMakeBne(INDEX iins, REG s1, NUMBER offset)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_bne;
    AINS_REG_A(iins) = s1;
    AINS_REG_B(iins) = REG_ZERO;
    AINS_REG_C(iins) = REG_ZERO;
    AINS_DATA(iins) = offset;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeBsr"     

extern void AinsMakeBsr(INDEX iins,REG ret)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_bsr;
    AINS_REG_A(iins) = REG_ZERO;
    AINS_REG_B(iins) = REG_ZERO;
    AINS_REG_C(iins) = ret;
    AINS_DATA(iins) = 0;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeJsr"     

extern void AinsMakeJsr(INDEX iins, REG target, REG ret)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_jsr;
    AINS_REG_A(iins) = REG_ZERO;
    AINS_REG_B(iins) = target;
    AINS_REG_C(iins) = ret;
    AINS_DATA(iins) = 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeRet"     

extern void AinsMakeRet(INDEX iins,REG ret)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_ret;
    AINS_REG_A(iins) = REG_ZERO;
    AINS_REG_B(iins) = ret;
    AINS_REG_C(iins) = REG_ZERO;
    AINS_DATA(iins) = 0;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeUnop"

extern void AinsMakeUnop(INDEX iins)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);

    AINS_CODE(iins) = I_unop;
    AINS_REG_A(iins) = REG_ZERO;
    AINS_REG_C(iins) = REG_ZERO;
    AINS_REG_B(iins) = REG_ZERO;
    AINS_DATA(iins) = 0;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeInop"

extern void AinsMakeInop(INDEX iins)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_bis;
    AINS_REG_A(iins) = REG_ZERO;
    AINS_REG_B(iins) = REG_ZERO;
    AINS_REG_C(iins) = REG_ZERO;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsMakeFnop"

extern void AinsMakeFnop(INDEX iins)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);
    
    AINS_CODE(iins) = I_cpys;
    AINS_REG_A(iins) = REG_FZERO;
    AINS_REG_B(iins) = REG_FZERO;
    AINS_REG_C(iins) = REG_FZERO;
}
/*=======================================================
=======================================================*/
extern BOOL AinsMakeMove(INDEX iins, REG sReg, DATATYPE sType, REG dReg, DATATYPE dType)
{
    AINS_FLAGS(iins) |= FI_CHANGED;
    AINS_FLAGS(iins) &= ~(FI_EVAL_C|FI_EVAL_B|FI_EVAL_A|FI_DEF_B|FI_DEF_A);

    CASSERT( sReg != REG_LIT,("bad source register\n"));
    CASSERT( dReg != REG_LIT,("bad dest register\n"));
    
    if( (sType == DT_LONG || sType == DT_QUAD) && dType == DT_LONG )
    {
        AINS_CODE(iins) = I_addl;
        AINS_REG_A(iins) = sReg;
        AINS_REG_B(iins) = REG_ZERO;
        AINS_DATA(iins) = 0;
        AINS_REG_C(iins) = dReg;
        return 1;
    }
    else if( sType == DT_GFLT && dType == DT_GFLT )
    {
        AINS_CODE(iins) = I_cpys;
        AINS_REG_A(iins) = sReg;
        AINS_REG_B(iins) = sReg;
        AINS_REG_C(iins) = dReg;
        return 1;
    }
    else if( sType == DT_TFLT && dType == DT_TFLT )
    {
        AINS_CODE(iins) = I_cpys;
        AINS_REG_A(iins) = sReg;
        AINS_REG_B(iins) = sReg;
        AINS_REG_C(iins) = dReg;
        return 1;
    }
    else if( sType == DT_QUAD && dType == DT_QUAD )
    {
        AINS_CODE(iins) = I_lda;
        AINS_REG_A(iins) = REG_ZERO;
        AINS_REG_B(iins) = sReg;
        AINS_DATA(iins) = 0;
        AINS_REG_C(iins) = dReg;
        return 1;
    }
#if 0
    else if( sType == DT_QUAD && dType == DT_TFLT )
    {
        AINS_CODE(iins) = I_itoft;
        AINS_REG_A(iins) = sReg;
        AINS_REG_B(iins) = REG_ZERO;
        AINS_REG_C(iins) = dReg;
        return 1;
    }
    else if( sType == DT_LONG && dType == DT_SFLT )
    {
        AINS_CODE(iins) = I_itofs;
        AINS_REG_A(iins) = sReg;
        AINS_REG_B(iins) = REG_ZERO;
        AINS_REG_C(iins) = dReg;
        return 1;
    }
    else if( sType == DT_TFLT && dType == DT_QUAD )
    {
        AINS_CODE(iins) = I_ftoit;
        AINS_REG_A(iins) = sReg;
        AINS_REG_B(iins) = REG_ZERO;
        AINS_REG_C(iins) = dReg;
        return 1;
    }
    else if( sType == DT_SFLT && dType == DT_LONG )
    {
        AINS_CODE(iins) = I_ftoit;
        AINS_REG_A(iins) = sReg;
        AINS_REG_B(iins) = REG_ZERO;
        AINS_REG_C(iins) = dReg;
        return 1;
    }
#endif
    else
    {
        return 0;
    }
    
}

/*=======================================================
=======================================================*/

