/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.merge.c,v 1.9 1999/03/03 21:13:22 muth Exp $
  Comment:  This file contains the code for the condbranch
            instruction merge optimization
=======================================================*/

#include "global.h"
#include "flags.h"
#include "functions.h"
#include "opcode.h"
#include "regs.h"



/*=======================================================
@FUN
Mark those instructions that have no dependencies
within the given bbl, unmark the others.
Dependecies can either be register or resources.
This functioned currently uses the default memory
disambiguation which assumes that every store conflicts
with all loads and stores.
@END
=======================================================*/
#if 0
#undef __PROC__
#define __PROC__ "AbblMarkRootIns"

static void AbblMarkRootIns(INDEX ibbl)
{
    REGLIST reg_def;
    REGLIST reg_use;
    REGLIST res_def;
    REGLIST res_use;
    
    INDEX iins;

    reg_def = REGS_NONE;
    reg_use = REGS_NONE;
    res_def = 0;
    res_use = 0;
    
    FOREACH_BBL_INS(ibbl,iins)
    {
        REGLIST curr_reg_def = AinsDef(iins);
        REGLIST curr_reg_use = AinsUse(iins);
        REGLIST curr_res_def = AINS_RES_DEF(iins);
        REGLIST curr_res_use = AINS_RES_USE(iins);
        
        if( (curr_reg_use & reg_def) ||
            (curr_reg_def & (reg_def|reg_use)) ||
            (curr_res_use & res_def) ||
            (curr_res_def & (res_def|res_use)) )
        {
            AINS_FLAGS(iins) &= ~F_MARKED;
        }
        else
        {
            AINS_FLAGS(iins) |= F_MARKED;
        }

        reg_use |= curr_reg_use;
        reg_def |= curr_reg_def;
        res_use |= curr_res_use;
        res_def |= curr_res_def;
    }
}
#endif
/*=======================================================
@FUN
Do alias analysis
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblMarkRootInsFancy"

static void AbblMarkRootInsFancy(INDEX ibbl)
{
    REGLIST reg_def;
    REGLIST reg_use;
    REGLIST res_def;
    REGLIST res_use;
    
    INDEX iins;

    reg_def = REGS_NONE;
    reg_use = REGS_NONE;
    res_def = 0;
    res_use = 0;
    
    FOREACH_BBL_INS(ibbl,iins)
    {
        REGLIST curr_reg_def = AinsDef(iins);
        REGLIST curr_reg_use = AinsUse(iins);
        REGLIST curr_res_def = AINS_RES_DEF(iins) & ~RES_MM;
        REGLIST curr_res_use = AINS_RES_USE(iins) & ~RES_MM;
        
        if( (curr_reg_use & reg_def) ||
            (curr_reg_def & (reg_def|reg_use)) ||
            (curr_res_use & res_def) ||
            (curr_res_def & (res_def|res_use)) )
        {
            AINS_FLAGS(iins) &= ~F_MARKED;
        }
        else
        {
            AINS_FLAGS(iins) |= F_MARKED;
        }

        reg_use |= curr_reg_use;
        reg_def |= curr_reg_def;
        res_use |= curr_res_use;
        res_def |= curr_res_def;
    }

    FOREACH_BBL_INS(ibbl,iins)
        {
        INDEX jins;
        
        if( AINS_TYPE(iins) != IT_UST &&
            AINS_TYPE(iins) != IT_IST &&
            AINS_TYPE(iins) != IT_CST &&
            AINS_TYPE(iins) != IT_FST ) continue;

        for( jins = AINS_NEXT(iins); VALID(jins);  jins = AINS_NEXT(jins) )
        {
            if( AINS_TYPE(jins) != IT_UST &&
                AINS_TYPE(jins) != IT_IST &&
                AINS_TYPE(jins) != IT_CST &&
                AINS_TYPE(jins) != IT_FST &&

                AINS_TYPE(jins) != IT_ULD &&
                AINS_TYPE(jins) != IT_ILD &&
                AINS_TYPE(jins) != IT_LLD &&
                AINS_TYPE(jins) != IT_FLD ) continue;

            if( CONFLICT_NO == DoMemoryAccessesConflict(iins,jins) ) continue;

            AINS_FLAGS(jins) &= ~F_MARKED;
        }
    }    
}

/*=======================================================
=======================================================*/
#if 0
static void AbblTestMarkRoot(INDEX jbbl)
{
    NUMBER x=0,y=0;
    INDEX jins;
    
    AbblMarkRootIns(jbbl);
    FOREACH_BBL_INS(jbbl,jins)
    {
        if( AINS_FLAGS(jins) & F_MARKED ) x++;
    }
    AbblMarkRootInsFancy(jbbl);
    FOREACH_BBL_INS(jbbl,jins)
    {
        if( AINS_FLAGS(jins) & F_MARKED ) y++;
    }
    
    if( x != y )
    {
        PrintBblSuper(stdout,jbbl);
        printf("Mark Root...................... %d -> %d\n",x,y);
        FOREACH_BBL_INS(jbbl,jins)
        {
                if( AINS_FLAGS(jins) & F_MARKED ) PrintIns(stdout,jins);
        }
    }
}

#endif
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsEqual"

extern BOOL AinsEqual(INDEX iins,INDEX jins)
{
    if( AINS_CODE(iins) != AINS_CODE(jins) ) return 0;
    if( AINS_REG_C(iins) != AINS_REG_C(jins) ) return 0;
    if( AINS_REG_B(iins) != AINS_REG_B(jins) ) return 0;
    if( AINS_REG_A(iins) != AINS_REG_A(jins) ) return 0;
    if( AINS_DATA(iins) != AINS_DATA(jins) ) return 0;

    return 1;
}

/*=======================================================
@OPT
This optimization tries to move instructions which appear
both in the true and the false branch of a conditional
before that conditional. It does this by determining a set
of movable instructions in both basic blocks
and the checking for identical instructions in those sets.
Once an instruction has been moved the set of movable instructions
might change, so we restart the process.

Requiers: Liveness

Benefits from Register Renaming (to reduce the number of registers used)
since this often normalizes the register names it also provides
free registers for renaming activities
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptMergeInsUp"

extern BOOL OptMergeInsUp(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Mergeinsup");
    BOOL   change;
    INDEX  ibbl;
    INDEX  ibbl_t;
    INDEX  ibbl_f;
    INDEX  iins_branch;
    INDEX  iins_t;
    INDEX  iins_f;
    REG    reg_condition;
    
    NUMBER count = 0;
    NUMBER counth = 0;
    NUMBER countr = 0;
    NUMBER countp = 0;
    NUMBER countph = 0;
        
    clock_t time = clock();

    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_BRANCH ) continue;
        
        iins_branch = ABBL_INS_LAST(ibbl);
        reg_condition = AINS_REG_A(iins_branch);
        
        ibbl_t = AEDG_TAIL(ABBL_EDG_TRUE(ibbl));
        ibbl_f = AEDG_TAIL(ABBL_EDG_FALSE(ibbl));

        if( AbblNpreds(ibbl_t) != 1 ) continue;

        if( AbblNpreds(ibbl_f) != 1 ) continue;

    restart:
        AbblMarkRootInsFancy(ibbl_t);
        /*        AbblTestMarkRoot(ibbl_t); */
        
        AbblMarkRootInsFancy(ibbl_f);
        /*        AbblTestMarkRoot(ibbl_f);*/
        
        FOREACH_BBL_INS(ibbl_t,iins_t)
        {
            if( !(AINS_FLAGS(iins_t) & F_MARKED) ) continue;
            
            FOREACH_BBL_INS(ibbl_f,iins_f)
            {
                if( !(AINS_FLAGS(iins_f) & F_MARKED) ) continue;

                if( !AinsEqual(iins_t,iins_f) ) continue;

                if( AINS_TYPE(iins_t) == IT_JMP ||
                    AINS_TYPE(iins_t) == IT_BRA ||
                    AINS_TYPE(iins_t) == IT_IBR ||
                    AINS_TYPE(iins_t) == IT_FBR ||
                    AINS_TYPE(iins_t) == IT_PAL )
                {
                    if( verbose )
                    {
                        POTENTIAL(("Optimization of branches not implemented\n"));
                        PrintIns(stdout,iins_t);
                        PrintIns(stdout,iins_f);
                        PrintBblSuper(stdout,ibbl);
                        PrintBblSuper(stdout,ibbl_t);
                        PrintBblSuper(stdout,ibbl_f);
                    }
                    continue;
                }

                countp++;
                if( ABBL_IS_HOT(ibbl) ) countph++;
                
                if( AINS_REG_C(iins_t) == reg_condition )
                {
                    REGLIST regs_avail;
                    INDEX iins_def;
                    REG newreg;
                    
                    if( AinsUse(iins_t) & REG2BIT(reg_condition) ) continue;
                    
                    CERROR( ABBL_OUT(ibbl) & REG2BIT(reg_condition),("condition register should be dead\n"));

                    if( ABBL_OUT(ibbl) & REG2BIT(reg_condition) )
                    {
                        PrintIns(stdout,iins_t);
                        PrintBblSuper(stdout,ibbl);
                        PrintBblSuper(stdout,ibbl_f);
                        PrintBblSuper(stdout,ibbl_t);
                    }
                    
                    iins_def = FindInstructionRegDef(iins_branch, reg_condition);

                    if( !VALID(iins_def) ) continue;
                    
                    regs_avail =  MyFindScratchRegs(iins_def,INVALID);
                    
                    regs_avail &= REG_IS_INT(reg_condition) ?  REGS_INT_BUT_ZERO : REGS_FLT_BUT_ZERO;
                        
                    if( regs_avail == REGS_NONE ) continue;

                    newreg = RegsGetFirst(regs_avail);
                    
                    if( verbose )
                    {
                        printf("\t[*] renaming condition register %d -> %d\n",reg_condition,newreg);
                    }

                    reg_condition = newreg;
                    
                    RenameRegister(iins_def,newreg);
                    countr++;
#if 0
                    PrintBblSuper(stdout,ibbl);
#endif
                    
                }
                    
                count++;
                if( ABBL_IS_HOT(ibbl) ) counth++;
                
                if( verbose && ABBL_COUNT(ibbl) > 0 )
                {
                    HURRAH(("Merging\n"));
                    PrintIns(stdout,iins_t);
                    PrintIns(stdout,iins_f);
                    PrintBblSuper(stdout,ibbl);
                    PrintBblSuper(stdout,ibbl_t);
                    PrintBblSuper(stdout,ibbl_f);
                }

                                  
                AinsKill(iins_t);
                iins_t = AinsDup(iins_f);
                AINS_BBL(iins_t) = ibbl;
                AinsKill(iins_f);
                AinsInsertBefore(iins_t,iins_branch);

                /* reconstruct the liveness information */

                ABBL_OUT(ibbl) = AbblRegsLiveIn(ibbl_f) | AbblRegsLiveIn(ibbl_t);
                
                goto restart;
            }
            
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = (count>0);

    InfoStatus(fp,"\t[OptMergeinsup%c] merged iins: %d#%d potential: %d#%d  regs renamed: %d {%d ms}\n",
            CHANGE(change),count,counth,countp,countph,countr,time);
    
    return 0;   /* there should be only one run of this optimization */
}
/*=======================================================
=======================================================*/

/*=======================================================
@FUN
Mark those instructions that have no dependencies
within the given bbl, unmark the others.
Dependecies can either be register or resources.
This functioned currently uses the default memory
disambiguation which assumes that every store conflicts
with all loads and stores.
@END
=======================================================*/
#if 0
#undef __PROC__
#define __PROC__ "AbblMarkLeafIns"

static void AbblMarkLeafIns(INDEX ibbl)
{
    REGLIST reg_defuse;
    REGLIST res_defuse;
    
    INDEX iins;

    reg_defuse = REGS_NONE;
    res_defuse = 0;
    
    FOREACH_BBL_INS_R(ibbl,iins)
    {
        REGLIST curr_reg_def = AinsDef(iins);
        REGLIST curr_reg_use = AinsUse(iins);
        REGLIST curr_res_def = AINS_RES_DEF(iins);
        REGLIST curr_res_use = AINS_RES_USE(iins);
        
        if( (curr_reg_use & reg_defuse) ||
            (curr_reg_def & reg_defuse) ||
            (curr_res_use & res_defuse) ||
            (curr_res_def & res_defuse) )
        {
            AINS_FLAGS(iins) &= ~F_MARKED;
        }
        else
        {
            AINS_FLAGS(iins) |= F_MARKED;
        }

        reg_defuse |= curr_reg_use;
        reg_defuse |= curr_reg_def;
        res_defuse |= curr_res_use;
        res_defuse |= curr_res_def;
    }
    
}
#endif
/*=======================================================
@FUN
Does Alias analysis
@END
=======================================================*/

static void AbblMarkLeafInsFancy(INDEX ibbl)
{
    REGLIST reg_defuse;
    REGLIST res_defuse;
    
    INDEX iins;

    reg_defuse = REGS_NONE;
    res_defuse = 0;
    
    FOREACH_BBL_INS_R(ibbl,iins)
    {
        REGLIST curr_reg_def = AinsDef(iins);
        REGLIST curr_reg_use = AinsUse(iins);
        REGLIST curr_res_def = AINS_RES_DEF(iins) & ~RES_MM;
        REGLIST curr_res_use = AINS_RES_USE(iins) & ~RES_MM;
        
        if( (curr_reg_use & reg_defuse) ||
            (curr_reg_def & reg_defuse) ||
            (curr_res_use & res_defuse) ||
            (curr_res_def & res_defuse) )
        {
            AINS_FLAGS(iins) &= ~F_MARKED;
        }
        else
        {
            AINS_FLAGS(iins) |= F_MARKED;
        }

        reg_defuse |= curr_reg_use;
        reg_defuse |= curr_reg_def;
        res_defuse |= curr_res_use;
        res_defuse |= curr_res_def;
    }

    FOREACH_BBL_INS_R(ibbl,iins)
    {
        INDEX jins;

        if( AINS_TYPE(iins) != IT_UST &&
            AINS_TYPE(iins) != IT_IST &&
            AINS_TYPE(iins) != IT_CST &&
            AINS_TYPE(iins) != IT_FST ) continue;

        for( jins = AINS_PREV(iins); VALID(jins);  jins = AINS_PREV(jins) )
        {
            if( AINS_TYPE(jins) != IT_UST &&
                AINS_TYPE(jins) != IT_IST &&
                AINS_TYPE(jins) != IT_CST &&
                AINS_TYPE(jins) != IT_FST &&

                AINS_TYPE(jins) != IT_ULD &&
                AINS_TYPE(jins) != IT_ILD &&
                AINS_TYPE(jins) != IT_LLD &&
                AINS_TYPE(jins) != IT_FLD ) continue;


            if( CONFLICT_NO == DoMemoryAccessesConflict(iins,jins) ) continue;

            AINS_FLAGS(jins) &= ~F_MARKED;
        }
    }
}





/*=======================================================
@OPT
This optimization tries to move instructions which appear
both in the true and the false branch of a conditional
before that conditional. It does this by determining a set
of movable instructions in both basic blocks
and the checking for identical instructions in those sets.
Once an instruction has been moved the set of movable instructions
might change, so we restart the process.

Requiers: Liveness

Benefits from Register Renaming (to reduce the number of registers used)
since this often normalizes the register names it also provides
free registers for renaming activities
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptMergeInsUp"

extern BOOL OptMergeInsDown(FILE *fp)
{
#if 0
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Mergeinsup");
#endif
    BOOL   change;
    INDEX  ibbl;
    INDEX  jbbl;
    INDEX  xbbl;

    INDEX  xedg;
    INDEX  iedg;
    INDEX  jedg;

    INDEX xins;
    INDEX jins;
    
    NUMBER n;
    NUMBER count = 0;
        
    clock_t time = clock();
    
    FOREACH_LIVE_BBL(ibbl)
    {
        n=0;
        FOREACH_BBL_PRED(ibbl,iedg)
        {
            n++;
            if( AEDG_TYPE(iedg) != BT_NORM ) break;
        }

        if( VALID(iedg) ) continue;
        if( n < 2 ) continue;
    
    restart:
        FOREACH_BBL_PRED(ibbl,iedg)
        {
            jbbl = AEDG_HEAD(iedg);
            AbblMarkLeafInsFancy(jbbl);
        }

        xedg = ABBL_PRED(ibbl);
        xbbl = AEDG_HEAD(xedg);

        FOREACH_BBL_INS(xbbl,xins)
        {
            if( !(AINS_FLAGS(xins) & F_MARKED) ) continue;

            FOREACH_BBL_PRED(ibbl,jedg)
            {
                if( jedg == xedg ) continue;
                jbbl = AEDG_HEAD(jedg);
                FOREACH_BBL_INS(jbbl,jins)
                {
                    if( !(AINS_FLAGS(jins) & F_MARKED) ) continue;
                    if( !AinsEqual(jins,xins) ) continue;
                    break;
                }
                if( !VALID(jins) ) break;
            }
            if( VALID(jedg) ) continue;

#if 0
            printf("\t[*] QQQQQQQQQQQQQQQQQQQQQQQQQQQQQQ\n");
            PrintIns(stdout,xins);
            printf("\t[*] QQQQQQQQQQQQQQQQQQQQQQQQQQQQQQ\n");
            PrintBblSuper(stdout,ibbl);

            FOREACH_BBL_PRED(ibbl,jedg)
            {
                jbbl = AEDG_HEAD(jedg);
                PrintBblSuper(stdout,jbbl);
            }
#endif
            xins = AinsDup(xins);
            AINS_BBL(xins) = ibbl;
            /*
              the next part is necessary to keep the eval values
              and use-def-info correct
              */

            AINS_FLAGS(xins) &= ~(FI_EVAL_B|FI_EVAL_B|FI_EVAL_C);
            AINS_A_EVAL(xins) = INVALID;
            AINS_B_EVAL(xins) = INVALID;
            AINS_C_EVAL(xins) = INVALID;
            
                                  
            AinsPrepend(xins);
            count++;
            
            FOREACH_BBL_PRED(ibbl,jedg)
            {
                jbbl = AEDG_HEAD(jedg);
                FOREACH_BBL_INS(jbbl,jins)
                {
                    if( !(AINS_FLAGS(jins) & F_MARKED) ) continue;
                    if( !AinsEqual(jins,xins) ) continue;
                    AinsKill(jins);
                    count++;
                    break;
                }

                CASSERT( VALID(jins),("weird\n"));
            }

#if 0
            PrintBblSuper(stdout,ibbl);
            FOREACH_BBL_PRED(ibbl,jedg)
            {
                jbbl = AEDG_HEAD(jedg);
                PrintBblSuper(stdout,jbbl);
            }
#endif
            goto restart;
        }
    }
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = (count>0);
    
    InfoStatus(fp,"\t[OptMergeinsdown%c] killed iins: %d {%d ms}\n",
            CHANGE(change),count,time);
    
    return 0;   /* there should be only one run of this optimization */
}




