/*=======================================================
  Author:   R. Muth
  Version:  $Id: peephole.c,v 1.27 1999/02/01 18:19:10 muth Exp $
  Comment:  This file contains the code for various peephole optimizations
=======================================================*/

#include "global.h"
#include "functions.h"
#include "opcode.h"
#include "regs.h"
#include "flags.h"
#include <string.h>

/*=======================================================
=======================================================*/
#undef __PROC__ 
#define __PROC__ "PeepHoleMUL"

static BOOL PeepHoleMUL(INDEX iins)
{
    INDEX ibbl = AINS_BBL(iins);
    INDEX ifun = ABBL_FUN(ibbl);
    REG   src;
    REG   dst;
    INT64 val;
    
    
    if( AINS_FLAGS(iins) & FI_EVAL_B )
    {
        src = AINS_REG_A(iins);
        val = AINS_B_EVAL(iins);
        dst = AINS_REG_C(iins);
    }
    else if( AINS_FLAGS(iins) & FI_EVAL_A )
    {
        src = AINS_REG_B(iins);
        val = AINS_A_EVAL(iins);
        dst = AINS_REG_C(iins);
    }
    else
    {
        return 0;
    }

    if( val == 0 )
    {
        HURRAH(("peephole evaluated %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
        AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
        return 1;
    }
    

    switch( AINS_CODE(iins) )
    {
    case I_mulq:
        if( val == 1 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_addq;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = REG_ZERO;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
        }
        else if( val == 2 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_addq;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = src;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
        }
        else if( val == 3 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_s4subq;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = src;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
            
        }
        else if( val == 4 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_s4addq;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = REG_ZERO;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
        }
        else if( val == 5 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_s4addq;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = src;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
        }
        else if( val == 7 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_s8subq;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = src;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
        }
        else if( val == 8 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_s8addq;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = REG_ZERO;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
        }
        else if( val == 9 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_s8addq;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = src;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
        }
        break;

    case I_mull:
        if( val == 1 )
        {
#if 1
            PrintIns(stdout,iins);
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_addl;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = REG_ZERO;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            PrintIns(stdout,iins);

            return 1;
#else
            return 0;
#endif            
        }
        else if( val == 2 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_addl;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = src;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
        }
        else if( val == 3 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_s4subl;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = src;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
            
        }
        else if( val == 4 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_s4addl;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = REG_ZERO;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
        }
        else if( val == 5 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_s4addl;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = src;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
        }
        else if( val == 7 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_s8subl;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = src;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
        }
        else if( val == 8 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_s8addl;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = REG_ZERO;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
        }
        else if( val == 9 )
        {
            HURRAH(("peephole str-red %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AINS_CODE(iins) = I_s8addl;
            AINS_REG_C(iins) = dst;
            AINS_REG_B(iins) = src;
            AINS_REG_A(iins) = src;
            AINS_FLAGS(iins) &= ~(FI_EVAL_A|FI_EVAL_B|FI_EVAL_C);
            return 1;
        }
        break;
    }

    
    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PeepHoleIOP"

static BOOL PeepHoleIOP(INDEX iins)
{
    INDEX ibbl = AINS_BBL(iins);
    INDEX ifun = ABBL_FUN(ibbl);

    if( AINS_REG_A(iins) == AINS_REG_B(iins) )
    {
        if( AINS_CODE(iins) == I_subl || AINS_CODE(iins) == I_subq )
        {
            HURRAH(("peephole evaluated %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;
        }
    }

    return 0;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PeepHoleLOG"

#define HURRAHLOG(mess) 

static BOOL PeepHoleLOG(INDEX iins)
{
    /*
    INDEX ibbl = AINS_BBL(iins);
    INDEX ifun = ABBL_FUN(ibbl);
    */

    if( AINS_REG_A(iins) == AINS_REG_B(iins) )
    {
        switch( AINS_CODE(iins) )
        {
        case I_and:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            if( AINS_REG_B(iins) == REG_LIT )
                AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), AINS_DATA(iins));
            else
                AinsMakeLda(iins,AINS_REG_B(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_bis:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            if( AINS_REG_B(iins) == REG_LIT )
                AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), AINS_DATA(iins));
            else
                AinsMakeLda(iins,AINS_REG_B(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_bic:
            HURRAHLOG(("peephole evaluated %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;            
        case I_eqv:
            HURRAHLOG(("peephole evaluated %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), -1);
            return 1;
        case I_ornot:
            HURRAHLOG(("peephole evaluated %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), -1);
            return 1;
        case I_xor:
            HURRAHLOG(("peephole evaluated %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;
        }
    }
    
    if( AINS_REG_A(iins) == REG_ZERO || ((AINS_FLAGS(iins) & FI_EVAL_A) && AINS_A_EVAL(iins) == 0) )
    {
        switch( AINS_CODE(iins) )
        {
        case I_and:
            HURRAHLOG(("peephole evaluated %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;
        case I_bis:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            if( AINS_REG_B(iins) == REG_LIT )
                AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), AINS_DATA(iins));
            else
                AinsMakeLda(iins,AINS_REG_B(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_bic:
            HURRAHLOG(("peephole evaluated %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;            
        case I_eqv:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeNot(iins,AINS_REG_B(iins),AINS_REG_C(iins),AINS_DATA(iins));
            break;
        case I_ornot:
            /* this is the normal form for NOT !!! */
            break;
        case I_xor:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            if( AINS_REG_B(iins) == REG_LIT )
                AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), AINS_DATA(iins));
            else
                AinsMakeLda(iins,AINS_REG_B(iins), AINS_REG_C(iins), 0);
            return 1;
        }
    }

    if( AINS_REG_B(iins) == REG_ZERO || ((AINS_FLAGS(iins) & FI_EVAL_B) && AINS_B_EVAL(iins) == 0) )
    {
        switch( AINS_CODE(iins) )
        {
        case I_and:
            HURRAHLOG(("peephole evaluated %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;
        case I_bis:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,AINS_REG_A(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_bic:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,AINS_REG_A(iins), AINS_REG_C(iins), 0);
            return 1;            
        case I_eqv:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeNot(iins,AINS_REG_A(iins),AINS_REG_C(iins),AINS_DATA(iins));
            return 1;
        case I_ornot:
            HURRAHLOG(("peephole evaluated %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), -1);
            return 1;
        case I_xor:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,AINS_REG_A(iins), AINS_REG_C(iins), 0);
            return 1;
        }
    }


    if( (AINS_FLAGS(iins) & FI_EVAL_A) && AINS_A_EVAL(iins) == -1 )
    {
        switch( AINS_CODE(iins) )
        {
        case I_and:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            if( AINS_REG_B(iins) == REG_LIT )
                AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), AINS_DATA(iins));
            else
                AinsMakeLda(iins,AINS_REG_B(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_bis:
            HURRAHLOG(("peephole evaluated %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), -1);
            return 1;
        case I_bic:
            HURRAHLOG(("peephole evaluated %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeNot(iins,AINS_REG_B(iins), AINS_REG_C(iins), AINS_DATA(iins));
            return 1;
        case I_eqv:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            if( AINS_REG_B(iins) == REG_LIT )
                AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), AINS_DATA(iins));
            else
                AinsMakeLda(iins,AINS_REG_B(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_ornot:
            HURRAHLOG(("peephole evaluated %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), -1);
            return 1;
        case I_xor:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeNot(iins,AINS_REG_B(iins), AINS_REG_C(iins), AINS_DATA(iins));
            return 1;
        }
    }
    
    
    if( (AINS_FLAGS(iins) & FI_EVAL_B) && AINS_B_EVAL(iins) == -1 )
    {
        switch( AINS_CODE(iins) )
        {
        case I_and:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,AINS_REG_A(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_bis:
            HURRAHLOG(("peephole evaluated %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), -1);
            return 1;
        case I_bic:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;
        case I_eqv:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,AINS_REG_A(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_ornot:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,AINS_REG_A(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_xor:
            HURRAHLOG(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeNot(iins,AINS_REG_A(iins), AINS_REG_C(iins), AINS_DATA(iins));
            return 1;
        }
    }
    
    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PeepHoleSFT"

static BOOL PeepHoleSFT(INDEX iins)
{
    if( AINS_REG_A(iins) == REG_ZERO )
    {
#if 0
        INDEX ibbl = AINS_BBL(iins);
        INDEX ifun = ABBL_FUN(ibbl);
        HURRAH(("peephole evaluated %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
#endif
        AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
        return 1;
    }

    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PeepHoleICP"

static BOOL PeepHoleICP(INDEX iins)
{
    
    INDEX ibbl = AINS_BBL(iins);
    INDEX ifun = ABBL_FUN(ibbl);
    
    if( AINS_REG_A(iins) == AINS_REG_B(iins) )
    {
        /*HURRAH(("peephole evaluated %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));*/
            
        switch( AINS_CODE(iins) )
        {
        case I_cmpbge:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0xff);
            return 1;
        case I_cmple:
        case I_cmpeq:
        case I_cmpule:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 1);
            return 1;
        case I_cmpult:
        case I_cmplt:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;
        default:
            ERROR(("unexpected instruction %s\n",AINS_NAME(iins)));
        }
    }

    if( AINS_CODE(iins) == I_cmpult && AINS_REG_B(iins) == REG_ZERO )
    {
        /*HURRAH(("peephole evaluated %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));*/
        AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
        return 1;
    }

    if( AINS_CODE(iins) == I_cmpbge && AINS_REG_B(iins) == REG_ZERO )
    {
        
        /*HURRAH(("peephole evaluated %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));*/
        AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0xff);
        return 1;
    }

    if( AINS_CODE(iins) == I_cmpule && AINS_REG_A(iins) == REG_ZERO )
    {
        HURRAH(("peephole evaluated %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
        AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 1);
        return 1;
    }

    
    if( AINS_CODE(iins) == I_cmplt && (AINS_FLAGS(iins) & FI_EVAL_A)  && AINS_A_EVAL(iins) == -1 )
    {
        HURRAH(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
        AINS_CODE(iins) = I_cmple;
        AINS_REG_A(iins) = REG_ZERO;
        AINS_A_EVAL(iins) = 0;
        return 1;
    }
    
    return 0;
}


/*=======================================================
@OPT
Peephole optimizations and strength reductions
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptPeepHole"

extern BOOL OptPeepHole(FILE *fp)
{
#if 0
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Peep");
#endif
    BOOL change;
    NUMBER ecount=0;
    NUMBER ecounth=0;
    NUMBER kcount=0;
    NUMBER kcounth=0;
    NUMBER mcount=0;
    NUMBER mcounth=0;
    INDEX iins;
    INDEX ibbl;
    INDEX ifun;
    
    FOREACH_LIVE_INS(iins)
    {
        if( AINS_FLAGS(iins) & FI_EVAL_C ) continue;

        ibbl = AINS_BBL(iins);
        ifun = ABBL_FUN(ibbl);

        switch( AINS_TYPE(iins) )
        {
        case IT_ICP:
            if( PeepHoleICP(iins) )
            {
                ecount++;
                if( ABBL_IS_HOT(ibbl) )
                    ecounth++;
            }
            break;

        case IT_IOP:
            if( PeepHoleIOP(iins) )
            {
                ecount++;
                if( ABBL_IS_HOT(ibbl) )
                    ecounth++;
            }
            break;
        case IT_LOG:
            if( PeepHoleLOG(iins) )
            {
                ecount++;
                if( ABBL_IS_HOT(ibbl) )
                    ecounth++;
            }
            break;
        case IT_SFT:
            if( PeepHoleSFT(iins) )
            {
                ecount++;
                if( ABBL_IS_HOT(ibbl) )
                    ecounth++;
            }
            break;

        case IT_MLL:
        case IT_MLQ:
        case IT_MLH:
            if( PeepHoleMUL(iins) )
            {
                mcount++;
                if( ABBL_IS_HOT(ibbl) )
                    mcounth++;
            }
            break;

        default:
                break;
        }
    }
    

    change = (ecount+kcount)>0;

    InfoStatus(fp,"\t[OptPeep%c] Eval Inss: %d#%d   Mod Ins: %d#%d  Killed Inss: %d#%d\n",
            CHANGE(change),ecount,ecounth,mcount,mcounth,kcount,kcounth);

        
    return change;
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "AbblFindSuitablePredecessor"

static INDEX AbblFindSuitablePredecessor(INDEX ibbl,INDEX succ)
{
    INDEX iedg;
    INDEX pred;
    

    FOREACH_BBL_PRED(ibbl,iedg)
    {
        pred = AEDG_HEAD(iedg);
        if( ABBL_TYPE(pred) != BT_BRANCH ) continue;
        if( AEDG_TAIL(ABBL_EDG_TRUE(pred)) != succ && AEDG_TAIL(ABBL_EDG_TRUE(pred)) != ibbl ) continue;
        if( AEDG_TAIL(ABBL_EDG_FALSE(pred)) != succ && AEDG_TAIL(ABBL_EDG_FALSE(pred)) != ibbl ) continue;

        return pred;
    }

    return INVALID;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblHasBadIns"

static BOOL AbblHasBadIns(INDEX ibbl)
{
    INDEX iins;
    
    FOREACH_BBL_INS(ibbl,iins)
    {
        switch( AINS_TYPE(iins) )
        {
        case IT_ILD:
        case IT_IST:
        case IT_ULD:
        case IT_UST:
        case IT_FLD:
        case IT_FST:
            return 1;
        case IT_LDA:
        case IT_ICP: 
        case IT_LOG:
        case IT_SFT:
        case IT_IOP:
        default:
            /* using live register we could do some tricks here like
               computing a value into an unused regsister and only use
               it, ie. conditionally move it, if the condition is true */
            break;
            

        }
    }
    
    return 0;
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "AbblsIsNotInteresting"

static BOOL AbblIsNotInteresting(INDEX ibbl,REG reg)
{
    INDEX iins;
    NUMBER new_instructions = 0;
    
    if( REG_IS_INT(reg) )
    {
        FOREACH_BBL_INS(ibbl,iins)
        {
            if( AINS_REG_C(iins) == reg && VALID(AINS_NEXT(iins)) ) return 1;
            
            switch( AINS_TYPE(iins) )
            {
            case IT_LDA:
                if( AINS_CODE(iins) != I_lda ) return 1;
                    if( AINS_DATA(iins) == 0 ) 
                    {
                        new_instructions++;
                    }
                    else if( AINS_DATA(iins) > 0 && AINS_DATA(iins) < 256 && AINS_REG_B(iins) == REG_ZERO )
                    {
                        new_instructions++;
                    }
                    else
                    {
                        return 1;
                    }
                    break;

            case IT_ICP: 
            case IT_LOG:
            case IT_SFT:
            case IT_IOP:
                /* using live register we could do some tricks here like
                   computing a value into an unused regsister and only use
                   it, ie. conditionally move it, if the condition is true */
                return 1;

            default:
                return 1;
            }
        }
        
        if( new_instructions > 2 ) return 1;
    }
    else
    {
        FOREACH_BBL_INS(ibbl,iins)
        {
            if( AINS_REG_C(iins) == reg && VALID(AINS_NEXT(iins)) ) return 1;
            
            switch( AINS_TYPE(iins) )
            {
            case IT_FCY:
                CASSERT( AINS_CODE(iins) == I_cpys,("weird\n"));
                if( AINS_REG_A(iins) != AINS_REG_B(iins) ) return 1;
                new_instructions++;
                break;

            default:
                return 1;
            }
        }
        
        if( new_instructions > 2 ) return 1;
        
    }
    
    return 0;
}

/*=======================================================
=======================================================*/



static struct
{
    NUMBER branch;
    NUMBER true;
    NUMBER false;
}
OpcodeTranslator[] =
{
    {I_blbc, I_cmovlbc, I_cmovlbs},
    {I_blbs, I_cmovlbs, I_cmovlbc},

    {I_beq,  I_cmoveq,  I_cmovne },
    {I_bne,  I_cmovne,  I_cmoveq },
    {I_ble,  I_cmovle,  I_cmovgt }, 
    {I_bgt,  I_cmovgt,  I_cmovle },
    {I_bge,  I_cmovge,  I_cmovlt },
    {I_blt,  I_cmovlt,  I_cmovge },

    {I_fbeq,  I_fcmoveq,  I_fcmovne },
    {I_fbne,  I_fcmovne,  I_fcmoveq },
    {I_fble,  I_fcmovle,  I_fcmovgt }, 
    {I_fbgt,  I_fcmovgt,  I_fcmovle },
    {I_fbge,  I_fcmovge,  I_fcmovlt },
    {I_fblt,  I_fcmovlt,  I_fcmovge },
    {0, 0, 0}
    
};

/*=======================================================
@OPT
Conditional move generation
Conditional moves are introduced for the basic block on the left hand side

@example
      O
      |\
      | O
      |/
      O
@end example

@END
=======================================================*/

#undef __PROC__
#define __PROC__ "OptCondMoveOne"

extern BOOL OptCondMoveOne(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Condmoveone");
    BOOL change;
    NUMBER count=0;
    NUMBER counth=0;
    NUMBER fcount=0;
    NUMBER fcounth=0;
    NUMBER i;
    UINT16 opcode;
    
    INDEX ibbl;
    INDEX pred;
    INDEX succ;
    INDEX iins;
    REG   reg_condition;
    
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_NORM ) continue;
        succ = AEDG_TAIL(ABBL_EDG_NORM(ibbl));
        pred = AbblFindSuitablePredecessor(ibbl,succ);
        if( !VALID(pred) ) continue;
        
        iins = ABBL_INS_LAST(pred);
        CASSERT( AINS_TYPE(iins) == IT_IBR ||
                 AINS_TYPE(iins) == IT_FBR,("unexpected instruction\n"));

        reg_condition = AINS_REG_A(iins);

        opcode = 0;
        
        for(i=0; OpcodeTranslator[i].branch!=0;i++)
        {
            if( OpcodeTranslator[i].branch == AINS_CODE(iins) )
            {
                if( AEDG_TAIL(ABBL_EDG_TRUE(pred)) == ibbl )
                    opcode = OpcodeTranslator[i].true;
                else
                    opcode = OpcodeTranslator[i].false;    
                break;
            }
            
        }
        CASSERT( opcode!=0,("weird\n"));

#if 0
        if( verbose )
        {
            fprintf(fp,"**Before:\n");
            
            PrintBblSuper(fp, pred);
            PrintBblSuper(fp, ibbl);
        }
#endif

#if 0
        if( ABBL_IS_HOT(ibbl) && !AbblHasBadIns(ibbl) )
        {
            HURRAH(("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"));
            PrintBblSuper(stdout,pred);
            PrintBblSuper(stdout,ibbl);
            PrintBblSuper(stdout,succ);
        }
#endif
        
        if( AbblIsNotInteresting(ibbl,reg_condition) ) continue;


        if( verbose )
        {
            fprintf(fp,"Before:\n");
            
            PrintBblSuper(fp, pred);
            PrintBblSuper(fp, ibbl);
        }

        
        AedgKillFancy(ABBL_EDG_TRUE(pred));
        AedgKillFancy(ABBL_EDG_FALSE(pred));
        AinsKill(iins);
        ABBL_TYPE(pred) = BT_NORM;
        AedgLinkFancy( AedgNewFancy(pred,succ,ET_NORM) );        


        FOREACH_BBL_INS(ibbl,iins)
        {
            INDEX jins = AinsNew();

            AINS_BBL(jins) = pred;
            AINS_CODE(jins) = opcode;

            switch( AINS_TYPE(iins) )
            {
            case IT_LDA:
                if( AINS_DATA(iins) == 0 )
                {
                    
                    AINS_REG_A(jins) = reg_condition;
                    AINS_REG_B(jins) = AINS_REG_B(iins);
                    AINS_REG_C(jins) = AINS_REG_C(iins);
                }
                else if( AINS_DATA(iins) > 0 && AINS_DATA(iins) < 256 && AINS_REG_B(iins) == REG_ZERO )
                {
                    AINS_REG_A(jins) = reg_condition;
                    AINS_REG_B(jins) = REG_LIT;
                    AINS_DATA(jins) = AINS_DATA(iins);
                    AINS_REG_C(jins) = AINS_REG_C(iins);                
                }
                else
                {
                    ERROR(("unexpected case\n"));
                }
                break;

            case IT_FCY:
                AINS_REG_A(jins) = reg_condition;
                AINS_REG_B(jins) = AINS_REG_B(iins);
                AINS_REG_C(jins) = AINS_REG_C(iins);
                break;

            default:
                ERROR(("unexpected case\n"));
            }
            
           
            AinsAppend(jins);
        }
        
        if( verbose )
        {
            fprintf(fp,"After:\n");
            PrintBblSuper(fp, pred);
        }

        if( REG_IS_INT(reg_condition) )
        {
            count++;
            if( ABBL_IS_HOT(pred) ) counth++;
        }
        else
        {
            fcount++;
            if( ABBL_IS_HOT(pred) ) fcounth++;
        }
        
    }

    change = (count)>0;

    InfoStatus(fp,"\t[OptCondMoveOne%c] Added ICMs: %d#%d  Added FCM: %d#%d\n",
            CHANGE(change),count,counth,fcount,fcounth);

    return change;
}


/*=======================================================
@OPT

Conditional move generation.
Conditional moves are introduced for the basic blocks in the middle

@example
      O
     / \
    O   O
     \ /
      O
@end example
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptCondMoveTwo"

extern BOOL OptCondMoveTwo(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Condmovetwo");
    BOOL change;
    NUMBER count=0;
    NUMBER counth=0;
    NUMBER fcount=0;
    NUMBER fcounth=0;
    NUMBER opcode;
    
    INDEX iins;
    INDEX jins;
    INDEX iins_t;
    INDEX iins_f;

    INDEX ibbl;
    INDEX ibbl_t;
    INDEX ibbl_f;
    INDEX ibbl_join;

    REGLIST avail;
    REG   reg;
    REG   reg_condition;
        
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_BRANCH ) continue;
        iins = ABBL_INS_LAST(ibbl);
        CASSERT( AINS_TYPE(iins) == IT_IBR ||
                 AINS_TYPE(iins) == IT_FBR,("unexpected instruction\n"));
        reg_condition = AINS_REG_A(iins);

        ibbl_t = AEDG_TAIL(ABBL_EDG_TRUE(ibbl));
        ibbl_f = AEDG_TAIL(ABBL_EDG_FALSE(ibbl));
        
        if( ABBL_TYPE(ibbl_f) != BT_NORM )  continue;
        if( ABBL_TYPE(ibbl_t) != BT_NORM )  continue;

        ibbl_join = AEDG_TAIL(ABBL_EDG_NORM(ibbl_t));
        if( AEDG_TAIL(ABBL_EDG_NORM(ibbl_f)) != ibbl_join ) continue;

#if 0
        if( ABBL_IS_HOT(ibbl) && !AbblHasBadIns(ibbl_f) && !AbblHasBadIns(ibbl_t) )
        {
            HURRAH(("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"));
            PrintBblSuper(stdout,ibbl);
            PrintBblSuper(stdout,ibbl_t);
            PrintBblSuper(stdout,ibbl_f);
            PrintBblSuper(stdout,ibbl_join);
        }
#endif
        
        if( AbblNInss(ibbl_t) > 1 ) continue;
        if( AbblNInss(ibbl_f) > 1 ) continue;

#if 0
        if( AbblNpreds(ibbl_f) > 1 ) continue;
        if( AbblNpreds(ibbl_t) > 1 ) continue;
#endif
        
        iins_f = ABBL_INS_FIRST(ibbl_f);
        iins_t = ABBL_INS_FIRST(ibbl_t);
#if 0        
        if( AbblNpreds(ibbl_join) > 2 ) continue;
#endif

            
        for(opcode=0; OpcodeTranslator[opcode].branch!=0;opcode++)
        {
            if( OpcodeTranslator[opcode].branch == AINS_CODE(iins) )
                  break;
         }
 
        /* check whether instructions iins_t and iins_f are suitable */
        
        if( REG_IS_INT(reg_condition) )
        {
            if( AINS_CODE(iins_f) != I_lda ) continue;
            if( AINS_DATA(iins_f) !=0  &&
                (AINS_DATA(iins_f) < 0 || AINS_DATA(iins_f) >= 256 || AINS_REG_B(iins_f) != REG_ZERO) ) continue;
            if( AINS_CODE(iins_t) != I_lda ) continue;
            if( AINS_DATA(iins_t) !=0  &&
                (AINS_DATA(iins_t) < 0 || AINS_DATA(iins_t) >= 256 || AINS_REG_B(iins_t) != REG_ZERO) ) continue;

#if 0
            /* we can handle this now gracefully */
            CASSERT( AINS_REG_C(iins_t) != reg_condition,("weird\n"));
            CASSERT( AINS_REG_C(iins_f) != reg_condition,("weird\n"));
#endif
        }
        else
        {
            if( AINS_CODE(iins_f) != I_cpys ) continue;
            if( AINS_REG_A(iins_f) != AINS_REG_B(iins_f) ) continue;
            if( AINS_CODE(iins_t) != I_cpys ) continue;
            if( AINS_REG_A(iins_t) != AINS_REG_B(iins_t) ) continue;
#if 0
            /* we can handle this now gracefully */
            CASSERT( AINS_REG_C(iins_t) != reg_condition,("weird\n"));
            CASSERT( AINS_REG_C(iins_f) != reg_condition,("weird\n"));
#endif
        }
        
        /*
          check whether the condition register might get overwritten 
          we need live register analysis here
         */

        if( AINS_REG_C(iins_f) == reg_condition || AINS_REG_C(iins_t) == reg_condition )
        {
            if( REG_IS_INT(reg_condition) )
            {
                avail = REGS_INT_BUT_ZERO & ~ABBL_OUT(ibbl);
                avail &= ~REG2BIT(reg_condition);
                avail &= ~REG2BIT(AINS_REG_C(iins_t));
                avail &= ~REG2BIT(AINS_REG_C(iins_f));
                if( avail == REGS_NONE )
                {
                    WARNING(("failed to find a free register \n"));
                    continue;
                }
                                    
                reg = RegsGetFirst(avail);
                jins = AinsNew();
                AINS_BBL(jins) = ibbl;
                AinsMakeLda(jins,reg_condition,reg,0);
                AinsAppend(jins);
            }
            else
            {
                avail = REGS_FLT_BUT_ZERO & ~ABBL_OUT(ibbl);
                if( avail == REGS_NONE )
                {
                    WARNING(("failed\n"));
                    continue;
                }
                WARNING(("succeeded\n"));
                
                reg = RegsGetFirst(avail);
                jins = AinsNew();
                AINS_BBL(jins) = ibbl;
                AinsMakeCpys(jins,reg_condition,reg_condition,reg);
                AinsAppend(jins);
            }

            reg_condition = reg;
        }
        
                
            
        if( verbose )
        {
            fprintf(fp,"Before\n");
            PrintBblSuper(fp, ibbl);
            PrintBblSuper(fp, ibbl_t);
            PrintBblSuper(fp, ibbl_f);
            PrintBblSuper(fp, ibbl_join);
        }
            


        if( REG_IS_INT(reg_condition) )
        {
            jins = AinsNew();
            AINS_BBL(jins) = ibbl;

            if( AINS_REG_C(iins_t) == AINS_REG_C(iins_f) )
            {
                AinsMakeLda(jins,AINS_REG_B(iins_t),AINS_REG_C(iins_t),AINS_DATA(iins_t));
            }
            else
            {
                AINS_REG_A(jins) = reg_condition;
                AINS_CODE(jins) = OpcodeTranslator[opcode].true;
                if( AINS_DATA(iins_t) == 0 )
                {
                    AINS_REG_B(jins) = AINS_REG_B(iins_t);
                    AINS_REG_C(jins) = AINS_REG_C(iins_t);
                }
                else 
                {
                    AINS_REG_B(jins) = REG_LIT;
                    AINS_DATA(jins) = AINS_DATA(iins_t);
                    AINS_REG_C(jins) = AINS_REG_C(iins_t);                
                }
            }
            
            AinsAppend(jins);
            
            jins = AinsNew();
            AINS_BBL(jins) = ibbl;
            AINS_REG_A(jins) = reg_condition;
            AINS_CODE(jins) = OpcodeTranslator[opcode].false;
                        
            if( AINS_DATA(iins_f) == 0 )
            {
                AINS_REG_B(jins) = AINS_REG_B(iins_f);
                AINS_REG_C(jins) = AINS_REG_C(iins_f);
            }
            else 
            {
                AINS_REG_B(jins) = REG_LIT;
                AINS_DATA(jins) = AINS_DATA(iins_f);
                AINS_REG_C(jins) = AINS_REG_C(iins_f);                
            }
            AinsAppend(jins);

        }
        else
        {
            jins = AinsNew();
            AINS_BBL(jins) = ibbl;

            if( AINS_REG_C(iins_t) == AINS_REG_C(iins_f) )
            {
                AinsMakeCpys(jins,AINS_REG_A(iins_t),AINS_REG_B(iins_t),AINS_REG_C(iins_t));
            }
            else
            {
                AINS_REG_A(jins) = reg_condition;
                AINS_CODE(jins) = OpcodeTranslator[opcode].true;
                AINS_REG_B(jins) = AINS_REG_B(iins_t);
                AINS_REG_C(jins) = AINS_REG_C(iins_t);
            }
            AinsAppend(jins);
                        
            jins = AinsNew();
            AINS_REG_A(jins) = reg_condition;
            AINS_CODE(jins) = OpcodeTranslator[opcode].false;
            AINS_BBL(jins) = ibbl;
            AINS_REG_B(jins) = AINS_REG_B(iins_f);
            AINS_REG_C(jins) = AINS_REG_C(iins_f);
            AinsAppend(jins);
        }
        
        AedgKillFancy(ABBL_EDG_TRUE(ibbl));
        AedgKillFancy(ABBL_EDG_FALSE(ibbl));
        AinsKill(iins);
        ABBL_TYPE(ibbl) = BT_NORM;
        AedgLinkFancy( AedgNewFancy(ibbl,ibbl_join,ET_NORM) );
        
        
        if( verbose )
        {
            fprintf(fp,"After\n");
            PrintBblSuper(fp, ibbl);
            PrintBblSuper(fp, ibbl_t);
            PrintBblSuper(fp, ibbl_f);
            PrintBblSuper(fp, ibbl_join);
        }


        if( REG_IS_INT(reg_condition) )
        {
            count++;
            if( ABBL_IS_HOT(ibbl) ) counth++;
        }
        else
        {
            fcount++;
            if( ABBL_IS_HOT(ibbl) ) fcounth++;
        }
    }
    
    change = (count)>0;

    InfoStatus(fp,"\t[OptCondMoveTwo%c] Added ICMs: %d#%d  Added FCM: %d#%d\n",
            CHANGE(change),count,counth,fcount,fcounth);

    return change;
}


/*=======================================================
@FUN

one and two set should be disjoint!

@END
=======================================================*/


static NUMBER OpTransNegateOpOne[] =
{
    I_blbc,I_blbs,
    I_blbs,I_blbc,
    I_beq,I_bne,
    I_bne,I_beq,

    /*  these should not occur */
#if 0
    I_ble,I_bgt, 
    I_bgt,I_ble,
    I_bge,I_blt,
    I_blt,I_bge,
#endif
    
    I_cmovne,I_cmoveq,
    I_cmoveq,I_cmovne,
    I_cmovgt,I_cmovle,
    I_cmovle,I_cmovgt,
    I_cmovlt,I_cmovge,
    I_cmovge,I_cmovlt,
    I_cmovlbc,I_cmovlbs,
    I_cmovlbs,I_cmovlbc,

    /* these require operand switching */
#if 0    
    I_bis,I_ornot,
    I_xor,I_eqv,
    I_and.I_bic,
#endif
    INVALID,INVALID
};


static NUMBER OpTransNegateOpTwo[] =
{
#if 0
    I_bis,I_ornot,
    I_ornot,I_bis,
    I_xor,I_eqv,
    I_eqv,I_xor,
    I_and,I_bic,
    I_bic,I_and,
#endif
    INVALID,INVALID
};


static NUMBER OpTransInvertOpOne[] =
{
    I_beq,I_beq,   /* no change !!! */
    I_bne,I_bne,   /* no change !!! */
    I_ble,I_bgt,
    I_bgt,I_ble,
    I_bge,I_blt,
    I_blt,I_bge,

    I_cmovne,I_cmovne, /* no change !!! */
    I_cmoveq,I_cmoveq, /* no change !!! */
    I_cmovge,I_cmovle,
    I_cmovle,I_cmovge,
    I_cmovlt,I_cmovgt,
    I_cmovgt,I_cmovlt,

    /* these require operand switching */
#if 0    
    I_addq,I_subq,
    I_addl,I_subl,
#endif
    INVALID,INVALID
};


static NUMBER OpTransInvertOpTwo[] =
{
    I_addq,I_subq,
    I_subq,I_addq,
    I_addl,I_subl,
    I_subl,I_addl,

    INVALID,INVALID
};



static NUMBER OpcodeTranslate(NUMBER opcode,NUMBER *OpTrans)
{
    while( *OpTrans != INVALID )
    {
        if( *OpTrans == opcode )
        {
            return OpTrans[1];
        }

        OpTrans+=2;
    }
    
    return INVALID;
}



/*=======================================================
@FUN
@END
=======================================================*/
static BOOL AinsTransform(INDEX iins_the,NUMBER *OpOneTrans, NUMBER *OpTwoTrans, BOOL DoTrans )
{
    INDEX reg;
    INDEX jins;
    NUMBER newop;
    
    reg = AINS_REG_C(iins_the);

    
    for(jins=AINS_NEXT(iins_the); VALID(jins); jins = AINS_NEXT(jins))
    {
        if( AINS_REG_B(jins) == reg )
        {
            newop = OpcodeTranslate(AINS_CODE(jins),OpTwoTrans);
            if( !VALID(newop) )
            {
                CERROR( DoTrans,("could not do transform!\n"));
                return 0;
            }

            if( DoTrans )
            {
                AINS_CODE(jins) = newop;
            }
            
        }
        
        if( AINS_REG_A(jins) == reg )
        {
            newop = OpcodeTranslate(AINS_CODE(jins),OpOneTrans);
            if( !VALID(newop) )
            {
                CERROR( DoTrans,("could not do transform!\n"));
                return 0;
            }
            
            if( DoTrans )
            {
                AINS_CODE(jins) = newop;
            }
        }
        if
            ( AINS_REG_C(jins) == reg )
        {
            if( AINS_TYPE(jins) == IT_ICM || AINS_TYPE(jins) == IT_FCM ) return 0;
            else break;  /* ok we are fine */
        }
    }

    if( !VALID(jins) )
    {
        if( ABBL_OUT(AINS_BBL(iins_the)) & REG2BIT(reg) )
        {
            CERROR( DoTrans,("could not do transform!\n"));
            return 0;
        }
        
    }

    return 1;
}

/*=======================================================
@FUN
@END
=======================================================*/

static void AinsSwitchOperands(INDEX iins)
{
    REG reg_tmp;
    INT64 val_tmp;
    
    CASSERT( AINS_REG_B(iins) != REG_LIT,("cannot switch iins %d\n",iins));

    reg_tmp = AINS_REG_A(iins);
    AINS_REG_A(iins) = AINS_REG_B(iins);
    AINS_REG_B(iins) = reg_tmp;

    val_tmp = AINS_A_EVAL(iins);
    AINS_A_EVAL(iins) = AINS_B_EVAL(iins);
    AINS_B_EVAL(iins) = val_tmp;


    if( AINS_FLAGS(iins) & FI_EVAL_A )
    {
        if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
        {
            AINS_FLAGS(iins) &= ~FI_EVAL_A;
            AINS_FLAGS(iins) |= FI_EVAL_B;            
        }
    }
    else
    {
        if( AINS_FLAGS(iins) & FI_EVAL_B )
        {
            AINS_FLAGS(iins) &= ~FI_EVAL_B;
            AINS_FLAGS(iins) |= FI_EVAL_A;            
        }
    }
}


/*=======================================================
@OPT
Some alpha instruction can take a small immediate value instead of register B.
Now suppose that register A evaluates to such a small value, this optimization
tries to switch register A and B so that we can use the immediate value feature.
Example: xor r1,r2,r3 (we knwo that r1 contains the value 66)
         ----> xor r2,#66,r3 (using commutativity)
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptLiteral"

extern NUMBER OptLiteral(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Literal");
    BOOL change;
    NUMBER count=0;
    NUMBER counth=0;

    INDEX ifun;
    INDEX ibbl;
    INDEX iins;
    
    FOREACH_LIVE_FUN(ifun)
    FOREACH_FUN_BBL(ifun,ibbl)
    FOREACH_BBL_INS(ibbl,iins)
    {
        
        /* use commutativity to make the second operand the constant one, do not switch REG_ZERO */
        
        if( (AINS_FLAGS(iins) & FI_EVAL_A)  &&
            !(AINS_FLAGS(iins) & FI_EVAL_B) &&
            AINS_REG_A(iins) != REG_ZERO &&
            AINS_REG_B(iins) != REG_LIT &&
            AINS_A_EVAL(iins) < 256 &&  AINS_A_EVAL(iins) > 0 ) 
        {
            if( AINS_CODE(iins) == I_addq ||
                AINS_CODE(iins) == I_addl ||
                AINS_CODE(iins) == I_and ||
                AINS_CODE(iins) == I_bis ||
                AINS_CODE(iins) == I_cmpeq ||
                AINS_CODE(iins) == I_eqv ||
                AINS_CODE(iins) == I_mull ||
                AINS_CODE(iins) == I_mulq ||
                AINS_CODE(iins) == I_xor )
            {
                if( verbose )
                {
                    fprintf(fp,"\t[*] switching operands %s %016lx in %d [%s] %ld\n",
                            AINS_NAME(iins),AINS_A_EVAL(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl));
                }
                

                AinsSwitchOperands(iins);
                count++;
                if( ABBL_IS_HOT(ibbl) )
                    counth++;
            }

            if( AINS_CODE(iins) == I_cmple ||
                AINS_CODE(iins) == I_cmplt ||
                AINS_CODE(iins) == I_cmpule ||
                AINS_CODE(iins) == I_cmpult)
            {
                if( AinsTransform(iins,OpTransNegateOpOne,OpTransNegateOpTwo,0) )
                {
                    if( verbose )
                    {
                        fprintf(fp,"\t[*] negating %s %016lx in %d [%s] %ld\n",
                                AINS_NAME(iins),AINS_A_EVAL(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl));
                        PrintIns(stdout,iins);
                        PrintBblSuper(stdout,ibbl);
                    }

                    AinsTransform(iins,OpTransNegateOpOne,OpTransNegateOpTwo,1);
                    AinsSwitchOperands(iins);
                    switch( AINS_CODE(iins) )
                    {
                    case I_cmple:
                        AINS_CODE(iins) = I_cmplt;
                        break;
                    case I_cmplt:
                        AINS_CODE(iins) = I_cmple;
                        break;
                    case I_cmpule:
                        AINS_CODE(iins) = I_cmpult;
                        break;
                    case I_cmpult:
                        AINS_CODE(iins) = I_cmpule;
                        break;
                    default:
                        ERROR(("unexpected opcode\n"));
                    }
                    
                    if( verbose )
                    {
                        PrintBblSuper(stdout,ibbl);
                    }

                    count++;
                    if( ABBL_IS_HOT(ibbl) )
                        counth++;
                }

            }
            

            if( AINS_CODE(iins) == I_subq ||
                AINS_CODE(iins) == I_subl )
            {
                if( AinsTransform(iins,OpTransInvertOpOne,OpTransInvertOpTwo,0) )
                {
                    if( verbose )
                    {
                        fprintf(fp,"\t[*] inverting %s %016lx in %d [%s] %ld\n",
                                AINS_NAME(iins),AINS_A_EVAL(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl));
                        PrintIns(stdout,iins);
                        PrintBblSuper(stdout,ibbl);
                    }
                                        
                    AinsTransform(iins,OpTransInvertOpOne,OpTransInvertOpTwo,1);
                    AinsSwitchOperands(iins);
                                    
                    if( verbose )
                    {
                        PrintBblSuper(stdout,ibbl);
                    }

                    count++;
                    if( ABBL_IS_HOT(ibbl) )
                        counth++;
                }
                
            }
        }

        if( (AINS_FLAGS(iins) & FI_EVAL_B) && ((INT32)AINS_B_EVAL(iins)) < 0 && ((INT32)AINS_B_EVAL(iins)) > -256)
        {
            if( AINS_CODE(iins) == I_addl )
            {
                HURRAH(("addl -> subl (untested)\n"));
                                
                AINS_CODE(iins) = I_subl;
                AINS_REG_B(iins) = REG_LIT;
                AINS_DATA(iins) =   -((INT32)AINS_B_EVAL(iins));
                AINS_B_EVAL(iins) = -((INT32)AINS_B_EVAL(iins));
                count++;
                if( ABBL_IS_HOT(ibbl) )
                    counth++;

            }
            else if( AINS_CODE(iins) == I_subl )
            {
                HURRAH(("subl -> addl (untested)\n"));
                
                AINS_CODE(iins) = I_addl;
                AINS_REG_B(iins) = REG_LIT;
                AINS_DATA(iins) =   -((INT32)AINS_B_EVAL(iins));
                AINS_B_EVAL(iins) = -((INT32)AINS_B_EVAL(iins));
                count++;
                if( ABBL_IS_HOT(ibbl) )
                    counth++;

            }
        }
    }
    
    
    change = (count)>0;

    InfoStatus(fp,"\t[OptLiteral%c] changed: %d#%d\n",CHANGE(change),count,counth);

    return change;
}

/*=======================================================
=======================================================*/

static NUMBER TabEqNe[] =
{
    I_beq,I_bne,
    I_bne,I_beq,
    I_cmoveq,I_cmovne,
    I_cmovne,I_cmoveq,
    INVALID,INVALID
};

static NUMBER TabEqEq[] =
{
    I_beq,I_beq,
    I_bne,I_bne,
    I_cmoveq,I_cmoveq,
    I_cmovne,I_cmovne,
    INVALID,INVALID
};

static NUMBER TabEqLt[] =
{
    I_beq,I_blt,
    I_bne,I_bge,
    I_cmoveq,I_cmovlt,
    I_cmovne,I_cmovge,
    INVALID,INVALID
};

static NUMBER TabEqGt[] =
{
    I_beq,I_bgt,
    I_bne,I_ble,
    I_cmoveq,I_cmovgt,
    I_cmovne,I_cmovle,
    INVALID,INVALID
};

static NUMBER TabEqLe[] =
{
    I_beq,I_ble,
    I_bne,I_bgt,
    I_cmoveq,I_cmovle,
    I_cmovne,I_cmovgt,
    INVALID,INVALID
};

static NUMBER TabEqGe[] =
{
    I_beq,I_bge,
    I_bne,I_blt,
    I_cmoveq,I_cmovge,
    I_cmovne,I_cmovlt,
    INVALID,INVALID
};



/*=======================================================
@FUN
@END
=======================================================*/
static BOOL AinsCanTransformAllUses(INDEX iins_the,NUMBER *OpOneTrans, REG the_reg)
{
    INDEX reg;
    INDEX jins;
    BOOL  def;

    def = 0;
    
    reg = AINS_REG_C(iins_the);

    for(jins=AINS_NEXT(iins_the); VALID(jins); jins = AINS_NEXT(jins))
    {
        if( AINS_REG_B(jins) == reg )
                return 0;
        if( AINS_REG_A(jins) == reg )
        {
            if(!VALID(OpcodeTranslate(AINS_CODE(jins),OpOneTrans)) )
                return 0;
            if( def )
                return 0;
        }
        
        if( AINS_REG_C(jins) == reg )
        {
            if( AINS_TYPE(jins) == IT_ICM || AINS_TYPE(jins) == IT_FCM ) return 0;
            else return 1;
        }

        if( AINS_REG_C(jins) == the_reg )
            def = 1;
        
    }
    
    if( ABBL_OUT(AINS_BBL(iins_the)) & REG2BIT(reg) )
        return 0;
    return 1;
}

/*=======================================================
@OPT
Optimize compare (cmpxx) instrutions and instructions that are used
as compare instruction (subq,xor).

Example:
            
@END
=======================================================*/
static NUMBER AinsTransform2(INDEX iins_the,NUMBER *OpOneTrans, REG the_reg )
{
    INDEX reg;
    INDEX jins;
    NUMBER newop;
    NUMBER n = 0;
    BOOL  def;

    def = 0;
    
    reg = AINS_REG_C(iins_the);

    for(jins=AINS_NEXT(iins_the); VALID(jins); jins = AINS_NEXT(jins))
    {
        if( AINS_REG_A(jins) == reg )
        {
            newop = OpcodeTranslate(AINS_CODE(jins),OpOneTrans);
            if( VALID(newop) && !def )
            {
                AINS_CODE(jins) = newop;
                AINS_REG_A(jins) = the_reg;
                n++;
            }
        }
        if( AINS_REG_C(jins) == reg )
            break;
        
        if( AINS_REG_C(jins) == the_reg )
            def = 1;
    }

    return n;
}
/*=======================================================
@OPT
Optimize compare instrutions with zero and their subsequent uses
Example:
            
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptCompare"

extern NUMBER OptCompare(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Compare");
    BOOL change;
    BOOL kill;
    NUMBER countc=0;
    NUMBER countch=0;
    NUMBER countk=0;
    NUMBER countkh=0;
    
    NUMBER n;
    
    INDEX ifun;
    INDEX ibbl;
    INDEX iins;
    
    FOREACH_LIVE_FUN(ifun)
    FOREACH_FUN_BBL(ifun,ibbl)
    FOREACH_BBL_INS(ibbl,iins)
        {

            if( ( AINS_FLAGS(iins) & FI_EVAL_A) && (AINS_FLAGS(iins) & FI_EVAL_B) )
                continue;

            if( AINS_REG_A(iins) == AINS_REG_B(iins) )
            {
                switch(AINS_CODE(iins))
                {
                case I_cmpeq:
                case I_cmple:
                case I_cmpule:
                    AINS_FLAGS(iins) |= FI_EVAL_C;
                    AINS_C_EVAL(iins) = 1;
                    countc++;
                    if( ABBL_IS_HOT(ibbl) )
                        countch++;
                    if( verbose )
                        fprintf(fp,"\t[*] Evaluated compare with identical operands %d %d\n",iins,ibbl);
                    break;
                case I_cmpult:
                case I_cmplt:
                    AINS_FLAGS(iins) |= FI_EVAL_C;
                    AINS_C_EVAL(iins) = 0;
                    countc++;
                    if( ABBL_IS_HOT(ibbl) )
                        countch++;
                    if( verbose )
                        fprintf(fp,"\t[*] Evaluated compare with identical operands %d %d\n",iins,ibbl);
                    break;
                }
            }

            
            if( AINS_CODE(iins)==I_cmplt && (AINS_FLAGS(iins) & FI_EVAL_B) && AINS_B_EVAL(iins)==1 )
            {
                AINS_REG_B(iins) = REG_ZERO;
                AINS_B_EVAL(iins) = 0;
                AINS_CODE(iins) = I_cmple;
                countc++;
                if( ABBL_IS_HOT(ibbl) )
                    countch++;
                if( verbose )
                    fprintf(fp,"\t[*] Simplified compare with one %d %d\n",iins,ibbl);
                break;
            }

            if( AINS_CODE(iins)==I_cmpult && (AINS_FLAGS(iins) & FI_EVAL_B) && AINS_B_EVAL(iins)==1 )
            {
                AINS_REG_B(iins) = REG_ZERO;
                AINS_B_EVAL(iins) = 0;
                AINS_CODE(iins) = I_cmpule;
                countc++;
                if( ABBL_IS_HOT(ibbl) )
                    countch++;
                if( verbose )
                    fprintf(fp,"\t[*] Simplified compare with one %d %d\n",iins,ibbl);
                break;
            }

            if( AINS_CODE(iins)==I_cmpeq && AINS_REG_A(iins)==REG_ZERO ) 
            {
                kill =  AinsCanTransformAllUses(iins,TabEqNe,AINS_REG_B(iins));
                if( AINS_REG_B(iins) == AINS_REG_C(iins) && !kill ) continue;

                if( verbose )
                        PrintBblSuper(stdout,ibbl);
                                    
                n = AinsTransform2(iins,TabEqNe,AINS_REG_B(iins));
                if( kill )
                {
                    countk++;
                    if( ABBL_IS_HOT(ibbl) )
                    countkh++;
                    AinsKill(iins);
                }
                countc+=n;
                if( ABBL_IS_HOT(ibbl) )
                    countch+=n;

                if( verbose )
                    PrintBblSuper(stdout,ibbl);
            }

#if 1
            if( AINS_CODE(iins)==I_lda && AINS_DATA(iins) < 0 && AINS_DATA(iins) >= -255 )
            {
                if( AinsCanTransformAllUses(iins,TabEqNe,AINS_REG_C(iins)) )
                {
                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                    
                    AinsMakeCmpeq(iins,AINS_REG_B(iins),REG_LIT, AINS_REG_C(iins),-AINS_DATA(iins));
                    n=AinsTransform2(iins,TabEqNe,AINS_REG_C(iins));
                    countc+=n;
                    if( ABBL_IS_HOT(ibbl) )
                        countch+=n;

                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                }
            }
                        
            if( AINS_CODE(iins)==I_subq )
            {
                if( AinsCanTransformAllUses(iins,TabEqNe,AINS_REG_C(iins)) )
                {
                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                                        
                    AinsMakeCmpeq(iins,AINS_REG_A(iins),AINS_REG_B(iins), AINS_REG_C(iins),AINS_DATA(iins));
                    n=AinsTransform2(iins,TabEqNe,AINS_REG_C(iins));
                    countc+=n;
                    if( ABBL_IS_HOT(ibbl) )
                        countch+=n;

                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                }
            }

            if( AINS_CODE(iins)==I_xor )
            {
                if( AinsCanTransformAllUses(iins,TabEqNe,AINS_REG_C(iins)) )
                {
                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                    AinsMakeCmpeq(iins,AINS_REG_A(iins),AINS_REG_B(iins), AINS_REG_C(iins),AINS_DATA(iins));
                    n=AinsTransform2(iins,TabEqNe,AINS_REG_C(iins));
                    countc+=n;
                    if( ABBL_IS_HOT(ibbl) )
                        countch+=n;
                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                }
            }
#endif
            
            if( AINS_CODE(iins)==I_cmpeq && AINS_REG_B(iins)==REG_ZERO )
            {
                kill = AinsCanTransformAllUses(iins,TabEqNe,AINS_REG_A(iins));
                if( AINS_REG_A(iins) == AINS_REG_C(iins) && !kill ) continue;

                if( verbose )
                    PrintBblSuper(stdout,ibbl);
                                    
                n=AinsTransform2(iins,TabEqNe,AINS_REG_A(iins));
                if( kill )
                {
                    countk++;
                    if( ABBL_IS_HOT(ibbl) )
                    countkh++;
                    AinsKill(iins);
                }
                countc+=n;
                if( ABBL_IS_HOT(ibbl) )
                    countch+=n;

                if( verbose )
                    PrintBblSuper(stdout,ibbl);
            }
            
            if( AINS_CODE(iins)==I_cmple && AINS_REG_A(iins)==REG_ZERO )
            {
                kill = AinsCanTransformAllUses(iins,TabEqLt,AINS_REG_B(iins));
                if( AINS_REG_B(iins) == AINS_REG_C(iins) && !kill ) continue;

                if( verbose )
                    PrintBblSuper(stdout,ibbl);

                n=AinsTransform2(iins,TabEqLt,AINS_REG_B(iins));
                if( kill )
                {
                    countk++;
                    if( ABBL_IS_HOT(ibbl) )
                    countkh++;
                    AinsKill(iins);
                }
                countc+=n;
                if( ABBL_IS_HOT(ibbl) )
                    countch+=n;

                if( verbose )
                    PrintBblSuper(stdout,ibbl);
            }
            

            if( AINS_CODE(iins)==I_cmple && AINS_REG_B(iins)==REG_ZERO )
            {
                kill =  AinsCanTransformAllUses(iins,TabEqGt,AINS_REG_A(iins));
                if( AINS_REG_A(iins) == AINS_REG_C(iins) && !kill ) continue;

                if( verbose )
                    PrintBblSuper(stdout,ibbl);
                n=AinsTransform2(iins,TabEqGt,AINS_REG_A(iins));
                if( kill )
                {
                    countk++;
                    if( ABBL_IS_HOT(ibbl) )
                    countkh++;
                    AinsKill(iins);
                }

                countc+=n;
                if( ABBL_IS_HOT(ibbl) )
                    countch+=n;
                if( verbose )
                    PrintBblSuper(stdout,ibbl);
            }


            if( AINS_CODE(iins)==I_cmplt && AINS_REG_A(iins)==REG_ZERO )
            {
                kill =  AinsCanTransformAllUses(iins,TabEqLe,AINS_REG_B(iins));
                if( AINS_REG_B(iins) != AINS_REG_C(iins) || kill )
                {
                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                    n=AinsTransform2(iins,TabEqLe,AINS_REG_B(iins));
                    if( kill )
                    {
                        countk++;
                        if( ABBL_IS_HOT(ibbl) )
                            countkh++;
                        AinsKill(iins);
                    }

                    countc+=n;
                    if( ABBL_IS_HOT(ibbl) )
                        countch+=n;
                    if( verbose )
                        PrintBblSuper(stdout,ibbl);
                }
            }
            
            if( AINS_CODE(iins)==I_cmplt && AINS_REG_B(iins)==REG_ZERO )
            {
                kill = AinsCanTransformAllUses(iins,TabEqGe,AINS_REG_A(iins));
                if( AINS_REG_A(iins) == AINS_REG_C(iins) && !kill ) continue;
                if( verbose )
                    PrintBblSuper(stdout,ibbl);
                n=AinsTransform2(iins,TabEqGe,AINS_REG_A(iins));
                if( kill )
                {
                    countk++;
                    if( ABBL_IS_HOT(ibbl) )
                    countkh++;
                    AinsKill(iins);
                }

                countc+=n;
                if( ABBL_IS_HOT(ibbl) )
                    countch+=n;
                if( verbose )
                    PrintBblSuper(stdout,ibbl);
            }
            
            if( AINS_CODE(iins)==I_cmpult && AINS_REG_A(iins)==REG_ZERO )
            {
                kill = AinsCanTransformAllUses(iins,TabEqEq,AINS_REG_B(iins));
                if( AINS_REG_B(iins) == AINS_REG_C(iins) && !kill ) continue;
                if( verbose )
                    PrintBblSuper(stdout,ibbl);
                n = AinsTransform2(iins,TabEqEq,AINS_REG_B(iins));
                if( kill )
                {
                    countk++;
                    if( ABBL_IS_HOT(ibbl) )
                    countkh++;
                    AinsKill(iins);
                }

                countc+=n;
                if( ABBL_IS_HOT(ibbl) )
                    countch+=n;
                if( verbose )
                    PrintBblSuper(stdout,ibbl);
            }
            
            if( AINS_CODE(iins)==I_cmpult && AINS_REG_B(iins)==REG_ZERO )
            {
                AINS_FLAGS(iins) |= FI_EVAL_C;
                AINS_C_EVAL(iins) = 0;

                countc++;
                if( ABBL_IS_HOT(ibbl) )
                    countch++;
                break;
            }


            if( AINS_CODE(iins)==I_cmpule && AINS_REG_A(iins)==REG_ZERO )
            {
                AINS_FLAGS(iins) |= FI_EVAL_C;
                AINS_C_EVAL(iins) = 1;

                countc++;
                if( ABBL_IS_HOT(ibbl) )
                    countch++;
                break;
            }

            /* same as I_cmpeq */
            
            if( AINS_CODE(iins)==I_cmpule && AINS_REG_B(iins)==REG_ZERO )
            {
                kill = AinsCanTransformAllUses(iins,TabEqNe,AINS_REG_A(iins));
                if( AINS_REG_A(iins) == AINS_REG_C(iins) && !kill ) continue;
                if( verbose )
                {
                    PrintIns(stdout,iins);
                    PrintBblSuper(stdout,ibbl);
                }
                
                n = AinsTransform2(iins,TabEqNe,AINS_REG_A(iins));
                if( kill )
                {
                    countk++;
                    if( ABBL_IS_HOT(ibbl) )
                    countkh++;
                    AinsKill(iins);
                }
                countc+=n;
                if( ABBL_IS_HOT(ibbl) )
                    countch+=n;
                if( verbose )
                    PrintBblSuper(stdout,ibbl);
            }
        }
    
    change = (countc+countk)>0;

    InfoStatus(fp,"\t[OptCompare%c] changed: %d#%d killed: %d#%d\n",CHANGE(change),countc,countch,countk,countkh);
        
    return change;
}
