/*=======================================================
  Author:   R. Muth
  Version:  $Id: schedule.c,v 1.29 1999/02/01 18:13:30 muth Exp $
  Comment:  scheduling code
                
=======================================================*/

#include "global.h"
#include "flags.h"
#include "functions.h"
#include "opcode.h"
#include "regs.h"


/*=======================================================
Data structures
=======================================================*/

#define SCHLOT_INSTR(is) aSchlot[is].Instr
#define SCHLOT_ISSUE_CYCLE(is) aSchlot[is].IssueCycle

#define CANDIDATE_SCHEDULEE(ic) aCandidate[ic].Schedulee
#define CANDIDATE_TIME_TO_END(ic) aCandidate[ic].TimeToEnd
#define CANDIDATE_EARLIEST_ISSUE_CYCLE(ic) aCandidate[ic].EarliestIssueCycle
#define CANDIDATE_BONUS_MALUS_HIGH(ic) aCandidate[ic].BonusMalusHigh
#define CANDIDATE_BONUS_MALUS_LOW(ic) aCandidate[ic].BonusMalusLow


#define FOREACH_SCHEDULEE_SUCC_CONSTRAINT(ib, is)  for(is=SCHEDULEE_SUCC(ib); VALID(is); is=CONSTRAINT_SUCC(is))
#define FOREACH_SCHEDULEE_PRED_CONSTRAINT(ib, ip)  for(ip=SCHEDULEE_PRED(ib); VALID(ip); ip=CONSTRAINT_PRED(ip))

#define FOREACH_SCHEDULEE_AFTER(is,js) for(js=is+1; js<SCHEDULEES_USED;js++)
     
#define FOREACH_SCHEDULEE(is) for(is=0; is < SCHEDULEES_USED; is++)

#define  FOREACH_CONSTRAINT(ic) for(ic=0; ic < CONSTRAINTS_USED; ic++)

#define    CONSTRAINT_PRED(ic) aConstraint[ic].Pred
#define    CONSTRAINT_SUCC(ic) aConstraint[ic].Succ
#define    CONSTRAINT_HEAD(ic) aConstraint[ic].Head
#define    CONSTRAINT_TAIL(ic) aConstraint[ic].Tail
#define    CONSTRAINT_TYPE(ic) aConstraint[ic].Type
#define    CONSTRAINT_FLAGS(ic) aConstraint[ic].Flags
#define    CONSTRAINT_LATENCY_AV(ic) aConstraint[ic].LatencyAv
#define    CONSTRAINT_LATENCY_HI(ic) aConstraint[ic].LatencyHi
#define    CONSTRAINT_REGRES(ic)  aConstraint[ic].Regres


#define    SCHEDULEE_FLAGS(is) aSchedulee[is].Flags
#define    SCHEDULEE_PRED(is) aSchedulee[is].Pred
#define    SCHEDULEE_SUCC(is) aSchedulee[is].Succ
#define    SCHEDULEE_INSTR(is) aSchedulee[is].Instr
#define    SCHEDULEE_TIME_TO_END(is) aSchedulee[is].TimeToEnd
#define    SCHEDULEE_LIVE_REGS(is) aSchedulee[is].LiveRegs
#define    SCHEDULEE_IS_SCHEDULED(is) (VALID(aSchedulee[is].IssueCycle))

#define    SCHEDULEE_ISSUE_CYCLE(is) aSchedulee[is].IssueCycle

#define FOREACH_CANDIDATE(ic)  for(ic=0; ic < CANDIDATES_USED; ic++)

#define FOREACH_SCHLOT(is)  for(is=0; is < SCHLOTS_USED; is++)

#define FOREACH_REG_IN_REGLIST(reg,reglist)   \
                                       for( reg = 0; reg <= 63; reg++)       \
                                       if( !((1L<<reg) & reglist) ) \
                                       continue; else

#define FOREACH_RES_IN_RESLIST(res,reslist)   \
                                       for( res = 0; res <= 63; res++)       \
                                       if( !((1L<<res) & reslist) ) \
                                       continue; else


/*=======================================================
  Local globals
=======================================================*/

static struct 
{
    INDEX IntMulBusyTime;
    INDEX FltDivBusyTime;
}FunctionalUnit;

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "CandidateNew"

static INDEX CandidateNew()
{
    CFATAL(CANDIDATES_USED >= CANDIDATES_ALLOCATED,("not enough edges\n"));
    return(CANDIDATES_USED++);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "CandidateNewFancy"

static INDEX CandidateNewFancy(INDEX ischedulee)
{
    INDEX icandidate = CandidateNew();
    CANDIDATE_SCHEDULEE(icandidate) = ischedulee;
    return icandidate;
}
/*=======================================================
@FUN
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintCandidate"

static void PrintCandidate(FILE *fp,INDEX icandidate)
{
    fprintf(fp,"%5d  ins  %5d  eic %d  mh %d  tte %d  ml %d\n",
            icandidate,
            SCHEDULEE_INSTR(CANDIDATE_SCHEDULEE(icandidate)),
            CANDIDATE_EARLIEST_ISSUE_CYCLE(icandidate),
            CANDIDATE_BONUS_MALUS_HIGH(icandidate),
            CANDIDATE_TIME_TO_END(icandidate),
            CANDIDATE_BONUS_MALUS_LOW(icandidate));
    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ScheduleeNew"

static INDEX ScheduleeNew()
{
    CFATAL(SCHEDULEES_USED >= SCHEDULEES_ALLOCATED,("not enough edges\n"));
    return(SCHEDULEES_USED++);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ScheduleeNewFancy"

extern INDEX ScheduleeNewFancy(INDEX iins, LATENCY latency, REGLIST liveregs)
{
    INDEX ischedulee = ScheduleeNew();

    SCHEDULEE_INSTR(ischedulee) = iins;
    SCHEDULEE_ISSUE_CYCLE(ischedulee) = latency;
    SCHEDULEE_PRED(ischedulee) = INVALID;
    SCHEDULEE_SUCC(ischedulee)  = INVALID;
    SCHEDULEE_TIME_TO_END(ischedulee) = INVALID;
    SCHEDULEE_LIVE_REGS(ischedulee) = liveregs;
    
    return ischedulee;
}
/*=======================================================
@FUN
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintSchedulee"

static void PrintSchedulee(FILE *fp,INDEX ischedulee)
{
    fprintf(fp,"%5d  ins  %5d  ic %d  tte  %d\n",
            ischedulee,
            SCHEDULEE_INSTR(ischedulee),
            SCHEDULEE_ISSUE_CYCLE(ischedulee),
            SCHEDULEE_TIME_TO_END(ischedulee));
    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "SchlotNew"

static INDEX SchlotNew()
{
    CFATAL(SCHLOTS_USED >= SCHLOTS_ALLOCATED,("not enough schlots\n"));
    return(SCHLOTS_USED++);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ConstraintNew"

static INDEX ConstraintNew()
{
    CFATAL(CONSTRAINTS_USED >= CONSTRAINTS_ALLOCATED,("not enough constraints\n"));
    return(CONSTRAINTS_USED++);
}

/*=======================================================
@FUN
Attach an  edge to its nodes. 
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ConstraintNewFancy"

static INDEX ConstraintNewFancy(INDEX head, INDEX tail, INT8 type, INT16 latency_av, INT16 latency_hi, INT8 regres )
{
    INDEX iconstraint;

    iconstraint = ConstraintNew();
    CONSTRAINT_HEAD(iconstraint) = head;
    CONSTRAINT_TAIL(iconstraint) = tail;
    CONSTRAINT_TYPE(iconstraint) = type;

    if( latency_av > latency_hi )
    {
        ERROR(("av latency is bigger than hi latency\n"));
        PrintIns(stdout,SCHEDULEE_INSTR(head));
        PrintIns(stdout,SCHEDULEE_INSTR(tail));
    }
    
    CONSTRAINT_LATENCY_AV(iconstraint) = latency_av;
    CONSTRAINT_LATENCY_HI(iconstraint) = latency_hi;


    CONSTRAINT_REGRES(iconstraint) = regres;
        
    
    CONSTRAINT_SUCC(iconstraint) = SCHEDULEE_SUCC( head );
    SCHEDULEE_SUCC( head ) = iconstraint;
    
    CONSTRAINT_PRED(iconstraint) = SCHEDULEE_PRED( tail );
    SCHEDULEE_PRED( tail ) = iconstraint;

    return iconstraint;
}

/*=======================================================
@FUN
Attach an  edge to its nodes. 
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintConstraint"

static char *ConstraintType[] =
{
    "BAD BAD BAD",
    "REG_PROD_CONS",
    "REG_PROD_PROD",
    "REG_CONS_PROD",
    "RES_PROD_CONS",
    "RES_PROD_PROD",
    "RES_CONS_PROD",
    "MEMORY       ",
    "HINT         "
};


static void PrintConstraint(FILE *fp,INDEX iconstraint)
{
    fprintf(fp,"%5d  %5d->%5d %s %d  %d,%d\n",
            iconstraint,
            SCHEDULEE_INSTR(CONSTRAINT_HEAD(iconstraint)),
            SCHEDULEE_INSTR(CONSTRAINT_TAIL(iconstraint)),
            ConstraintType[CONSTRAINT_TYPE(iconstraint)],
            CONSTRAINT_REGRES(iconstraint),
            CONSTRAINT_LATENCY_AV(iconstraint),
            CONSTRAINT_LATENCY_HI(iconstraint));
}

/*=======================================================
@FUN

@END
=======================================================*/
#if 0
#undef __PROC__
#define __PROC__ "ConstraintIsOnlyLink"

static BOOL ConstraintIsOnlyLink(INDEX iconstraint)
{
    INDEX xschedulee;
    INDEX xconstraint;
    BOOL  flag;
    
    FOREACH_SCHEDULEE(xschedulee)
    {
        SCHEDULEE_FLAGS(xschedulee) = 0;
    }

    SCHEDULEE_FLAGS(CONSTRAINT_HEAD(iconstraint)) = 1;

    do
    {
        flag = 0;
        
        FOREACH_CONSTRAINT(xconstraint)
        {
            if( CONSTRAINT_TYPE(xconstraint) == CT_REG_CONS_PROD) continue;
            if( CONSTRAINT_TYPE(xconstraint) == CT_REG_PROD_PROD) continue;
            
            if( SCHEDULEE_FLAGS(CONSTRAINT_HEAD(xconstraint)) )
            {
                if( !SCHEDULEE_FLAGS(CONSTRAINT_TAIL(xconstraint)) )
                {
                    flag = 1;
                    SCHEDULEE_FLAGS(CONSTRAINT_TAIL(xconstraint)) = 1;
                    if( CONSTRAINT_TAIL(xconstraint) == CONSTRAINT_TAIL(iconstraint) )
                        return 0;
                }
                
            }
        }
    }
    while( flag );
    
    return 1;
    
}
#endif
/*=======================================================
@FUN
This is a heuristic!
This is routine is optimized and hence a little hard to understand
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "CountAndMarkRelevantConsProdAndProdProdConstraints"

extern NUMBER CountAndMarkRelevantConsProdAndProdProdConstraints()
{
    INDEX ischedulee;
    INDEX jschedulee;
    INDEX iconstraint;
    INDEX relevant = 0;
    BOOL  flag;

    FOREACH_CONSTRAINT(iconstraint)
    {
        CONSTRAINT_FLAGS(iconstraint) &= ~F_MARKED;
    }
    
                    
    FOREACH_SCHEDULEE(ischedulee)
    {
        FOREACH_SCHEDULEE(jschedulee)
        {
            SCHEDULEE_FLAGS(jschedulee) &= ~F_MARKED;
        }

        SCHEDULEE_FLAGS(ischedulee) |= F_MARKED;

        /* and propagate the mark over all edges which are not of type CONS_PROD or PROD_PROD */
          
        do
        {
            flag = 0;

            FOREACH_SCHEDULEE(jschedulee)
            {
                if( SCHEDULEE_FLAGS(jschedulee) & F_MARKED )
                {
                    FOREACH_SCHEDULEE_SUCC_CONSTRAINT(jschedulee,iconstraint)
                    {
                        if( CONSTRAINT_TYPE(iconstraint) == CT_REG_CONS_PROD) continue;
                        if( CONSTRAINT_TYPE(iconstraint) == CT_REG_PROD_PROD) continue;
                        if( SCHEDULEE_FLAGS(CONSTRAINT_TAIL(iconstraint)) & F_MARKED ) continue;
                        flag = 1;
                        SCHEDULEE_FLAGS(CONSTRAINT_TAIL(iconstraint)) |= F_MARKED;
                    }
                }
            }
        }
        while( flag );

        FOREACH_SCHEDULEE_SUCC_CONSTRAINT(ischedulee,iconstraint)
        {
            if( CONSTRAINT_TYPE(iconstraint) != CT_REG_CONS_PROD &&
                CONSTRAINT_TYPE(iconstraint) != CT_REG_PROD_PROD) continue;
            
            if( SCHEDULEE_FLAGS(CONSTRAINT_TAIL(iconstraint)) & F_MARKED ) continue;
            relevant++;
            
            CONSTRAINT_FLAGS(iconstraint) |= F_MARKED;
        }
    }

    return relevant;    
}

/*=======================================================
@FUN

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ScheduleesMarkInstrWithRelavantConsProdAndProdProdEdges"

extern void ScheduleesMarkInstrWithRelavantConsProdAndProdProdEdges()
{
    INDEX ischedulee;
    INDEX iconstraint;
    
    FOREACH_SCHEDULEE(ischedulee)
    {
        AINS_FLAGS(SCHEDULEE_INSTR(ischedulee)) &= ~F_MARKED;
    }
    
    FOREACH_CONSTRAINT(iconstraint)
    {
        if(CONSTRAINT_FLAGS(iconstraint) & F_MARKED)
        {
            ischedulee = CONSTRAINT_HEAD(iconstraint);
            AINS_FLAGS(SCHEDULEE_INSTR(ischedulee)) |= F_MARKED;
            ischedulee = CONSTRAINT_TAIL(iconstraint);
            AINS_FLAGS(SCHEDULEE_INSTR(ischedulee)) |= F_MARKED;        
        }
    }
}


/*=======================================================
@FUN

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "CandidatesSort"

static int cmp(const void *x, const void *y)
{
    CANDIDATE *c = (CANDIDATE *)x;
    CANDIDATE *d = (CANDIDATE *)y;

    /* smaller issue cycle first */
    
    if( c->EarliestIssueCycle != d->EarliestIssueCycle )
        return c->EarliestIssueCycle - d->EarliestIssueCycle;

    /* higher high bonus malus first */
    
    if( d->BonusMalusHigh != c->BonusMalusHigh )
        return d->BonusMalusHigh - c->BonusMalusHigh;

    /* higher time to end first */

    if( d->TimeToEnd != c->TimeToEnd )
        return d->TimeToEnd - c->TimeToEnd;
    
    /* higher low bonus malus first */
    
    if( d->BonusMalusLow != c->BonusMalusLow )
        return d->BonusMalusLow - c->BonusMalusLow;
    
    return 0;
}


static void CandidatesSort()
{
    qsort(aCandidate,CANDIDATES_USED,sizeof(CANDIDATE),cmp);
}

            
/*=======================================================
@FUN

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "CandidatesFind"

static void CandidatesFind()
{
    
    INDEX ischedulee;
    INDEX iconstraint;
    INDEX jschedulee;

    CANDIDATES_USED = 0;
    FOREACH_SCHEDULEE(ischedulee)
    {
        if( VALID(SCHEDULEE_ISSUE_CYCLE(ischedulee)) ) continue;

        FOREACH_SCHEDULEE_PRED_CONSTRAINT(ischedulee,iconstraint)
        {
            jschedulee = CONSTRAINT_HEAD(iconstraint);
            if( VALID(SCHEDULEE_ISSUE_CYCLE(jschedulee)) ) continue;
            break;
        }

        if( VALID(iconstraint) ) continue;

        CandidateNewFancy(ischedulee);
    }
}


/*=======================================================
@FUN

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ScheduleesComputeTimeToEnd"

static void ScheduleesComputeTimeToEnd()
{
    BOOL cont;
    INDEX ischedulee;
    INDEX iconstraint;
    INDEX jschedulee;
    LATENCY n;
    LATENCY time_to_end;
    
    /* latency to already scheduled instructions */

    FOREACH_SCHEDULEE(ischedulee)
    {
        SCHEDULEE_TIME_TO_END(ischedulee)  = INVALID;
    }

    
    do
    {
        cont=0;
        FOREACH_SCHEDULEE(ischedulee)
        {
            if( VALID(SCHEDULEE_TIME_TO_END(ischedulee)) ) continue;

            time_to_end = 0;  /* this takes care of the no successor case */

            FOREACH_SCHEDULEE_SUCC_CONSTRAINT(ischedulee,iconstraint)
            {
                jschedulee = CONSTRAINT_TAIL(iconstraint);
                if( VALID(SCHEDULEE_TIME_TO_END(jschedulee)) )
                {
                    n = SCHEDULEE_TIME_TO_END(jschedulee) + CONSTRAINT_LATENCY_AV(iconstraint);
                    if( n > time_to_end ) time_to_end = n;
                }
                else
                {
                    time_to_end = INVALID;
                    break;
                }
            }

            if( VALID(time_to_end) )
            {
                SCHEDULEE_TIME_TO_END(ischedulee) = time_to_end;
            }
            else
            {
                cont = 1 ;
            }
        }
    }
    while( cont );
}


/*=======================================================
@FUN

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ProdProdLatencyAv"

static LATENCY ProdProdLatencyAv(INDEX iins, INDEX jins)
{
    return AINS_LATENCY_AV(iins) - AINS_LATENCY_AV(jins) + 1;
}

/*=======================================================
@FUN

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ProdProdLatencyHi"

static LATENCY ProdProdLatencyHi(INDEX iins, INDEX jins)
{
    if( AINS_TYPE(iins) == IT_ULD || AINS_TYPE(iins) == IT_ILD || AINS_TYPE(iins) == IT_FLD )
    {
        if( AINS_REG_B(iins) == REG_SP || AINS_REG_B(iins) == REG_GP )
            return AINS_LATENCY_AV(iins) - AINS_LATENCY_AV(jins) + 1;
    }

    return AINS_LATENCY_HI(iins) - AINS_LATENCY_AV(jins) + 1;
}

/*=======================================================
@FUN

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ProdConsLatencyAv"

static LATENCY ProdConsLatencyAv(INDEX iins, INDEX jins)
{
    if( (AINS_TYPE(iins) == IT_LOG || AINS_TYPE(iins) == IT_ICP) && (AINS_TYPE(jins) == IT_IBR || AINS_TYPE(jins) == IT_ICM) )
        return 0;

    return AINS_LATENCY_AV(iins);
}


/*=======================================================
@FUN

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ProdConsLatencyHi"

static LATENCY ProdConsLatencyHi(INDEX iins, INDEX jins)
{
    if( (AINS_TYPE(iins) == IT_LOG || AINS_TYPE(iins) == IT_ICP) && (AINS_TYPE(jins) == IT_IBR || AINS_TYPE(jins) == IT_ICM) )
        return 0;

    if( AINS_TYPE(iins) == IT_ULD || AINS_TYPE(iins) == IT_ILD || AINS_TYPE(iins) == IT_FLD )
    {
        if( AINS_REG_B(iins) == REG_SP || AINS_REG_B(iins) == REG_GP )
            return AINS_LATENCY_AV(iins);
    }
    
    return AINS_LATENCY_HI(iins);
}
/*=======================================================
@FUN
Returns the number of incoming edges into ibbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ScheduleeNpreds"

extern NUMBER ScheduleeNpreds(INDEX ischedulee)
{
    INDEX  iconstraint;
    NUMBER i=0;
    FOREACH_SCHEDULEE_PRED_CONSTRAINT(ischedulee,iconstraint) i++;
    return i;
}

/*=======================================================
@FUN
Returns the number of outgoing edges from ibbl
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ScheduleeNsuccs"

extern NUMBER ScheduleeNsuccs(INDEX ischedulee)
{
    INDEX  iconstraint;
    NUMBER i=0;
    FOREACH_SCHEDULEE_SUCC_CONSTRAINT(ischedulee,iconstraint) i++;
    return i;
}

/*=======================================================
@FUN
This function contains most of the scheduling heuristic/logic

@END
=======================================================*/

#undef __PROC__
#define __PROC__ "CandidatesNStores"

static NUMBER CandidatesNStores()
{
    INDEX icandidate;
    INDEX iins;
    NUMBER count=0;
    
    FOREACH_CANDIDATE(icandidate)
    {
        iins = SCHEDULEE_INSTR(CANDIDATE_SCHEDULEE(icandidate));
    
        if( AINS_TYPE(iins) == IT_UST ||
            AINS_TYPE(iins) == IT_IST ||
            AINS_TYPE(iins) == IT_CST ||
            AINS_TYPE(iins) == IT_FST )
            count++;
    }

    return count;
}

/*=======================================================
@FUN
This function contains most of the scheduling heuristic/logic
It sets the priority 
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "CandidateSetPriorities"

static void CandidateSetPriorities(INDEX icandidate, LATENCY cycle_av)
{    
    INDEX iconstraint;
    INDEX ischedulee = CANDIDATE_SCHEDULEE(icandidate);
    INDEX iins = SCHEDULEE_INSTR(ischedulee);
    INDEX jschedulee;
    LATENCY n;
    LATENCY cycle_hi = cycle_av;
    
    /* latency to already scheduled instructions */
    
    FOREACH_SCHEDULEE_PRED_CONSTRAINT(ischedulee,iconstraint)
    {
        jschedulee = CONSTRAINT_HEAD(iconstraint);
        CASSERT( SCHEDULEE_IS_SCHEDULED(jschedulee),("weird\n"));
        
        n = SCHEDULEE_ISSUE_CYCLE(jschedulee) + CONSTRAINT_LATENCY_AV(iconstraint);
        if( n > cycle_av ) cycle_av = n;

        n = SCHEDULEE_ISSUE_CYCLE(jschedulee) + CONSTRAINT_LATENCY_HI(iconstraint);
        if( n > cycle_hi ) cycle_hi = n;
    }

    CASSERT( cycle_hi >= cycle_av,("weird\n"));
    
    /* integer multiply */

    /* not implemented */

    /* floating point divide */

    /* not implemented */
    
    CANDIDATE_EARLIEST_ISSUE_CYCLE(icandidate) = cycle_av;



#if 0
    if( cycle_hi > cycle_av )
    {
        CANDIDATE_BONUS_MALUS_HIGH(icandidate) = -(cycle_hi - cycle_av);
    }
    else
    {
#if 0
        CANDIDATE_BONUS_MALUS_HIGH(icandidate) = AINS_LATENCY_HI(iins);
        
        if( !(AINS_PIPELINES(iins) & (AINS_PIPELINES(iins)-1)) )
            CANDIDATE_BONUS_MALUS_HIGH(icandidate) ++;
        if( AINS_TYPE(iins) == IT_UST ||
            AINS_TYPE(iins) == IT_IST ||
            AINS_TYPE(iins) == IT_CST ||
            AINS_TYPE(iins) == IT_FST )
            CANDIDATE_BONUS_MALUS_HIGH(icandidate) += 2;
#endif

#if 1
        CANDIDATE_BONUS_MALUS_HIGH(icandidate) = AINS_LATENCY_HI(iins);
        if( AINS_TYPE(iins) == IT_UST ||
            AINS_TYPE(iins) == IT_IST ||
            AINS_TYPE(iins) == IT_CST ||
            AINS_TYPE(iins) == IT_FST )
            CANDIDATE_BONUS_MALUS_HIGH(icandidate) = 3;
        CANDIDATE_BONUS_MALUS_HIGH(icandidate) += ScheduleeNsuccs(ischedulee);
        
#endif
    }
    
    
    
    CANDIDATE_TIME_TO_END(icandidate) = SCHEDULEE_TIME_TO_END(ischedulee) / 4;

#else

    CANDIDATE_BONUS_MALUS_HIGH(icandidate) = 0;

#if 0
    if( cycle_hi > cycle_av )
    {
        CANDIDATE_BONUS_MALUS_HIGH(icandidate) -= (cycle_hi - cycle_av);
    }
#endif
    
    CANDIDATE_BONUS_MALUS_HIGH(icandidate)  += ScheduleeNsuccs(ischedulee);
    CANDIDATE_BONUS_MALUS_HIGH (icandidate) += SCHEDULEE_TIME_TO_END(ischedulee) ;

    /* stores have often no dependees and hence zero time to end but they represent a bottle neck
       since we can only issue one per cycle. The following code
       is intended to compensate for that
       */
       
    if( AINS_TYPE(iins) == IT_UST ||
        AINS_TYPE(iins) == IT_IST ||
        AINS_TYPE(iins) == IT_CST ||
        AINS_TYPE(iins) == IT_FST )
        CANDIDATE_BONUS_MALUS_HIGH(icandidate) += CandidatesNStores();
            
    CANDIDATE_TIME_TO_END(icandidate) = 0;
#endif
    
    CANDIDATE_BONUS_MALUS_LOW(icandidate) = -AINS_DATA(iins);  /* smallest load offset first */
}

/*=======================================================
@FUN

@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AinsSchedule"

static void AinsSchedule(INDEX iins, LATENCY cycle)
{
    INDEX ischlot = SchlotNew();

    SCHLOT_ISSUE_CYCLE(ischlot) = cycle;
    SCHLOT_INSTR(ischlot) = iins;
    if( AINS_TYPE(iins) == IT_MLL )
    {
        FunctionalUnit.IntMulBusyTime = cycle + 6;
    }
    else if( AINS_TYPE(iins) == IT_MLQ )
    {
        FunctionalUnit.IntMulBusyTime = cycle + 10;
    }
    else if( AINS_TYPE(iins) == IT_MLH )
    {
        FunctionalUnit.IntMulBusyTime = cycle + 10;
    }

    if( VALID(FunctionalUnit.IntMulBusyTime) && FunctionalUnit.IntMulBusyTime < cycle )
        FunctionalUnit.IntMulBusyTime = INVALID;
}

/*=======================================================
@FUN
Create memory constraints needed for a correct schedule
using alias analysis
Requires:
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "ScheduleesCreateMemoryConstraints"

extern void ScheduleesCreateMemoryConstraints()     
{
    INDEX ischedulee;
    INDEX jschedulee;
    INDEX iins;
    INDEX jins;
    NUMBER result;
        
    /* resource constraints */
    
    FOREACH_SCHEDULEE(ischedulee)
    {
        iins = SCHEDULEE_INSTR(ischedulee);
        
        if( AINS_TYPE(iins) == IT_ULD ||
            AINS_TYPE(iins) == IT_ILD ||
            AINS_TYPE(iins) == IT_LLD ||
            AINS_TYPE(iins) == IT_FLD )
            
        {
            FOREACH_SCHEDULEE_AFTER(ischedulee,jschedulee)
            {
                jins = SCHEDULEE_INSTR(jschedulee);
                if( AINS_TYPE(jins) != IT_UST &&
                    AINS_TYPE(jins) != IT_IST &&
                    AINS_TYPE(jins) != IT_CST &&
                    AINS_TYPE(jins) != IT_FST ) continue;
                if( DoMemoryAccessesConflict(iins,jins) == CONFLICT_NO ) continue;
                
                ConstraintNewFancy(ischedulee,jschedulee,CT_MEMORY,0,0,0);
            }
        }

        if( AINS_TYPE(iins) == IT_UST ||
            AINS_TYPE(iins) == IT_IST ||
            AINS_TYPE(iins) == IT_CST ||
            AINS_TYPE(iins) == IT_FST )
        {
            FOREACH_SCHEDULEE_AFTER(ischedulee,jschedulee)
            {
                jins = SCHEDULEE_INSTR(jschedulee);

                if( AINS_TYPE(jins) == IT_ULD ||
                    AINS_TYPE(jins) == IT_ILD ||
                    AINS_TYPE(jins) == IT_LLD ||
                    AINS_TYPE(jins) == IT_FLD )
                {
                    result =  DoMemoryAccessesConflict(iins,jins);
                    if( result == CONFLICT_NO )
                        continue;
                    else if( result == CONFLICT_MATCH || result == CONFLICT_EXACT )
                        ConstraintNewFancy(ischedulee,jschedulee,CT_MEMORY,3,3,0);
                    else   
                        ConstraintNewFancy(ischedulee,jschedulee,CT_MEMORY,0,0,0);
                        
                }
                
                if( AINS_TYPE(jins) == IT_UST ||
                    AINS_TYPE(jins) == IT_IST ||
                    AINS_TYPE(jins) == IT_CST ||
                    AINS_TYPE(jins) == IT_FST )
                {
                    if( DoMemoryAccessesConflict(iins,jins) == CONFLICT_NO ) continue;
                    ConstraintNewFancy(ischedulee,jschedulee,CT_MEMORY,0,0,0);
                }
                
            }
        }
    }
}

/*=======================================================
@FUN
Create the register constraints needed for a correct schedule
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "ScheduleesCreateRegisterConstraints"

extern void ScheduleesCreateRegisterConstraints()     
{
    INDEX ischedulee;
    INDEX jschedulee;
    REGLIST ireg_use;
    REGLIST ireg_def;
    REGLIST maskdef;
    REGLIST maskuse;
    
    FOREACH_SCHEDULEE(ischedulee)
    {
        ireg_use = AinsUse(SCHEDULEE_INSTR(ischedulee));
        ireg_def = AinsDef(SCHEDULEE_INSTR(ischedulee));
        maskdef = 0;
        maskuse = 0;

        FOREACH_SCHEDULEE_AFTER(ischedulee,jschedulee)
        {
            REGLIST jreg_use = AinsUse(SCHEDULEE_INSTR(jschedulee));
            REGLIST jreg_def = AinsDef(SCHEDULEE_INSTR(jschedulee));
            RESLIST reglist;
            REG  reg;

            maskuse |= jreg_use;

            /* use -> def */

            reglist = ireg_use & jreg_def & ~maskdef;
            if( reglist )
                FOREACH_REG_IN_REGLIST(reg,reglist)
                {
                    ConstraintNewFancy(ischedulee,jschedulee,CT_REG_CONS_PROD,0,0,reg);
                }
            
            /* def -> use */

            reglist = ireg_def & jreg_use & ~maskdef;
            if( reglist )
                FOREACH_REG_IN_REGLIST(reg,reglist)
                {
                    ConstraintNewFancy(ischedulee,
                                       jschedulee,
                                       CT_REG_PROD_CONS,
                                       ProdConsLatencyAv(SCHEDULEE_INSTR(ischedulee),SCHEDULEE_INSTR(jschedulee)),
                                       ProdConsLatencyHi(SCHEDULEE_INSTR(ischedulee),SCHEDULEE_INSTR(jschedulee)),
                                       reg);
                }
            
            /* def -> def */

            reglist = ireg_def & jreg_def & ~maskdef & ~maskuse;
            if( reglist )
                FOREACH_REG_IN_REGLIST(reg,reglist)
                {
                    ConstraintNewFancy(ischedulee,
                                       jschedulee,
                                       CT_REG_PROD_PROD,
                                       ProdProdLatencyAv(SCHEDULEE_INSTR(ischedulee),SCHEDULEE_INSTR(jschedulee)),
                                       ProdProdLatencyHi(SCHEDULEE_INSTR(ischedulee),SCHEDULEE_INSTR(jschedulee)),
                                       reg);
                }

            maskdef |= jreg_def;
        }
    }
}


/*=======================================================
@FUN
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "ScheduleesCreateControlConstraints"

extern void ScheduleesCreateControlConstraints()     
{
    INDEX ischedulee;
    INDEX jschedulee;
    RESLIST ires_use;
    RESLIST ires_def;
    REGLIST live;
    
        
    /* resource constraints */
    
    FOREACH_SCHEDULEE(ischedulee)
    {
        ires_use = AINS_RES_USE(SCHEDULEE_INSTR(ischedulee));
        ires_def = AINS_RES_DEF(SCHEDULEE_INSTR(ischedulee));

        FOREACH_SCHEDULEE_AFTER(ischedulee,jschedulee)
        {
            RESLIST jres_def = AINS_RES_DEF(SCHEDULEE_INSTR(jschedulee));

            /* def -> def */

            if( ires_def &  jres_def & RES_PC )
            {
                ConstraintNewFancy(ischedulee,jschedulee,CT_CONTROL,0,0,RES_PC);
                break;
            }

            /* use -> def */
            
            if( ires_use &  jres_def & RES_PC )
            {
                ConstraintNewFancy(ischedulee,jschedulee,CT_CONTROL,0,0,RES_PC);
                break;
            }
        }


        if( !(ires_def & RES_PC) ) continue;

        live = SCHEDULEE_LIVE_REGS(ischedulee);
              
        /* now handle def->use */
        
        FOREACH_SCHEDULEE_AFTER(ischedulee,jschedulee)
        {
            INDEX iins = SCHEDULEE_INSTR(jschedulee);
            RESLIST jres_use = AINS_RES_USE(iins);

            if( !(jres_use & RES_PC) ) continue;
            
            if( AinsMayCauseException(iins) || AinsHasSideEffect(iins) )
                ConstraintNewFancy(ischedulee,jschedulee,CT_CONTROL,0,0,0);
            if( REG2BIT(AINS_REG_C(iins)) & live )
                ConstraintNewFancy(ischedulee,jschedulee,CT_CONTROL,0,0,0);
            
        }
    }
    
}

/*=======================================================
@FUN
Create resource constraints needed for a correct schedule.
A simple memory disambiguater is part of this it can be
enabled by passing a nonzero argument to the function
A simple scheme is also available for control dependencies
which can be disabled as well.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "ScheduleesCreateResourceConstraints"

extern void ScheduleesCreateResourceConstraints(BOOL do_mm, BOOL do_pc)     
{
    INDEX ischedulee;
    INDEX jschedulee;
    RESLIST ires_use;
    RESLIST ires_def;
    RESLIST maskdef;
    RESLIST maskuse;

    /* resource constraints */
    
    FOREACH_SCHEDULEE(ischedulee)
    {
        ires_use = AINS_RES_USE(SCHEDULEE_INSTR(ischedulee));
        ires_def = AINS_RES_DEF(SCHEDULEE_INSTR(ischedulee));
        maskdef = 0;
        maskuse = 0;

        if( !do_mm)
        {
            ires_def &= ~RES_MM;
            ires_use &= ~RES_MM;
        }

        if( !do_pc)
        {
            ires_def &= ~RES_PC;
            ires_use &= ~RES_PC;
        }
        
            
        FOREACH_SCHEDULEE_AFTER(ischedulee,jschedulee)
        {
            RESLIST jres_use = AINS_RES_USE(SCHEDULEE_INSTR(jschedulee));
            RESLIST jres_def = AINS_RES_DEF(SCHEDULEE_INSTR(jschedulee));
            RESLIST res_set;
            RES res;

            maskuse |= jres_use;
                      
            /* use -> def */

            res_set = ires_use & jres_def & ~maskdef;
            if( res_set )
                FOREACH_RES_IN_RESLIST(res,res_set)
                {
                    ConstraintNewFancy(ischedulee,jschedulee,CT_RES_CONS_PROD,0,0,res);
                }
            
            /* def -> use */

            res_set = ires_def & jres_use & ~maskdef;
            if( res_set )
                FOREACH_RES_IN_RESLIST(res,res_set)
                {
                    ConstraintNewFancy(ischedulee,jschedulee,CT_RES_PROD_CONS,0,0,res);
                }
            
            /* def -> def */

            res_set = ires_def & jres_def & ~maskdef & ~maskuse;
            if( res_set )
                FOREACH_RES_IN_RESLIST(res,res_set)
                {
                    ConstraintNewFancy(ischedulee,jschedulee,CT_RES_PROD_PROD,0,0,res);
                }

            maskdef |= jres_def;
        }
    }
    
}


/*=======================================================
@FUN
This is a hack for speed. It works only if the register
constraints were added last.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "ScheduleesRemoveRegisterConstraints"

extern void ScheduleesRemoveRegisterConstraints()
{
    INDEX ischedulee;
    INDEX iconstraint;

    FOREACH_SCHEDULEE(ischedulee)
    {
        FOREACH_SCHEDULEE_SUCC_CONSTRAINT(ischedulee,iconstraint)
        {
            if( CONSTRAINT_TYPE(iconstraint) == CT_REG_CONS_PROD) continue;
            if( CONSTRAINT_TYPE(iconstraint) == CT_REG_PROD_PROD) continue;
            if( CONSTRAINT_TYPE(iconstraint) == CT_REG_PROD_CONS) continue;
                    
            SCHEDULEE_SUCC(ischedulee) = iconstraint;
            break;
        }

        if( !VALID(iconstraint) )
            SCHEDULEE_SUCC(ischedulee) = INVALID;
        

        FOREACH_SCHEDULEE_PRED_CONSTRAINT(ischedulee,iconstraint)
        {
            if( CONSTRAINT_TYPE(iconstraint) == CT_REG_CONS_PROD) continue;
            if( CONSTRAINT_TYPE(iconstraint) == CT_REG_PROD_PROD) continue;
            if( CONSTRAINT_TYPE(iconstraint) == CT_REG_PROD_CONS) continue;
                    
            SCHEDULEE_PRED(ischedulee) = iconstraint;
            break;
        }

        if( !VALID(iconstraint) )
            SCHEDULEE_PRED(ischedulee) = INVALID;

    }
    
}


/*=======================================================
@FUN
Find a schedule for the instructions in the bbl
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "ScheduleesComputeSchedule"

static void ScheduleesComputeSchedule()
{
    NUMBER n;
    INDEX iins;
    INDEX ischlot;
    INDEX ischedulee;
    INDEX icandidate;
    LATENCY cycle;
    BOOL  hasLoad;
    BOOL  hasStore;
    NUMBER nInts;
    NUMBER nFlts;
    UINT8 pipelines_avail;
    UINT8 pipelines;
    
    /* count number of schedulees */

    n=0;
    FOREACH_SCHEDULEE(ischedulee)
    {
        if( SCHEDULEE_IS_SCHEDULED(ischedulee) ) continue;
        n++;
    }

    /* do the scheduling */

    while( n-- )
    {
        cycle = SCHLOT_ISSUE_CYCLE(SCHLOTS_USED-1);

        /*
          check whether we are dealing with a new block of 4 instructions
          also collect information of the currently slotted instructions
         */

        hasLoad = 0;
        hasStore = 0;
        nInts = 0;
        nFlts = 0;
        
        pipelines_avail = PIPE_E0 | PIPE_E1 | PIPE_FA | PIPE_FM;
            
        if( (SCHLOTS_USED & 3) == 0 )
        {
            cycle++;
        }
        else
        {
            for( ischlot = SCHLOTS_USED-1; ischlot >= 0 ; ischlot-- )
            {
                if( SCHLOT_ISSUE_CYCLE(ischlot) < cycle) break;
                iins = SCHLOT_INSTR(ischlot);

                if( AINS_TYPE(iins) == IT_ULD || AINS_TYPE(iins) == IT_ILD || AINS_TYPE(iins) == IT_FLD )
                    hasLoad = 1;

                if( AINS_TYPE(iins) == IT_UST || AINS_TYPE(iins) == IT_IST || AINS_TYPE(iins) == IT_FST )
                    hasStore = 1;

                pipelines =  AINS_PIPELINES(iins);

                if( pipelines & (PIPE_FA|PIPE_FM) )
                {
                    nFlts++;
                }
                else if( pipelines & (PIPE_E0|PIPE_E1) )
                {
                    nInts++;
                }
                else
                {
                    CASSERT(AINS_CODE(iins)==I_unop,("No pipeline for instruction %s\n",AINS_NAME(iins)));
                }

                
                if( pipelines && (pipelines & (pipelines-1)) == 0 )
                {
                    /* just one pipeline is easy */
                    pipelines_avail &= ~pipelines;
                }

                if( nFlts==2 )
                    pipelines_avail  &= ~(PIPE_FA|PIPE_FM);

                if( nInts==2 )
                    pipelines_avail  &= ~(PIPE_E0|PIPE_E1);
            }
        }
        

        CandidatesFind();
#if 0
        if( AINS_BBL(SCHEDULEE_INSTR(CANDIDATE_SCHEDULEE(0))) == 33 )
        {
            printf("%d schedulees left cycle %d\n",n+1,cycle);
        
            FOREACH_CANDIDATE(icandidate)
                {
                    PrintCandidate(stdout,icandidate);
                }
        }
        
#endif
        
        FOREACH_CANDIDATE(icandidate)
        {
            CandidateSetPriorities(icandidate,cycle);
            
            ischedulee = CANDIDATE_SCHEDULEE(icandidate);
            iins = SCHEDULEE_INSTR(ischedulee);

            /* do slotting corrections for schedulees potentially issuing in the same cycle */

            CASSERT(  CANDIDATE_EARLIEST_ISSUE_CYCLE(icandidate) >= cycle,("weird earliest issue cycle\n"));


            if( VALID(FunctionalUnit.IntMulBusyTime) && (AINS_TYPE(iins) == IT_MLH || AINS_TYPE(iins) == IT_MLQ || AINS_TYPE(iins) == IT_MLL) )
            {
                if( FunctionalUnit.IntMulBusyTime > CANDIDATE_EARLIEST_ISSUE_CYCLE(icandidate) )
                    CANDIDATE_EARLIEST_ISSUE_CYCLE(icandidate) = FunctionalUnit.IntMulBusyTime;
            }
            
            if( CANDIDATE_EARLIEST_ISSUE_CYCLE(icandidate) > cycle ) continue;


            if( hasStore && (AINS_TYPE(iins) == IT_ULD || AINS_TYPE(iins) == IT_ILD || AINS_TYPE(iins) == IT_FLD) )
            {
                CANDIDATE_EARLIEST_ISSUE_CYCLE(icandidate)++;
                continue;
            }

            if( hasLoad && (AINS_TYPE(iins) == IT_UST || AINS_TYPE(iins) == IT_IST || AINS_TYPE(iins) == IT_FST) )
            {
                CANDIDATE_EARLIEST_ISSUE_CYCLE(icandidate)++;
                continue;
            }

            if( (AINS_PIPELINES(iins) & pipelines_avail) == 0 )
            {
                CANDIDATE_EARLIEST_ISSUE_CYCLE(icandidate)++;
                continue;
            }
        }

        CandidatesSort();
#if 0
        
        if( AINS_BBL(SCHEDULEE_INSTR(CANDIDATE_SCHEDULEE(0))) == 6522 ) 
        {
            printf("after sorting\n");
        
            FOREACH_CANDIDATE(icandidate)
                {
                    PrintCandidate(stdout,icandidate);
                }
        }
#endif
        /* zero is now our candidate */
        
        ischedulee = CANDIDATE_SCHEDULEE(0);
        iins = SCHEDULEE_INSTR(ischedulee);
        cycle = CANDIDATE_EARLIEST_ISSUE_CYCLE(0);
        SCHEDULEE_ISSUE_CYCLE(ischedulee) = cycle;
        AinsSchedule(iins,cycle);
    }
}

/*=======================================================
@FUN
The scheduler assumes that the bbls of the program have
been placed in a sequences and put a single (dummy) function
which is passed to it as an argument.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "Schedule"

extern void PrintSchedulees(FILE *fp)
{
    INDEX ischedulee;
    INDEX iconstraint;

    FOREACH_SCHEDULEE(ischedulee)
    {
        printf("schedulee: %d (%d)\n",ischedulee, SCHEDULEE_INSTR(ischedulee));
        FOREACH_SCHEDULEE_SUCC_CONSTRAINT(ischedulee,iconstraint)
            PrintConstraint(stdout,iconstraint);
    }
}

/*=======================================================
@FUN
The scheduler assumes that the bbls of the program have
been placed in a sequences and put a single (dummy) function
which is passed to it as an argument.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "Schedule"

/* #define NO_SCHEDULING_ACROSS_BBL 1 */

extern void Schedule(FILE *fp, INDEX ifun)
{
    INDEX ibbl;
    INDEX next;
    
    INDEX ibbl_first;
    INDEX ibbl_last;
    INDEX iins;
    INDEX first_schlot;
    INDEX history_schlot;
    INDEX ischlot;
    INDEX ischedulee;
    NUMBER added_unops = 0;    
    NUMBER nprodprod = 0;
    NUMBER nprodprodh = 0;
    LATENCY cycle;
    
    NUMBER max_schedulees = 0;
    NUMBER max_constraints = 0;

    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Schedule");
    BOOL   bblverbose;
    

    clock_t time = clock();

    SCHLOTS_USED = 0;

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        SCHEDULEES_USED = 0;

        if( ABBL_FLAGS(ibbl) & FB_ALIGNED_4 )
        {
            CWARNING( SCHLOTS_USED & 3,("bad 4 alignment for bbl: %d\n",ibbl) );
        }

        if( ABBL_FLAGS(ibbl) & FB_ALIGNED_2 )
        {
            CWARNING( SCHLOTS_USED & 1,("bad 2 alignment for bbl: %d\n",ibbl));
        }
        
        /* keep a few of the already scheduled instructions (history) for dependency improvement */

        history_schlot = SCHLOTS_USED & ~3;
        first_schlot = SCHLOTS_USED;

        for(ischlot= history_schlot; ischlot < first_schlot; ischlot++)
        {
            cycle =  SCHLOT_ISSUE_CYCLE(ischlot);
            CASSERT( VALID(cycle),("previously issued instruction does not appear to be issued anymore\n"));
                            
            ScheduleeNewFancy(SCHLOT_INSTR(ischlot), cycle, REGS_ALL_BUT_ZERO);
        }

        /* extract schedulees from extended bbl, but add unop first */

        ischlot = first_schlot;
        
        /* add some noops for alignment purposes */
        
        ibbl_first = ibbl;
        ibbl_last = ibbl;

        if( verbose && ABBL_IS_HOT(ibbl_first) )
            bblverbose = 1;
        else
            bblverbose = 0;


        /*
          DO SUPERBLOCK
          Figure which bbls belong to this superblock while extracting instructions  
         */
        
        for(;;)
        {
            next = ABBL_NEXT(ibbl);

            ischlot += AbblNInss(ibbl);

            if( VALID(next) && (ABBL_FLAGS(next) & FB_ALIGNED_4) )
            {
                while( ischlot & 3 )
                {
                    iins = AinsNew();
                    AINS_BBL(iins) = ibbl;
                    AinsMakeUnop(iins);
                    AinsAppend(iins);
                    added_unops++;
                    ischlot++;
                }
            }
            
                
            if( VALID(next)  && (ABBL_FLAGS(next) & FB_ALIGNED_2) )
            {
                while( ischlot & 1 )
                {
                    iins = AinsNew();
                    AINS_BBL(iins) = ibbl;
                    AinsMakeUnop(iins);
                    AinsAppend(iins);
                    added_unops++;
                    ischlot++;
                }
            }
            
            FOREACH_BBL_INS(ibbl,iins)
            {
                REGLIST liveregs;
                INDEX iedg;

                if( AINS_TYPE(iins) == IT_IBR || AINS_TYPE(iins) == IT_FBR )
                {
#ifdef NO_SCHEDULING_ACROSS_BBL
                    liveregs = REGS_ALL_BUT_ZERO;
#else
                    liveregs = REGS_NONE;
#endif
                    FOREACH_BBL_SUCC(ibbl,iedg)
                    {
                        INDEX succ = AEDG_TAIL(iedg);
                        if( succ != next )
                            liveregs |= AbblRegsLiveIn(succ) & REGS_ALL_BUT_ZERO;                        
                    }
                }
                else
                {
                    liveregs = REGS_ALL_BUT_ZERO;
                }

                ScheduleeNewFancy(iins, INVALID,liveregs);
            }

            if( ABBL_TYPE(ibbl) != BT_BRANCH && ABBL_TYPE(ibbl) != BT_NORM ) break;
            
            if( AbblNpreds(next) != 1 ) break;
            if( ibbl < 5 ) break;
            if( next < 5 ) break;

            ibbl = next;
            ibbl_last = ibbl;
        }


        if( bblverbose )
        {
                printf("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n");
                printf("$$$$ BEFORE\n");
                printf("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n");

                next = ibbl_first;
                
                for(;;)
                {
                    PrintBblSuper(stdout,next);
                    if( next == ibbl_last ) break;
                    next = ABBL_NEXT(next);

                }
        }
        
        if( SCHEDULEES_USED > max_schedulees ) max_schedulees = SCHEDULEES_USED;
        

        /* if( ibbl_first < 5 || ibbl_first != PRINT_OPT_NUM)*/
        if( ibbl_first < 5 )
        {
            FOREACH_SCHEDULEE(ischedulee)
            {
                if( SCHEDULEE_IS_SCHEDULED(ischedulee) ) continue;
                AinsSchedule(SCHEDULEE_INSTR(ischedulee),0);
            }
        }
        else
        {
            
            /* create Constraints */

            CONSTRAINTS_USED = 0;

#if 1
                ScheduleesCreateResourceConstraints(0,0);
#else
                ScheduleesCreateResourceConstraints(0,1);
#endif
            ScheduleesCreateRegisterConstraints();
            ScheduleesCreateControlConstraints();
            ScheduleesCreateMemoryConstraints();

            if( CONSTRAINTS_USED > max_constraints ) max_constraints = CONSTRAINTS_USED;

            if( bblverbose )
            {
                printf("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n");
                printf("$$$$ ANTI DEP\n");
                printf("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n");
            }
            
#if 0
            FOREACH_CONSTRAINT(iconstraint)
            {
                if( CONSTRAINT_TYPE(iconstraint) != CT_REG_PROD_PROD &&
                    CONSTRAINT_TYPE(iconstraint) != CT_REG_CONS_PROD ) continue;
                if( !ConstraintIsOnlyLink(iconstraint) ) continue;
                
                nprodprod++;
                if( ABBL_IS_HOT(ibbl_first) )
                    nprodprodh++;

                if( ABBL_IS_HOT(ibbl_first) )
                {
                    PrintConstraint(stdout,iconstraint);
                }

            }
#endif
            
#if 0
            FOREACH_SCHEDULEE(ischedulee)
            {
                printf("schedulee: %d (%d)\n",ischedulee, SCHEDULEE_INSTR(ischedulee));
                FOREACH_SCHEDULEE_SUCC_CONSTRAINT(ischedulee,iconstraint)
                    PrintConstraint(stdout,iconstraint);
            }
#endif           
            
            /* create piece of information used by the scheduling heuristic*/

            
            ScheduleesComputeTimeToEnd();

#if 0
            PrintBblSuper(stdout,ibbl);
            FOREACH_SCHEDULEE(ischedulee)
                PrintSchedulee(stdout,ischedulee);
#endif               
            /* do the scheduling */

            FunctionalUnit.FltDivBusyTime = INVALID;
            FunctionalUnit.IntMulBusyTime = INVALID;
            
            ScheduleesComputeSchedule();
#if 0
            printf("AFTER SCHEDULING\n");
            PrintBblSuper(stdout,ibbl);
#endif       
        }

        /*
          unlink instructions and assign new bbl to those where it has changed
          this is done backward because ????
         */

        next = ibbl_last;
        for( ischlot = SCHLOTS_USED-1; ischlot >= first_schlot;ischlot--)
        {    
            iins = SCHLOT_INSTR(ischlot);
            if( AINS_RES_DEF(iins) & RES_PC )
                next = AINS_BBL(iins);
            
            AinsUnlink(iins);

            if( AINS_BBL(iins) != next )
            {
#if 1
                if( AINS_FLAGS(iins) & F_RELOC )
                {
                    WARNING(("moving reloc instr %d from %d to %d\n",iins,AINS_BBL(iins),next));
                    printf("OLD INS\n");                    
                    PrintIns(stdout,iins);
                    printf("OLD BBL\n");
                    PrintBblSuper(stdout,AINS_BBL(iins));
                    printf("NEW BBL\n");
                    PrintBblSuper(stdout,next);
                    SCHLOT_INSTR(ischlot) = AinsDup(iins);
                    AINS_FLAGS(SCHLOT_INSTR(ischlot)) &= ~F_RELOC;
                    AINS_FLAGS(iins) |= F_KILLED;
                    iins = SCHLOT_INSTR(ischlot);
                    printf("NEW INS\n");                    
                    PrintIns(stdout,iins);

                }
#endif           
                
#if 0
                if( AINS_CODE(iins) != I_unop )
                {
                    printf("non-unop instruction %d moved from %d to %d\n",iins,AINS_BBL(iins),next);
                }
#endif                
                AINS_BBL(iins) = next;
            }
            
        }

                
        /* relink instructions  */
        if( bblverbose )
        {
            printf("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n");
            printf("$$$$ SCHEDULE AFTER\n");
            printf("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n");
        }
        
        for( ischlot=first_schlot; ischlot < SCHLOTS_USED; ischlot++)
        {
            if( bblverbose )
            {
                printf("%5d %5d ",ischlot,SCHLOT_ISSUE_CYCLE(ischlot));
                PrintIns(stdout,SCHLOT_INSTR(ischlot));
                if( (ischlot & 3) == 3 )
                    printf("**\n");
            }
            
            AinsAppend(SCHLOT_INSTR(ischlot));
        }

        if( bblverbose )
        {
            printf("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n");
            printf("$$$$ BBL AFTER\n");
            printf("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n");
            
            next = ibbl_first;
                
            for(;;)
            {
                PrintBblSuper(stdout,next);
                if( next == ibbl_last ) break;
                next = ABBL_NEXT(next);
            }

        }
        
    }
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    

    InfoStatus(fp,"\t[Schedule] relcon: %d#%d unops: %d   maxs: %d  maxc: %d lastcycle: %d  {%d ms}\n",
            nprodprod, nprodprodh, added_unops, max_schedulees, max_constraints, SCHLOT_ISSUE_CYCLE(SCHLOTS_USED-1),time);

}
/*=======================================================
EOF
=======================================================*/

