/*********************************************************************************
  ComputeUDChains - computes use-define chains within each function.
  ********************************************************************************
   
  QUESTIONS/TODO:

  o WARNING!!! - AFUN_SAVED_REGS(f) are treated as though they were never
    not defined in f.

  o add unknowns for pall code
  o does changed registers include functions called within functions?

  o TEST TEST TEST TEST TEST
  o redo expirment with size
  o verify correctness (compare with PE)
  o add kill/gen? (probably isn't worth it)
  o improve dsAddSet? (smart merge?)
  o add work list?
  o use memcopy or memmove in dsAddInstr?
  o stats bbl size for space efficiency
  o conserve space at the end of analysis?
  o can you make it work on a function at a time?

  Context sensitive interprocedural analysis:
  
  First compute the local kill set for each basic block using a
  fixpoint computation over each function.  This represents the set of
  definitions that are definitely kill along any path from the
  function entry point to the program point after a given basic
  block. (The kill[b] includes the definitions killed by b).

  Next compute the IN/OUT in the usual way for edges *within* the
  function.  In the case where basic block p, calls a function with
  exit block f, and returns to block b, compute the IN[b] as follows:

  IN[b] = IN[b] U ((OUT[p] - (kill[f] - saved[f]))) U OUT[f]

  where the OUT[f] is the out of f's exit block, saved[f] is the
  registers saved and restored by f and kill[f] is the local kill set
  as defined above.  
  
  Imprecision in CFG: (look in unreachable code) jsrs - after strength
  reductions functions f that are called from nowhere?  look for
  F_RELOC ABBL_FLAGS(AFUN_BBL_INIT(f))&F_GPREL32 relocatable basic
  blocks - ABBL_FLAGS(b)&F_GPREL32||ABBL_FLAGS(b)&F_REFQUAD has an
  edge from the zero block could be jumped to from somewhere else
  
  */


#include "global.h"
#include "functions.h"
#include "regs.h"
#include "flags.h"
#include <stdlib.h>
#include <string.h>

/* @@@@ NEEDS WORK */

#define AINS_IS_STRONG(ii) (\
AINS_TYPE(ii) != FMT_PAL &&\
AINS_TYPE(ii) != FMT_ICM &&\
AINS_TYPE(ii) != FMT_FCM)

#define ALL_ARG_REGS REGSET_UNION(ARG_REGS, FARG_REGS)
#define ALL_RET_REGS REGSET_UNION(RET_REGS, FRET_REGS)
#define REG_ZEROS    REGSET_UNION(B_ZERO, B_FZERO)

#define AINSDEF(ins) REGSET_DIFF(REGSET_SINGLETON(AINS_REG_C(ins)), REG_ZEROS)
     
#define UD_USELIVE

/*
  ADT - definition set
  set of pairs (a, i) where a in some register defined by instruction i
  
  operations on definition set
  dsEmpty(S) => S := Empty Set
  dsAddInstr(S, i) S := S Union {(destReg(i), i)}
  dsCopy(A, B) A := B
  dsAddSet(A, B) A := A Union B
  dsRemoveAll(A, r) A := A - D_r, where D_r is set of all defs of reg r
  dsContains(S, i) returns true iff i in S

  obj iterator -  to get all definitions (id) of a particular
      use of reg r, with instruction ii
      
      for(id = dsIterInit(r, ii); VALID(id); id = dsIterNext())

  IMPLEMENTATION NOTES:
  
  A defSet, or definition set, is a list of branches, each of which is
  a sorted array of instruction indexes.  Each branch is
  correspondingly sorted.  While maintain the sorted order list my for
  random data be O(n), since the instructions are generally processed
  in order the cost was found to be relatively small, for the
  advantage of being able to quickly find duplicates.

  The initial branch of each instruction has a bit vector that
  indicates which (if any) registers may have definitions that could
  not be located due to imprecision in the CFG.

  */
#ifndef DS_SET_MIN
#define DS_SET_MIN 40  /* must be less than 256 */
#endif
     
#if DS_SET_MIN >= 256
#error "DS_SET_MIN must be less than 256!"
#endif

typedef struct defSet {
  INDEX instr[DS_SET_MIN];  /* instruction index */
  UINT8 end;                /* points to next free slot in instr */
  UINT8 size;               /* number of valid instr's (only used in remove) */
  REGSET unknown;            /* set of registers that may have definitions
			       I could not find */
  struct defSet *next;
} defSet;

static void  dsEmpty(defSet *s);
static void  dsFree(defSet *s);
static BOOL  dsAddInstr(defSet *s, INDEX i);
static void  dsAddtoEnd(defSet *s, INDEX i);
static void  dsAddtoBegin(defSet *s, INDEX i);
static void  dsCopy(defSet *dest, defSet *src);
static BOOL  dsAddSet(defSet *a, defSet *b, REGSET except);
static BOOL  getNextHoleAndInstr(defSet **hb, INDEX *hi, defSet **ib, INDEX *ii);
static void  dsRemoveAll(defSet *A, REGSET rl);

extern INDEX dsIterInit(REG r, INDEX ii);
extern INDEX dsIterNext(void);

extern BOOL BblUnknownEdg(INDEX ib);
extern INDEX AbblReturn2link(INDEX ret);
extern void ComputeUDChains();

#ifndef FOREACH_REACHING_DEF
#define FOREACH_REACHING_DEF(r, ins, ii)\
if (r != REG_LIT)\
for(ii = dsIterInit(r, ins); VALID(ii); ii = dsIterNext()) \
if (VALID(ii))
#endif

/**********
  globals
  *********/
defSet *IN = NULL;
int maxbbl = 0;

#ifdef UD_DEBUG
int callocCount = 0;
int freeCount = 0;
#define cCount (callocCount++)
#define fCount (freeCount++)
#else
#define cCount
#define fCount 
#endif

static void
printem(FILE *f, defSet *s) {
  int i;
  defSet *p;

  for (p = s; p; p = p->next)
    for (i = 0; i < p->end; i++) {
      fprintf(f, "BBL: %d INS: %d", p->instr[i], AINS_BBL(p->instr[i]));
      PrintIns(f, p->instr[i]);
    }
}



/*
  find the basic block that called the function that returns on edge r
  */
#undef __PROC__
#define __PROC__ "AbblReturn2Link"

INDEX
AbblReturn2Link(INDEX r) {
  INDEX c = r - 1; /*  c should be the call edge for this call site */
  INDEX b;
  CASSERT(c < CONFIG_UEDG && c >= 0, ("invalid edge\n"));

  b = AEDG_FROM(c);
  CASSERT(b < CONFIG_UBBL && b >= 0, ("invalid block\n"));

  return b;
} /* AbblLinkt2Return */



BOOL
BblUnknownEdg(INDEX ib) {
  INDEX p;
  BOOL r = FALSE;

  /* this case is dealt with in the fix point computation
  FOREACH_BBL_PRED(ib, p) 
    if (AEDG_FROM(p) == ABBL_JMP)
      r = TRUE; */

  return r ||
    (ABBL_IS_FIRST(ib) && (ABBL_FLAGS(ib)&(F_REFQUAD | F_GPREL32)));

} /* BblUnknownEdg */



static void
dsEmpty(defSet *s) {
  s->unknown = REGSET_THE_EMPTY_SET;
  s->end = 0;
  if (s->next != NULL) {
      dsFree(s->next);
      s->next = NULL;
  }
} /* dsEmpty */



#undef __PROC__
#define __PROC__ "dsFree"
/* assumes s != NULL */
static void
dsFree(defSet *s) {
  CASSERT(s, ("s->next == NULL\n"));
  if (s->next != NULL)
    dsFree(s->next);

  free(s); fCount;
} /* dsFree */


#undef __PROC__
#define __PROC__ "dsAddtoEnd"

/* add an elm to the end of branch p */
static void
dsAddtoEnd(defSet *p, INDEX ins) {
  if (p->end < DS_SET_MIN) {
    /* we have room in this branch */
    p->instr[p->end] = ins;
    p->end++;
  }
  else if (p->next == NULL) {
    /* need to allocate more room */
    p->next = calloc(1, sizeof(defSet)); cCount;
    CASSERT(p->next, ("Can't allocate\n"));
    p->next->instr[0] = ins;
    p->next->end = 1;
  }
  else 
    dsAddtoBegin(p->next, ins);
} /* dsAddtoEnd */


/* add an elm to the beginning of branch s */
static void
dsAddtoBegin(defSet *s, INDEX ins) {
  int i, tmp;

  for (i = 0; i < s->end; i++, ins = tmp) {
    tmp = s->instr[i];
    s->instr[i] = ins;
  }

  dsAddtoEnd(s, ins);
} /* dsAddtoBegin */

#undef __PROC__
#define __PROC__ "dsAddInstr"

/* returns true if this operation changed s */
static BOOL
dsAddInstr(defSet *s, INDEX ins) {
  defSet *p;
  int i = 0;
  int h, l, m;  /* high, low and middle for bin search */
  int tmp;

  if (s->end == 0) {
    /* set is empty */
    s->end++;
    s->instr[0] = ins;
    return TRUE;
  }
  
  /* maintain a sorted order (sort on ins index)*/

  for (p = s; p->next && p->instr[p->end-1] < ins; p = p->next)
    /* noop */;

  /* CP - ins goes in branch p (if not already there)
          ins <= p->instr[p->end-1] or p->next == NULL */

  
  /* check the boundary conditions */
  if (ins > p->instr[p->end-1]) {
    /* ins goes in the end */
    dsAddtoEnd(p, ins);
    return TRUE;
  }

  if (ins == p->instr[p->end-1] || ins == p->instr[0]) 
    /* ins is in the last/first slot */
    return FALSE;
  

  if (ins < p->instr[0]) {
    dsAddtoBegin(p, ins);
    return TRUE;
  }
    
  /* CP - p->instr[0] < ins < p->instr[p->end] */
  
  /* binary search for ins */
  h = p->end;
  l = 0;
  do {
    m = (h + l)/2;
    
    if (ins == p->instr[m]) 
      return FALSE;
    
    if (ins < p->instr[m])
      h = m;
    else
      l = m;
 
  } while(h > l+1);

  /* CP - p->instr[l] < ins < p->instr[h] AND h = l+1
          so h is the slot we want
   */

  /* put ins in the slot and move everyone over */
  for (i = h; i < p->end; i++, ins = tmp) {
    tmp = p->instr[i];
    p->instr[i] = ins;
  }

  /* put the last one at the end of this branch */
  dsAddtoEnd(p, ins);

  /* we changed the set */
  return TRUE;
} /* dsAddInstr */


#undef __PROC__
#define __PROC__ "dsCopy"

/* assumes space for first branch in dest */
static void
dsCopy(defSet *dest, defSet *src) {
  defSet *pd, *ps;

  CASSERT(dest && src, ("src or dest null\n"));
  dsEmpty(dest);
  for (pd = dest, ps = src; ps; ps = ps->next, pd = pd->next) {
    bcopy((char *)ps, (char *)pd, sizeof(defSet)); 
    
    if (ps->next) {
      /* we need space for the next branch */
      pd->next = calloc(1, sizeof(defSet));  cCount;
      CASSERT(ps->next, ("Can't allocate\n"));
    } /* if  */
  } /* for */
} /* dsCopy */



/* implements union operation - i.e. a := a U b
   doesn't add instructions that define registers in except
   returns true if this operation changed a */
static BOOL
dsAddSet(defSet *a, defSet *b, REGSET except) {
  defSet *p;
  int i;
  BOOL changed = FALSE;

  /* add unknown definitions */
  a->unknown = REGSET_UNION(a->unknown, REGSET_DIFF(b->unknown, except));
  
  for (p = b; p; p = p->next)
    for (i = 0; i < p->end; i++) {
      if (REGSET_IS_SUBSET(AINSDEF(p->instr[i]), except))
	/* don't add if instr defines a register in except */
	continue;
      
      changed = dsAddInstr(a, p->instr[i]) || changed;
    } /* for */

  return changed;
} /* dsAddSet */



/* utility routine for removall finds the next "hole" and then the
   next valid instruction after that */
static BOOL
getNextHoleAndInstr(defSet **hb, INDEX *hi, defSet **ib, INDEX *ii) {
  /* a hole is an instruction marked as invalid or space at the end of
     a branch */
  BOOL found;

  if (*hb == NULL)
    /* none left */
    return FALSE;
  
  /* next hole */
  found = FALSE;
  (*hi)++;  /* start one past the previous hole */
  do {
    if (!VALID((*hb)->instr[*hi]) || *hi > (*hb)->end)
      /* found a invalid instr, or space at the end of the branch */
      found = TRUE;
    else {
      /* inc (keep looking) */
      (*hi)++;
      if ((*hi) >= DS_SET_MIN) {
	(*hi) = 0;
	*hb = (*hb)->next;
      } /* if */
    } /* else */      
  } while(*hb && !found);

  if (!found)
    /* didn't find a hole, must be done */
    return FALSE;

  /* get the next instruction past the current hole */
  *ib = *hb;
  *ii = *hi;
  found = FALSE;
  do {
    /* increment */
    (*ii)++;
    if (*ii >= (*ib)->end) {
      *ii = 0;
      *ib = (*ib)->next;
    }
    
    if (*ib && VALID((*ib)->instr[*ii]))
      found = TRUE;
  } while(*ib && !found);

  if (!found)
    /* didn't find a valid instr, must be done */
    return FALSE;
  
  return TRUE;    
} /* getNextHoleAndInstr */

#undef __PROC__
#define __PROC__ "dsRemoveAll"

/* remove all instructions in the reg list rl,
   must maintain sorted order */
static void
dsRemoveAll(defSet *A, REGSET rl) {
  defSet *p;
  int i;

  /* for compression */
  INDEX ii;    /* index of next instruction */
  defSet *ib;  /* branch of next instruction */
  INDEX hi;    /* index of next hole */
  defSet *hb;  /* branch of next hole */

  if (A->end == 0)
    /* A is empy */
    return;

  /* remove unknow definitions */
  A->unknown = REGSET_DIFF(A->unknown, rl);
  
  /* mark definitions for removal */
  for (p = A; p; p = p->next) {
    p->size = p->end;
    for (i = 0; i < p->end; i++) {
      CASSERT(VALID(p->instr[i]), ("Invalid instruction in def. set\n"));
#ifdef UD_DEBUG
      if (!VALID(p->instr[i]) || p->instr[i] >= CONFIG_UINS)
	fprintf(stderr, "Bad News!!!\n");
#endif
      if (REGSET_IS_SUBSET(AINSDEF(p->instr[i]),rl)) {
	/* mark this instruction as removed for later */
	p->instr[i] = INVALID;
	p->size--;
      } /* if */
    } /* for */
  } /* for */

  /* init the hb, hi */
  hb = A;
  hi = -1;
  
  /* compress */
  while(getNextHoleAndInstr(&hb, &hi, &ib, &ii)) {
    /* move instruction to the hole */
    hb->instr[hi] = ib->instr[ii];
    hb->size++;

    /* mark the location of the instruction as a hole */
    ib->instr[ii] = INVALID;
    ib->size--;
  }

  /* set end to be the size (after compression this should be true) */
  A->end = A->size;
  for (ib = A; ib; ib = ib->next)
    ib->end = ib->size;

  /* find the last full branch */
  for (ib = A; ib && ib->end == DS_SET_MIN; ib = ib->next)
    /* noop */;
  
  /* ib points to the first branch with less than DS_SET_MIN elms */
  if (ib && ib->next) {
    dsFree(ib->next);
    ib->next = NULL;
  }
} /* dsRemoveAll */



static defSet *curr_ds; /* definition set for this iter */
static INDEX curr_i;    /* instr index into the bbl for this iter */
static INDEX curr_ei;   /* index into the def set for this iter */
static BOOL iterDone;       
static REGLIST curr_rl; /* register used for this iter */

#undef __PROC__
#define __PROC__ "dsIterInit"

INDEX
dsIterInit(REG r, INDEX ii) {
  /* initialize state to remember */
  if (AINS_BBL(ii) >= maxbbl) {
    WARNING(("Invalid instruction (May need to rerun ComputeUDChains)"));
    iterDone = TRUE;
    return INVALID;
  }
  curr_ds = &IN[AINS_BBL(ii)];
  curr_i = AINS_PREV(ii);
  iterDone = FALSE;
  curr_rl = REG2BIT(r);
  curr_ei = 0;

  return dsIterNext();
} /* dsIterInit */


 INDEX
dsIterNext(void) {
  BOOL found;
  INDEX reti = INVALID;
  
  if (iterDone)
    return INVALID;
  
  /* check for a definition in the current block */
  for (found = FALSE; !found && VALID(curr_i); curr_i = AINS_PREV(curr_i)) 
    if (REGSET_INTERSECT(AINSDEF(curr_i), curr_rl)) {
      
      if (AINS_IS_STRONG(curr_i))
	/* strong update */
	iterDone = TRUE;
      else 
	iterDone = FALSE;
      
      reti = curr_i;
      found = TRUE;
    }

  if (found)
    return reti;

  /* none found, now try in the current def set */
  found = FALSE;
  do {
    if (REGSET_INTERSECT(AINSDEF(curr_ds->instr[curr_ei]), curr_rl)){
      reti = curr_ds->instr[curr_ei];
      found = TRUE;
    } /* if */

    /* increment */
    curr_ei++;
    if (curr_ei >= curr_ds->end) {
      curr_ei = 0;
      curr_ds = curr_ds->next;
    }

    /* are we done ? */
    if (!curr_ds)
      iterDone = TRUE;
  } while(curr_ds && !found);

  if (found)
    return reti;
  
  /* none found */
  iterDone = TRUE;
  return INVALID;
  
} /* dsIterNext */

/************
  stats
  ***********/
#if 0 /* def UD_DEBUG */
  {
    /* hist[i] counts the number of defSets that have i instructions */
    int hist[DS_SET_MIN], i;
    INDEX b;
       
    for (i = 0; i < DS_SET_MIN; i++)
      hist[i] = 0;
    
    FOREACH_BBL(b) {
      hist[IN[b].end]++;
    }
    
    for (i = 0; i < DS_SET_MIN; i++)
      fprintf(stderr, "%d %d\n", i, hist[i]);
  }
#endif

#undef __PROC__
#define __PROC__ "ComputeUDChains"
#ifndef B_FZERO
#define B_FZERO B_FZER0
#endif

void
ComputeUDChains() {
  /* index vars */
  INDEX b; /* basic blocks */
  INDEX p; /* basic block predecessors */
  INDEX i; /* instructions */
  INDEX f;
  REGLIST rl;
  defSet *OUT;
  defSet tmp;
  BOOL change, localChange;

  /* required analysis... */
  ComputeSavedChangedRegisters();
  ComputeLiveness();

  tmp.next = NULL;
  tmp.end = 0;
#ifdef UD_DEBUG
  setbuf(stderr, NULL);
  fprintf(stderr, "********************************************************\n");
  fprintf(stderr, "ComputeUDChains\n");
  fprintf(stderr, "********************************************************\n");
#endif  

#ifdef UD_DEBUG
  fprintf(stderr, "ComputeUDChains - allocate..");
#endif  

  maxbbl = CONFIG_UBBL;
  IN = calloc(CONFIG_UBBL, sizeof(defSet));
  OUT = calloc(CONFIG_UBBL, sizeof(defSet));
  CASSERT(IN && OUT,
	  ("Can't allocate IN, OUT, killin, or killout (Total bytes 2x%dx%d = %d)\n",
	  CONFIG_UBBL, sizeof(defSet), 2*CONFIG_UBBL*sizeof(defSet)));
	  

#ifdef UD_DEBUG
  fprintf(stderr, "Total bytes 2x%dx%d = %d)\n",
	  CONFIG_UBBL, (int)sizeof(defSet),
	  (CONFIG_UBBL)*2*(int)sizeof(defSet));
  fprintf(stderr, "ComputeUDChains - init IN and OUT\n");
#endif  
  /* initialize IN/OUT and kill for all blocks */
  FOREACH_BBL(b) {
    dsEmpty(&IN[b]);
    dsEmpty(&OUT[b]);

    if (BblUnknownEdg(b)) 
      IN[b].unknown = ALL_REGS;

    /* compute OUT as all the defitions generated by basic block b */
    FOREACH_BBL_INS(b, i) {
      if ((rl = REGSET_DIFF(AINSDEF(i), REG_ZEROS))) {
	/* if this instr defines non zero registers */

	if (AINS_TYPE(i) == FMT_PAL)
	  /* what does pal define?? */
	  OUT[b].unknown = ALL_REGS;

	if (AINS_IS_STRONG(i))
	  /* this instruction kills previous definitions */
	  dsRemoveAll(&OUT[b], rl);
	
	dsAddInstr(&OUT[b], i);
      } /* if */
    } /* FOREACH_BBL_INS */
  } /* FOREACH_BBL */

  OUT[ABBL_JMP].unknown = ALL_REGS;
  
  /* compute IN/OUT with fixpoint computation  */
  do {
    
#ifdef UD_DEBUG
    fprintf(stderr, "fix point. overflow = %d(%d - %d)\n",
	    callocCount - freeCount, callocCount, freeCount);
#endif
    change = FALSE;
    FOREACH_BBL(b) {

      /* its dumb to do this for the dummy */
      if (b == 0)
	continue;

#ifdef UD_DEBUG
      if (ABBL_FUN(b) == 78)
	putc('.', stderr);
#endif
      
      /* IN[b] is union of the OUT[p] for all predecessors p */
      localChange = FALSE; 
      FOREACH_BBL_PRED(b, p) {
	CASSERT(p >= 0 && AEDG_FROM(p) >= 0 && AEDG_FROM(p) < CONFIG_UBBL,
		("bad edge"));
	if (AEDG_FLAGS(p)&FE_FUN_LINK)
	  /* LINK - deal with the function call on the corresponding return edge
	     (so we know which function is called)
	   */
	  continue;
	else if (AEDG_FLAGS(p)&FE_FUN_CALL) {
	  /* propagate only arg regs into the function */
	  rl = ALL_ARG_REGS;
#ifdef UD_USELIVE
	  rl = REGSET_INTERSECT(rl, ABBL_IN(b));
#endif
	  localChange = dsAddSet(&IN[b], &OUT[AEDG_FROM(p)], REGSET_COMPL(rl)) || localChange;

	  /* other registers add unknown definitions */
	  rl = REGSET_UNION(IN[b].unknown, REGSET_COMPL(rl));
	  if (IN[b].unknown != rl) {
	    localChange = TRUE;
	    IN[b].unknown = rl;
	  }
	} /* else if call */
	else if (AEDG_FLAGS(p)&FE_FUN_RETURN) {	  
	  /* RETURN EDG
	     propagate only return regs from the function return */
	  rl = ALL_RET_REGS;
#ifdef UD_USELIVE
	  rl = REGSET_INTERSECT(rl, ABBL_IN(b));
#endif
	  localChange = dsAddSet(&IN[b], &OUT[AEDG_FROM(p)], REGSET_COMPL(rl)) || localChange;

	  /* add unknown defs of everything else that the function may have changed */
	  f = ABBL_FUN(AEDG_FROM(p));
	  rl = REGSET_UNION(IN[b].unknown, REGSET_DIFF(AFUN_CHANGED_REGS(f), rl));
	  if (IN[b].unknown != rl) {
	    localChange = TRUE;
	    IN[b].unknown = rl;
	  }
	  
	  /* LINK EDG */
	  /* everything makes it through the function (nothing killed)*/
	  rl = REGSET_THE_EMPTY_SET;

#ifdef UD_USELIVE
	  rl = REGSET_UNION(rl, REGSET_COMPL(ABBL_IN(b)));
#endif
 	  localChange = dsAddSet(&IN[b], &OUT[AbblReturn2Link(p)], rl) || localChange;
	  
	} /* else if return */
	else {
#ifdef UD_USELIVE
	  rl = REGSET_COMPL(ABBL_IN(b));
#else
	  rl = REGSET_THE_EMPTY_SET;
#endif
	  /* normal edge */
 	  localChange = dsAddSet(&IN[b], &OUT[AEDG_FROM(p)], rl) || localChange;
	} /* else */
      } /* FOREACH_BBL_PRED */
      
      /* if IN changed recompute OUT */
      if (localChange) {

	/* compute the affect of block on the IN */
	dsCopy(&tmp, &IN[b]);
	FOREACH_BBL_INS(b, i) {
	  if ((rl = REGSET_DIFF(AINSDEF(i), REG_ZEROS))) {
	    /* if this instr defines non zero registers */

	    if (AINS_IS_STRONG(i)) 
	      dsRemoveAll(&tmp, rl);

	    dsAddInstr(&tmp, i);
	  } /* if */
	} /* FOREACH */
	
	/* its possible that all the changes to IN will be killed
	   so recompute changed flag */
	localChange = dsAddSet(&OUT[b], &tmp, REGSET_THE_EMPTY_SET);
      } /* if (localChange) */
      
      /* keep track of previous changes, if any */ 
      change = change || localChange;
    } /* foreach basic block */
  } while(change);
  
#ifdef UD_DEBUG
  fprintf(stderr, "\nComputeUDChains - free \n");
#endif      

#if 0 /* def UD_DEBUG */
  b = PRINT_OPT_NUM;
  fprintf(stderr, "overflow = %d\n", callocCount - freeCount);
  if (b != -1) {
    fprintf(stderr, "*** IN[%d] ***\n", b);
    printem(stderr, &IN[b]);
    fprintf(stderr, "*** OUT[%d] ***\n", b);
    printem(stderr, &OUT[b]);
    fprintf(stderr, "blocks reaching %d\n", b);
    FOREACH_BBL_PRED(b, p) 
      if (!AEDG_IS_INTER(p)) {
	/* stay inside this function */
	fprintf(stderr, "**** OUT[%d] ***\n", AEDG_FROM(p));
	printem(stderr, &OUT[AEDG_FROM(p)]);
      }
  }
#endif

  /* free extra branches */
  FOREACH_BBL(b) 
      dsEmpty(&OUT[b]);
  
  /* free OUT array */
  free(OUT);


#ifdef UD_DEBUG
  if (PRINT_OPT_NUM != -1) {
    int instr = Addr2Ins(PRINT_OPT_NUM);
    int i;
    REG reg;
    
    fprintf(stderr, "%s definitions Reaching BBL: %d    ",
	    REGSET_IS_SUBSET(AinsUse(instr), IN[AINS_BBL(instr)].unknown) ? "Some" : "All",
	    AINS_BBL(instr));
    PrintIns(stderr, instr);
    FOREACH_REACHING_DEF(AINS_REG_A(instr), instr, i) {
      fprintf(stderr, "BBL: %d    ", AINS_BBL(i));
      PrintIns(stderr, i);
    }
    FOREACH_REACHING_DEF(AINS_REG_B(instr), instr, i) {
      fprintf(stderr, "BBL: %d    ", AINS_BBL(i));
      PrintIns(stderr, i); 
    }

    fprintf(stderr, "Reaching Definitions for BBL: %d\n", AINS_BBL(instr));
    printem(stderr, &IN[AINS_BBL(instr)]);
    fprintf(stderr, "Unknown registers: ");
    REGSET_FOREACH_REG(IN[AINS_BBL(instr)].unknown, reg)
      fprintf(stderr, "%3d", reg);
    putc('\n', stderr);
  }
#endif

} /* computeUDChains */
